@@ -52,7 +52,7 @@ inline bool isStdOut(const std::string& out){ return out == "-" || out == "stdou
 // ClaspAppOptions
 /////////////////////////////////////////////////////////////////////////////////////////
 namespace Cli {
-ClaspAppOptions::ClaspAppOptions() : outf(0), onlyPre(false), printPort(false), ifs(' '), outLbd(Activity::MAX_LBD), inLbd(Activity::MAX_LBD) {
+ClaspAppOptions::ClaspAppOptions() : outf(0), ifs(' '), hideAux(false), onlyPre(false), printPort(false), outLbd(Activity::MAX_LBD), inLbd(Activity::MAX_LBD) {
 	quiet[0] = quiet[1] = quiet[2] = static_cast<uint8>(UCHAR_MAX);
 }
 void ClaspAppOptions::initOptions(ProgramOptions::OptionContext& root) {
@@ -65,14 +65,16 @@ void ClaspAppOptions::initOptions(ProgramOptions::OptionContext& root) {
 		 "      <o>: print {0=all|1=last|2=no} optimize values [<m>]\n"
 		 "      <c>: print {0=all|1=last|2=no} call steps      [2]")
 		("pre" , flag(onlyPre), "Run preprocessing and exit")
-		("print-portfolio,g" , flag(printPort), "Print default portfolio and exit")
+		("print-portfolio" , flag(printPort), "Print default portfolio and exit")
 		("outf,@1", storeTo(outf)->arg("<n>"), "Use {0=default|1=competition|2=JSON|3=no} output")
+		("out-atomf,@1" , storeTo(outAtom), "Set atom format string (<Pre>?%%s<Post>?)")
+		("out-ifs,@1"   , notify(this, &ClaspAppOptions::mappedOpts), "Set internal field separator")
+		("out-hide-aux,@1", flag(hideAux), "Hide auxiliary atoms in answers")
 		("lemma-out,@1" , storeTo(lemmaOut)->arg("<file>"), "Write learnt lemmas to %A on exit")
 		("lemma-out-lbd,@1",notify(this, &ClaspAppOptions::mappedOpts)->arg("<n>"), "Only write lemmas with lbd <= %A")
 		("lemma-in,@1"  , storeTo(lemmaIn)->arg("<file>"), "Read additional lemmas from %A")
 		("lemma-in-lbd,@1", notify(this, &ClaspAppOptions::mappedOpts)->arg("<n>"), "Initialize lbd of additional lemmas to <n>")
 		("hcc-out,@1", storeTo(hccOut)->arg("<file>"), "Write non-hcf programs to %A.#scc")
-		("ifs,@1", notify(this, &ClaspAppOptions::mappedOpts), "Internal field separator")
 		("file,f,@2", storeTo(input)->composing(), "Input files")
 	;
 	root.add(basic);
@@ -86,7 +88,7 @@ bool ClaspAppOptions::mappedOpts(ClaspAppOptions* this_, const std::string& name
 		for (int i = 0; i != parsed; ++i) { this_->quiet[i] = static_cast<uint8>(q[i]); }
 		return parsed && *err == 0;
 	}
-	else if (name == "ifs") {
+	else if (name == "out-ifs") {
 		if (value.empty() || value.size() > 2) { return false;}
 		if (value.size() == 1) { this_->ifs = value[0]; return true; }
 		if (value[1] == 't')   { this_->ifs = '\t'; return true; }
@@ -107,8 +109,8 @@ bool ClaspAppOptions::validateOptions(const ProgramOptions::ParsedOptions&) {
 /////////////////////////////////////////////////////////////////////////////////////////
 // ClaspAppBase
 /////////////////////////////////////////////////////////////////////////////////////////
-ClaspAppBase::ClaspAppBase() {}
-ClaspAppBase::~ClaspAppBase(){}
+ClaspAppBase::ClaspAppBase() { }
+ClaspAppBase::~ClaspAppBase(){ }
 const int* ClaspAppBase::getSignals() const {
 	static const int signals[] = { 
 		SIGINT, SIGTERM
@@ -127,19 +129,18 @@ bool ClaspAppBase::parsePositional(const std::string& t, std::string& out) {
 void ClaspAppBase::initOptions(ProgramOptions::OptionContext& root) {
 	claspConfig_.addOptions(root);
 	claspAppOpts_.initOptions(root);
-	ProgramOptions::OptionContext::option_iterator it = root.find("verbose");
-	it->get()->value()->defaultsTo("1");
+	root.find("verbose")->get()->value()->defaultsTo("1");
 }
 
-void ClaspAppBase::validateOptions(const ProgramOptions::OptionContext&, const ProgramOptions::ParsedOptions& parsed) {
-	if (!claspAppOpts_.validateOptions(parsed) || !claspConfig_.finalize(parsed)) {
-		error("command-line error!");
-		exit(E_NO_RUN);
-	}
+void ClaspAppBase::validateOptions(const ProgramOptions::OptionContext&, const ProgramOptions::ParsedOptions& parsed, const ProgramOptions::ParsedValues& values) {
 	if (claspAppOpts_.printPort) {
 		printTemplate();
 		exit(E_UNKNOWN);
 	}
+	if (!claspAppOpts_.validateOptions(parsed) || !claspConfig_.finalize(parsed, getProblemType(), true)) {
+		error("command-line error!");
+		exit(E_NO_RUN);
+	}
 	ClaspAppOptions& app = claspAppOpts_;
 	if (!app.lemmaOut.empty() && !isStdOut(app.lemmaOut)) {
 		if (std::find(app.input.begin(), app.input.end(), app.lemmaOut) != app.input.end() || app.lemmaIn == app.lemmaOut) {
@@ -151,6 +152,13 @@ void ClaspAppBase::validateOptions(const ProgramOptions::OptionContext&, const P
 		error("'lemma-in': could not open file!");
 		exit(E_NO_RUN);
 	}
+	for (std::size_t i = 1; i < app.input.size(); ++i) {
+		if (!isStdIn(app.input[i]) && !std::ifstream(app.input[i].c_str()).is_open()) {
+			error(clasp_format_error("'%s': could not open input file!", app.input[i].c_str()));
+			exit(E_NO_RUN);
+		}
+	}
+	storeCommandArgs(values);
 }
 void ClaspAppBase::setup() {
 	ProblemType pt = getProblemType();
@@ -163,7 +171,7 @@ void ClaspAppBase::setup() {
 		EventHandler::setVerbosity(Event::subsystem_load   , verb);
 		EventHandler::setVerbosity(Event::subsystem_prepare, verb);
 		EventHandler::setVerbosity(Event::subsystem_solve  , verb);
-		clasp_->ctx.eventHandler(this);
+		clasp_->ctx.setEventHandler(this);
 	}
 	else if (pt != Problem_t::ASP) { 
 		error("Option '--pre' only supported for ASP!"); 
@@ -252,26 +260,30 @@ int ClaspAppBase::exitCode(const RunSummary& run) const {
 }
 
 void ClaspAppBase::printTemplate() const {
-	printf("# clasp %s portfolio file\n", CLASP_VERSION);
-	printf("# A portfolio file contains a (possibly empty) list of configurations.\n"
-	       "# Each of which must have the following format:\n"
-	       "#   [<name>]: <cmd>\n"
-	       "# where <name> is a string that must not contain ']'\n"
-	       "# and   <cmd>  is a command-line style list of options\n"
-	       "# from \"Search Options\" and/or \"Lookback Options\".\n"
-	       "#\n"
-	       "# SEE: clasp --help\n"
-	       "#\n"
-	       "# NOTE: Options given on the command-line are added to all configurations in a\n"
-	       "#       portfolio file. If an option is given both on the command-line and in a\n"
-	       "#       portfolio configuration, the one from the command-line is preferred.\n"
-	       "#\n"
-	       "# NOTE: If, after adding command-line options, a portfolio configuration\n"
-	       "#       contains mutually exclusive options an error is raised.\n"
-	       "#\n");
-	for (const char* p = ClaspCliConfig::getPortfolio(ClaspCliConfig::portfolio_def); *p;) {
-		printf("%s\n", p);
-		p += strlen(p) + 1;
+	printf(
+		"# clasp %s configuration file\n"
+		"# A configuration file contains a (possibly empty) list of configurations.\n"
+		"# Each of which must have the following format:\n"
+		"#   <name>: <cmd>\n"
+		"# where <name> is a string that must not contain ':'\n"
+		"# and   <cmd>  is a command-line string of clasp options in long-format, e.g.\n"
+		"# ('--heuristic=vsids --restarts=L,100').\n"
+		"#\n"
+		"# SEE: clasp --help\n"
+		"#\n"
+		"# NOTE: The options '--configuration' and '--tester' must not occur in a\n"
+		"#       configuration file. Furthermore, global options are ignored in all\n"
+		"#       but the first configuration.\n"
+		"#\n"
+		"# NOTE: Options given on the command-line are added to all configurations in a\n"
+		"#       configuration file. If an option is given both on the command-line and\n"
+		"#       in a configuration file, the one from the command-line is preferred.\n"
+		"#\n"
+		"# NOTE: If, after adding command-line options, a configuration\n"
+		"#       contains mutually exclusive options an error is raised.\n"
+		"#\n", CLASP_VERSION);
+	for (ConfigIter it = ClaspCliConfig::getConfig(Clasp::Cli::config_many); it.valid(); it.next()) {
+		printf("%s: %s\n", it.name(), it.args());
 	}
 }
 
@@ -309,10 +321,10 @@ void ClaspAppBase::printHelp(const ProgramOptions::OptionContext& root) {
 void ClaspAppBase::printDefaultConfigs() const {
 	uint32 minW = 2, maxW = 80;
 	std::string cmd;
-	for (int i = ClaspCliConfig::config_default+1; i != ClaspCliConfig::config_max_value+1; ++i) {
-		ClaspCliConfig::NamedConfig nc = ClaspCliConfig::getDefaults(static_cast<ClaspCliConfig::DefaultConfig>(i));
-		printf("[%s]:\n%*c", nc.name, minW-1, ' ');
-		cmd = nc.opts;
+	for (int i = Clasp::Cli::config_default+1; i != Clasp::Cli::config_default_max_value; ++i) {
+		ConfigIter it = ClaspCliConfig::getConfig(static_cast<Clasp::Cli::ConfigKey>(i));
+		printf("%s:\n%*c", it.name(), minW-1, ' ');
+		cmd = it.args();
 		// split options into formatted lines
 		uint32 sz = cmd.size(), off = 0, n = maxW - minW;
 		while (n < sz) {
@@ -395,13 +407,32 @@ std::istream& ClaspAppBase::getStream() {
 }
 
 bool ClaspAppBase::updateConfig(const std::string& solvingOptions, bool resetOptions) {
-	if (resetOptions) { claspConfig_.reset(); }
-	ProgramOptions::OptionContext opts;
+	using namespace ProgramOptions;
+	claspConfig_.reset();
+	if (resetOptions) { argVec_.clear(); }
+	OptionContext opts;
 	claspConfig_.addOptions(opts);
-	ProgramOptions::ParsedOptions parsed;
-	parsed.assign(ProgramOptions::parseCommandString(solvingOptions, opts));
-	if (resetOptions) { opts.assignDefaults(parsed); }
-	return claspConfig_.finalize(parsed);
+	ParsedOptions exclude;
+	ParsedValues  values(ProgramOptions::parseCommandString(solvingOptions, opts, false, 0, ProgramOptions::command_line_allow_flag_value));
+	exclude.assign(values);
+	claspConfig_.addDisabled(exclude);
+	// reassign previous options
+	ArgVec::iterator j = argVec_.begin();
+	for (ArgVec::iterator it = j, end = argVec_.end(); it != end; ++it) {
+		OptionContext::option_iterator opt = opts.tryFind(it->first.c_str());
+		if (opt != opts.end() && (exclude.count(it->first) == 0 || (*opt)->value()->isComposing())) {
+			(*opt)->value()->parse(it->first, it->second, Value::value_unassigned);
+			exclude.add(it->first);
+			*j++ = *it;
+		}
+	}
+	argVec_.erase(j, argVec_.end());
+	// remember new options
+	for (ParsedValues::iterator it = values.begin(), end = values.end(); it != end; ++it) {
+		argVec_.push_back(ArgVal(it->first->name(), it->second));
+	}
+	opts.assignDefaults(exclude);
+	return claspConfig_.finalize(exclude, getProblemType(), true);
 }
 
 // Creates output object suitable for given input format
@@ -417,7 +448,7 @@ Output* ClaspAppBase::createOutput(ProblemType f) {
 		else if (f == Problem_t::ASP && claspAppOpts_.outf == ClaspAppOptions::out_comp) {
 			outFormat = TextOutput::format_aspcomp;
 		}
-		out.reset(new TextOutput(verbose(), outFormat, claspAppOpts_.ifs));
+		out.reset(new TextOutput(verbose(), outFormat, claspAppOpts_.outAtom.c_str(), claspAppOpts_.ifs));
 	}
 	else {
 		out.reset(new JsonOutput(verbose()));
@@ -431,8 +462,17 @@ Output* ClaspAppBase::createOutput(ProblemType f) {
 	if (claspAppOpts_.quiet[2] != static_cast<uint8>(UCHAR_MAX)) {
 		out->setCallQuiet((Output::PrintLevel)std::min(uint8(Output::print_no), claspAppOpts_.quiet[2]));
 	}
+	if (claspAppOpts_.hideAux) {
+		out->setHide('_');
+	}
 	return out.release();
 }
+void ClaspAppBase::storeCommandArgs(const ProgramOptions::ParsedValues& values) { 
+	argVec_.reserve(std::distance(values.begin(), values.end()));
+	for (ProgramOptions::ParsedValues::iterator it = values.begin(), end = values.end(); it != end; ++it) {
+		argVec_.push_back(ArgVal(it->first->name(), it->second));
+	}
+}
 
 bool ClaspAppBase::handlePostGroundOptions(ProgramBuilder& prg) {
 	if (!claspAppOpts_.onlyPre || prg.type() != Problem_t::ASP) { return true; }
@@ -476,6 +516,9 @@ void ClaspApp::printHelp(const ProgramOptions::OptionContext& root) {
 	printf("Get help/report bugs via : http://sourceforge.net/projects/potassco/support\n");
 	fflush(stdout);
 }
+void ClaspApp::storeCommandArgs(const ProgramOptions::ParsedValues&) {
+	/* We don't need the values */
+}
 /////////////////////////////////////////////////////////////////////////////////////////
 // WriteLemmas
 /////////////////////////////////////////////////////////////////////////////////////////
@@ -510,7 +553,7 @@ void WriteLemmas::flush(Constraint_t::Set x, uint32 maxLbd) {
 	Solver& s         = *ctx_->master();
 	const LitVec& t   = s.trail();
 	Antecedent trueAnte(posLit(0));
-	for (uint32 i = ctx_->topLevelSize(), end = s.decisionLevel() ? s.levelStart(1) : t.size(); i != end; ++i) {
+	for (uint32 i = ctx_->numUnary(), end = s.decisionLevel() ? s.levelStart(1) : t.size(); i != end; ++i) {
 		const Antecedent& a = s.reason(t[i]);
 		if (a.isNull() || a.asUint() == trueAnte.asUint()) {
 			os_ << t[i] << " 0\n";
