@@ -19,6 +19,12 @@
 
 // }}}
 
+#ifdef WITH_PYTHON
+#  include <Python.h>
+#endif
+#ifdef WITH_LUA
+#  include <lua.h>
+#endif
 #include "clingo_app.hh"
 #include <clasp/parser.h>
 #include <climits>
@@ -48,6 +54,27 @@ char const *ClingoStatistics::getKeys(char const* key) const {
 ClingoApp::ClingoApp() {
 }
 
+static bool parseConst(const std::string& str, std::vector<std::string>& out) {
+	out.push_back(str);
+	return true;
+}
+
+static bool parseWarning(const std::string& str, GringoOptions& out) {
+    if (str == "no-atom-undefined")        { out.wNoAtomUndef    = true;  return true; }
+    if (str ==    "atom-undefined")        { out.wNoAtomUndef    = false; return true; }
+    if (str == "no-define-cyclic")         { out.wNoCycle        = true;  return true; }
+    if (str ==    "define-cyclic")         { out.wNoCycle        = false; return true; }
+    if (str == "no-define-redfinition")    { out.wNoRedef        = true;  return true; }
+    if (str ==    "define-redfinition")    { out.wNoRedef        = false; return true; }
+    if (str == "no-file-included")         { out.wNoFileIncluded = true;  return true; }
+    if (str ==    "file-included")         { out.wNoFileIncluded = false; return true; }
+    if (str == "no-nonmonotone-aggregate") { out.wNoNonMonotone  = true;  return true; }
+    if (str ==    "nonmonotone-aggregate") { out.wNoNonMonotone  = false; return true; }
+    if (str == "no-term-undefined")        { out.wNoTermUndef    = true;  return true; }
+    if (str ==    "term-undefined")        { out.wNoTermUndef    = false; return true; }
+    return false;
+}
+
 void ClingoApp::initOptions(ProgramOptions::OptionContext& root) {
     using namespace ProgramOptions;
     BaseType::initOptions(root);
@@ -56,7 +83,7 @@ void ClingoApp::initOptions(ProgramOptions::OptionContext& root) {
     OptionGroup gringo("Gringo Options");
     gringo.addOptions()
         ("text"                     , flag(grOpts_.text = false)     , "Print plain text format")
-        ("const,c"                  , storeTo(grOpts_.defines)->composing()->arg("<id>=<term>"), "Replace term occurences of <id> with <term>")
+        ("const,c"                  , storeTo(grOpts_.defines, parseConst)->composing()->arg("<id>=<term>"), "Replace term occurences of <id> with <term>")
         ("lparse-rewrite"           , flag(grOpts_.lpRewrite = false), "Use together with --text to inspect lparse rewriting")
         ("lparse-debug"             , storeTo(grOpts_.lparseDebug = Gringo::Output::LparseDebug::NONE, values<Gringo::Output::LparseDebug>()
           ("none"  , Gringo::Output::LparseDebug::NONE)
@@ -67,12 +94,13 @@ void ClingoApp::initOptions(ProgramOptions::OptionContext& root) {
          "      plain : print rules as in plain output (prefix %%)\n"
          "      lparse: print rules as in lparse output (prefix %%%%)\n"
          "      all   : combines plain and lparse\n")
-        ("Wno-atom-undefined"       , flag(grOpts_.wNoAtomUndef = false) , "a :- b.")
-        ("Wno-define-cyclic"        , flag(grOpts_.wNoCycle = false)     , "#const a=b. #const b=a.")
-        ("Wno-define-redfinition"   , flag(grOpts_.wNoRedef = false)     , "#const a=1. #const a=2.")
-        ("Wno-file-included"        , flag(grOpts_.wNoFileIncluded=false), "#include \"a.lp\". #include \"a.lp\".")
-        ("Wno-nonmonotone-aggregate", flag(grOpts_.wNoNonMonotone=false) , "a :- #sum { 1:a; -1:a } >= 0.")
-        ("Wno-term-undefined"       , flag(grOpts_.wNoTermUndef=false)   , "p(1/0).")
+        ("warn,W"                   , storeTo(grOpts_, parseWarning)->arg("<warn>")->composing(), "Enable/disable warnings:\n"
+         "      [no-]atom-undefined:        a :- b.\n"
+         "      [no-]define-cyclic:         #const a=b. #const b=a.\n"
+         "      [no-]define-redfinition:    #const a=1. #const a=2.\n"
+         "      [no-]file-included:         #include \"a.lp\". #include \"a.lp\".\n"
+         "      [no-]nonmonotone-aggregate: a :- #sum { 1:a; -1:a } >= 0.\n"
+         "      [no-]term-undefined       : p(1/0).\n")
         ;
     root.add(gringo);
 
@@ -83,8 +111,8 @@ void ClingoApp::initOptions(ProgramOptions::OptionContext& root) {
     root.add(basic);
 }
 
-void ClingoApp::validateOptions(const ProgramOptions::OptionContext& root, const ProgramOptions::ParsedOptions& parsed) {
-    BaseType::validateOptions(root, parsed);
+void ClingoApp::validateOptions(const ProgramOptions::OptionContext& root, const ProgramOptions::ParsedOptions& parsed, const ProgramOptions::ParsedValues& vals) {
+    BaseType::validateOptions(root, parsed, vals);
     if (grOpts_.text) {
         if (mode_ == mode_clasp) { error("'--text' and '--mode=clasp' are mutually exclusive!"); exit(E_NO_RUN); }
         mode_ = mode_gringo;
@@ -113,6 +141,19 @@ void ClingoApp::printVersion() {
     ProgramOptions::Application::printVersion();
     printf("\n");
     printf("libgringo version " GRINGO_VERSION "\n");
+    printf("Configuration: "
+#ifdef WITH_PYTHON
+        "with Python " PY_VERSION
+#else
+        "without Python"
+#endif
+        ", "
+#ifdef WITH_LUA
+        "with " LUA_RELEASE
+#else
+        "without Lua"
+#endif
+        "\n");
     printf("Copyright (C) Roland Kaminski\n");
     printf("License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n");
     printf("Gringo is free software: you are free to change and redistribute it.\n");
@@ -134,6 +175,15 @@ struct ClingoModel : Gringo::Model {
     virtual Gringo::ValVec atoms(int atomset) const {
         return out.atoms(atomset, [this](unsigned uid) { return model.isTrue(lp.getLiteral(uid)); });
     }
+    virtual Gringo::Int64Vec optimization() const {
+        Gringo::Int64Vec values;
+        if (model.costs) {
+            for (unsigned level = 0, n = model.costs->numRules(); level < n; ++level) {
+                values.emplace_back(model.costs->optimum(level));
+            }
+        }
+        return values;
+    }
     virtual ~ClingoModel() { }
     Clasp::Asp::LogicProgram const   &lp;
     Gringo::Output::OutputBase const &out;
@@ -147,7 +197,7 @@ bool ClingoApp::onModel(Clasp::Solver const& s, Clasp::Model const& m) {
 }
 Gringo::Value ClingoApp::getConst(std::string const &name) {
     auto ret = grd->defs.defs().find(name);
-    return ret != grd->defs.defs().end() && std::get<2>(ret->second)->getInvertibility() == Gringo::Term::CONSTANT 
+    return ret != grd->defs.defs().end()
         ? std::get<2>(ret->second)->eval()
         : Gringo::Value();
 }
@@ -173,52 +223,16 @@ bool ClingoApp::prepare(ModelHandler h) {
     }
     Asp::LogicProgram& prg = static_cast<Asp::LogicProgram&>(clasp_->update(configUpdate != ConfigUpdate::KEEP));
     configUpdate = ConfigUpdate::KEEP;
-#ifdef DEBUG_OUTPUT
-    std::cerr << "clasp.update();\n";
-#endif
     if (!prg.ok()) { return false; }
     grd->ground(params, parts);
     if (!grd->lpOut || grd->lpOut->disposeMinimize()) { prg.disposeMinimizeConstraint(); }
-    for (auto &ext : freeze) {
-        Gringo::AtomState const* atm;
-        if (grd->out && (atm = grd->out->find(ext.first)) && atm->hasUid()) {
-#ifdef DEBUG_OUTPUT
-            std::cerr << "pb.freeze(" << atm->uid() << "," << (ext.second ? "Clasp::value_true" : "Clasp::value_false") << ");\n";
-#endif
-            prg.freeze(atm->uid(), ext.second ? Clasp::value_true : Clasp::value_false);
-        }
-    }
-    freeze.clear();
-    for (auto &ext : unfreeze) {
-        Gringo::AtomState const * atm;
-        if (grd->out && (atm = grd->out->find(ext)) && atm->hasUid()) {
-#ifdef DEBUG_OUTPUT
-            std::cerr << "pb.unfreeze(" << atm->uid() << ");\n";
-#endif
-            prg.unfreeze(atm->uid());
-        }
-    }
-    unfreeze.clear();
-    if (grOpts_.text) { exit(EXIT_SUCCESS); }
+    //if (grOpts_.text) { exit(EXIT_SUCCESS); }
     if (!handlePostGroundOptions(prg)) { return false; }
-#ifdef DEBUG_OUTPUT
-    std::cerr << "clasp.prepare(ClaspFacade::dispose_no);\n";
-#endif
     return clasp_->prepare(enableEnumAssupmption_ ? ClaspFacade::enum_volatile : ClaspFacade::enum_static) && handlePreSolveOptions(*clasp_);
 }
 Gringo::SolveFuture *ClingoApp::asolve(ModelHandler mh, FinishHandler fh) {
 #if WITH_THREADS
     prepare(mh);
-#ifdef DEBUG_OUTPUT
-    static bool defined = false;
-    if (!defined) {
-        std::cerr << "Clasp::ClaspFacade::AsyncResult f = clasp.solveAsync();\n";
-        defined = true;
-    }
-    else {
-        std::cerr << "f = clasp.solveAsync();\n";
-    }
-#endif
     solveFuture_.reset(fh);
     solveFuture_.reset(clasp_->solveAsync());
     return &solveFuture_;
@@ -235,14 +249,6 @@ void ClingoApp::onEvent(Event const& ev) {
 #if WITH_THREADS
     Clasp::ClaspFacade::StepReady const *r = Clasp::event_cast<Clasp::ClaspFacade::StepReady>(ev);
     if (r) {
-#ifdef DEBUG_OUTPUT
-        std::cerr <<
-            "switch (f.get()) {\n"
-            "    case ClaspFacade::Result::SAT:     { printf(\"SAT\\n\"); break; }\n"
-            "    case ClaspFacade::Result::UNSAT:   { printf(\"UNSAT\\n\"); break; }\n"
-            "    case ClaspFacade::Result::UNKNOWN: { printf(\"UNKNOWN\\n\"); break; }\n"
-            "}\n";
-#endif
         solveFuture_.ready(r->summary->result);
     }
 #endif
@@ -250,19 +256,16 @@ void ClingoApp::onEvent(Event const& ev) {
 }
 Gringo::SolveResult ClingoApp::solve(ModelHandler h) {
     if (!prepare(h)) { return convert(clasp_->result()); }
-#ifdef DEBUG_OUTPUT
-    std::cerr << "clasp.solve();\n";
-#endif
     return convert(clasp_->solve());
 }
 std::string ClingoApp::str() {
     return "[object:IncrementalControl]";
 }
 void ClingoApp::assignExternal(Gringo::Value ext, bool val) {
-    freeze.emplace_back(ext, val);
+    grd->freeze.emplace_back(ext, val ? Gringo::Output::ExternalType::E_TRUE : Gringo::Output::ExternalType::E_FALSE);
 }
 void ClingoApp::releaseExternal(Gringo::Value ext) {
-    unfreeze.emplace_back(ext);
+    grd->freeze.emplace_back(ext, Gringo::Output::ExternalType::E_FREE);
 }
 
 ClingoStatistics *ClingoApp::getStats() {
@@ -277,16 +280,6 @@ void ClingoApp::run(Clasp::ClaspFacade& clasp) {
     bool incremental    = mode_ != mode_clasp;
     ProblemType     pt  = getProblemType();
     ProgramBuilder* prg = &clasp.start(claspConfig_, pt, incremental);
-#ifdef DEBUG_OUTPUT
-    std::cerr 
-        << "#include <clasp/clasp_facade.h>\n"
-        << "#include <cstdio>\n"
-        << "using namespace Clasp;\n"
-        << "int main() {\n"
-        << "ClaspFacade clasp;\n"
-        << "ClaspConfig cfg;\n"
-        << "Asp::LogicProgram& pb = static_cast<Asp::LogicProgram&>(clasp.start(cfg, Problem_t::ASP, true));\n";
-#endif
     if (incremental) {
         grOpts_.verbose = verbose() == UINT_MAX;
         Asp::LogicProgram* lp = mode_ != mode_gringo ? static_cast<Asp::LogicProgram*>(prg) : 0;
