@@ -168,6 +168,15 @@ public:
 	 * The default implementation returns 0.
 	 */
 	virtual ClauseHead* clause();
+
+	//! Shall return whether the constraint is valid (i.e. not conflicting) w.r.t the current assignment in s.
+	/*!
+	 * \pre The assignment in s is not conflicting and fully propagated.
+	 * \post A changed assignment if the assignment was not valid.
+	 * \note The default implementation always returns true and assumes
+	 *       that conflicts are detected by Constraint::propagate().
+	 */
+	virtual bool valid(Solver& s);
 protected:
 	virtual ~Constraint();
 private:
@@ -274,7 +283,7 @@ public:
 	 */
 	virtual uint32 priority() const = 0;
 
-	//! Called when this object is added as a post propagator to s.
+	//! Called during initialization of s.
 	/*!
 	 * \note During initialization a post propagator may assign variables
 	 *       but it must not yet propagate them.
@@ -297,9 +306,18 @@ public:
 	 * \note 
 	 *   The function shall not call Solver::propagate()
 	 *   or any other function that would result in a recursive chain 
-	 *   of propagate() calls. On the other hand, it must call
+	 *   of propagate() calls. On the other hand, it shall call
 	 *   Solver::propagateUntil(this) after enqueuing new assignments 
 	 *   to initiate propagation up to this propagator.
+	 * 
+	 * Typically, propagateFixpoint() should implemet a loop like this:
+	 * \code
+	 * for (;;) {
+	 *   if (!assign_newly_implied_literals(s)){ return false; }
+	 *   if (s.queueSize() == 0)               { return true;  }
+	 *   if (!s.propagateUntil(this))          { return false; }
+	 * }
+	 * \endcode
 	 */
 	virtual bool   propagateFixpoint(Solver& s, PostPropagator* ctx) = 0;
 
@@ -322,6 +340,7 @@ public:
 	 *  - false otherwise
 	 * \post If the function returned true:  s.numFreeVars() == 0 && !s.hasConflict().
 	 *       If the function returned false: s.numFreeVars() > 0 || s.hasConflict().
+	 * \note The default implementation returns Constraint::valid(s);
 	 */
 	virtual bool   isModel(Solver& s);
 protected:
@@ -335,25 +354,6 @@ private:
 	PostPropagator& operator=(const PostPropagator&);
 };
 
-//! Base class for legacy post propagators.
-template <class T>
-class LegacyPostPropagator : public PostPropagator {
-public:
-	//! Propagates the current assignment until a fixpoint is reached w.r.t this post propagator.
-	/*!
-	 * This function implements a basic propagation loop that calls
-	 * T::propagate(Solver& s) until a fixpoint is reached or
-	 * a conflict is detected.
-	 */
-	bool propagateFixpoint(Solver& s, PostPropagator*);
-	T*   derived()           { return static_cast<T*>(this); }
-protected:
-	template <typename U, bool (U::*)(Solver&, PostPropagator*)> struct enable { typedef bool type; };
-	template <typename U>
-	typename enable<U, &U::propagate>::type _propagate(Solver& s, PostPropagator* ctx, U* x){ return x->propagate(s, ctx); }
-	bool                                    _propagate(Solver& s, PostPropagator*, ...)     { return derived()->propagate(s); }
-};
-
 class MessageHandler : public PostPropagator {
 public:
 	MessageHandler();
