@@ -57,35 +57,11 @@ struct HeuParams {
  * \see E. Goldberg, Y. Navikov: "BerkMin: a Fast and Robust Sat-Solver"
  *
  * \note
- * This version of the BerkMin heuristic varies mostly in three points from the original BerkMin:
+ * This version of the BerkMin heuristic varies mostly in the following points from the original BerkMin:
  *  -# it considers loop-nogoods if requested (this is the default)
  *  -# it uses a MOMS-like heuristic as long as there are no conflicts (and therefore no activities)
- *  -# it uses a MOMS-like score to break ties whenever multiple variables from an unsatisfied learnt
- * constraint have equal activities
- * 
- * The algorithm is as follows:
- * - For each variable a score s is maintained that is initially 0.
- * - During conflict analysis: For each variable v contained in a constraint CL that
- * participates in the derivation of the asserting clause, s(v) is incremented.
- * - Every 512 decisions, the score of each variable is divided by 4
- *
- * For selecting a literl DL the following algorithm is used:
- * - let CL be the most recently derived conflict clause that is currently not satisfied.
- * - if no such CL exists and look-back loops is enabled let CL be last learnt loop nogood that is currently not satisfied.
- * - if a CL was found:
- *    - let v be a free variable of CL with the highest activity. Break ties using MOMS
- *    - let p be the literal of v that occured more often in learnt nogoods. 
- *      - If both lits of v occurred equally often, set p to the preferred literal of v
- *    - set DL to p
- * - if no CL was found and #conflicts != 0
- *    - let v be the free variable with the highest activity. Break ties randomly.
- *    - let p and ~p be the two literals of v.
- *    - let s1 be the result of estimateBCP(p), let s2 be the result of estimateBCP(~p)
- *    - set DL to p iff s1 > s2, to ~p iff s2 > s1 and to the preferred decision literal of v otherwise.
- * - if #conflicts == 0
- *    - Select DL according to a MOMS-like heuristic
- * - return DL;
- * .
+ *  -# it uses a MOMS-like score to break ties whenever multiple variables from an unsatisfied learnt constraint have equal activities
+ *  -# it uses a lazy decaying scheme that only touches active variables
  */
 class ClaspBerkmin : public DecisionHeuristic {
 public:
@@ -100,14 +76,7 @@ public:
 	void updateReason(const Solver& s, const LitVec& lits, Literal resolveLit);
 	bool bump(const Solver& s, const WeightLitVec& lits, double adj);
 	void undoUntil(const Solver&, LitVec::size_type);
-	void resurrect(const Solver&, Var v) {
-		if (order_.score.size() <= v) {
-			order_.score.resize(v+1);
-		}
-		front_ = 1;
-		cache_.clear();
-		cacheFront_ = cache_.end();
-	}
+	void updateVar(const Solver& s, Var v, uint32 n);
 protected:
 	Literal doSelect(Solver& s);
 private:
@@ -198,7 +167,7 @@ private:
  * \see Lawrence Ryan: "Efficient Algorithms for Clause Learning SAT-Solvers"
  *
  * \note This implementation of VMTF differs from the original implementation in three points:
- *  - if considerLoops is true, it moves to the front a selection of variables from learnt loop nogoods
+ *  - it optionally moves to the front a selection of variables from learnt loop nogoods
  *  - it measures variable activity by using a BerkMin-like score scheme
  *  - the initial order of the var list is determined using a MOMs-like score scheme
  */
@@ -215,13 +184,7 @@ public:
 	bool bump(const Solver& s, const WeightLitVec& lits, double adj);
 	void simplify(const Solver&, LitVec::size_type);
 	void undoUntil(const Solver&, LitVec::size_type);
-	void resurrect(const Solver&, Var v) {
-		if (score_.size() <= v) {
-			score_.resize(v+1, VarInfo(vars_.end()));
-		}
-		if (score_[v].pos_ == vars_.end()) { score_[v].pos_ = vars_.insert(vars_.end(), v); }
-		else { front_ = vars_.begin(); }
-	}
+	void updateVar(const Solver& s, Var v, uint32 n);
 protected:
 	Literal doSelect(Solver& s);
 private:
@@ -267,6 +230,20 @@ private:
 	const LitVec::size_type MOVE_TO_FRONT;
 };
 
+//! Score type for VSIDS heuristic.
+/*!
+ * \see ClaspVsids
+ */
+struct VsidsScore {
+	typedef VsidsScore SC;
+	VsidsScore() : value(0.0) {}
+	double get()                  const { return value; }
+	bool   operator>(const SC& o) const { return value > o.value; }
+	double inc(double f)                { return (value += f); }
+	void   set(double f)                { value = f; }
+	double value;
+};
+
 //! A variable state independent decision heuristic favoring variables that were active in recent conflicts.
 /*!
  * \ingroup heuristic
@@ -276,26 +253,16 @@ private:
  * \see M. W. Moskewicz, C. F. Madigan, Y. Zhao, L. Zhang, and S. Malik: 
  * "Chaff: Engineering an Efficient SAT Solver"
  * \see E. Goldberg, Y. Navikov: "BerkMin: a Fast and Robust Sat-Solver"
- * 
- * The heuristic works as follows:
- * - For each variable v an activity a(v) is maintained that is initially 0.
- * - For each literal p an occurrence counter c(p) is maintained that is initially 0.
- * - When a new constraint is learnt (conflict or loop), c(p) is incremented 
- * for each literal p in the newly learnt constraint.
- * - During conflict analysis: For each variable v contained in a constraint CL that
- * participates in the derivation of the asserting clause, a(v) is incremented.
- * - Periodically the activity of each variable is divided by a constant that is a power
- * of two (currently every 512 decision variable activity is dived by 4).
- * - When a decision is necessary a free variable with the highest activity is taken
- * and from this variable the literal that occured in more learnt clauses is picked.
- * If c(v) == c(~v) the preferred choice literal of v is selected (v for bodies, ~v for heads).
+ *
+ * \note The implementation uses the exponential VSIDS scheme from MiniSAT.
  */
-class ClaspVsids : public DecisionHeuristic {
+template <class ScoreType>
+class ClaspVsids_t : public DecisionHeuristic {
 public:
 	/*!
 	 * \param d Set initial inc-factor to 1.0/d.
 	 */
-	explicit ClaspVsids(double d = 0.95, const HeuParams& params = HeuParams());
+	explicit ClaspVsids_t(double d = 0.95, const HeuParams& params = HeuParams());
 	void startInit(const Solver& s);
 	void endInit(Solver&);
 	void newConstraint(const Solver& s, const Literal* first, LitVec::size_type size, ConstraintType t);
@@ -303,48 +270,149 @@ public:
 	bool bump(const Solver& s, const WeightLitVec& lits, double adj);
 	void undoUntil(const Solver&, LitVec::size_type);
 	void simplify(const Solver&, LitVec::size_type);
-	void resurrect(const Solver&, Var v) {
-		if (v >= score_.size()) {
-			score_.resize(v+1);
-		}
-		vars_.update(v);  
-	}
+	void updateVar(const Solver& s, Var v, uint32 n);
 protected:
 	Literal doSelect(Solver& s);
-private:
+	virtual void initScores(Solver& s, bool moms);
+	typedef typename PodVector<ScoreType>::type ScoreVec;
+	typedef PodVector<int32>::type              OccVec;
 	Literal selectRange(Solver& s, const Literal* first, const Literal* last);
 	void updateVarActivity(Var v, double f = 1.0) {
-		if ( (score_[v].first += (inc_*f)) > 1e100 ) {
-			normalize();
-		}
+		double old = score_[v].get(), n = score_[v].inc((inc_*f));
+		if (n > 1e100) { normalize(); }
 		if (vars_.is_in_queue(v)) {
-			vars_.increase(v);
+			if (n >= old) { vars_.increase(v); }
+			else          { vars_.decrease(v); }
 		}
 	}
+	void incOcc(Literal p) { occ_[p.var()] += 1 - (int(p.sign()) << 1); }
+	int  occ(Var v) const  { return occ_[v]; }
 	void normalize();
-	// HScore.first: activity of a variable
-	// HScore.second: occurrence counter for a variable's literal
-	//  > 0: positive literal occurred more often
-	//  < 0: negative literal occurred more often
-	//  = 0: literals occurred equally often
-	typedef std::pair<double, int32>  HScore;
-	typedef PodVector<HScore>::type   Scores;
-	struct GreaterActivity {
-		explicit GreaterActivity(const Scores& s) : sc_(s) {}
-		bool operator()(Var v1, Var v2) const {
-			return sc_[v1].first > sc_[v2].first;
-		}
+	struct CmpScore {
+		explicit CmpScore(const ScoreVec& s) : sc_(s) {}
+		bool operator()(Var v1, Var v2) const { return sc_[v1] > sc_[v2]; }
 	private:
-		GreaterActivity& operator=(const GreaterActivity&);
-		const Scores& sc_;
+		CmpScore& operator=(const CmpScore&);
+		const ScoreVec& sc_;
 	};
-	typedef bk_lib::indexed_priority_queue<GreaterActivity> VarOrder;
-	Scores       score_;
+	typedef bk_lib::indexed_priority_queue<CmpScore> VarOrder;
+	ScoreVec     score_;
+	OccVec       occ_;
 	VarOrder     vars_;
 	const double decay_;
 	double       inc_;
 	TypeSet      types_;
 };
+typedef ClaspVsids_t<VsidsScore> ClaspVsids;
 
+//! Score type for DomainHeuristic.
+/*!
+ * \see DomainHeuristic
+ */
+struct DomScore {
+	typedef DomScore SC;
+	DomScore() : value(0.0), level(0), factor(1), domKey(UINT32_MAX) {}
+	double get()                  const { return value; }
+	bool   operator>(const SC& o) const { return (level > o.level) || (level == o.level && value > o.value); }
+	bool   isDom()                const { return domKey != UINT32_MAX; }
+	void   setDom(uint32 key)           { domKey = key; }
+	double inc(double f)                { return (value += (f*factor)); }
+	void   set(double f)                { value = f; }
+	double value;  // activity as in VSIDS
+	int16  level;  // priority level
+	int16  factor; // factor used when bumping activity
+	uint32 domKey; // index into dom-table if dom var	
+};
+
+//! A VSIDS heuristic supporting additional domain knowledge.
+/*!
+ * \ingroup heuristic
+ *
+ * \see M. Gebser, B. Kaufmann, R. Otero, J. Romero, T. Schaub, P. Wanko:
+ *      "Domain-specific Heuristics in Answer Set Programming",
+ *      http://www.cs.uni-potsdam.de/wv/pdfformat/gekaotroscwa13a.pdf
+ */
+class DomainHeuristic : public ClaspVsids_t<DomScore>
+                      , private Constraint {
+public:
+	typedef ClaspVsids_t<DomScore> BaseType;
+	enum GlobalPreference { gpref_atom = 1, gpref_scc = 2, gpref_hcc = 4, gpref_disj = 8, gpref_min  = 16, gpref_show = 32 };
+	enum GlobalModifier   { gmod_level = 1, gmod_spos = 2, gmod_true = 3, gmod_sneg  = 4, gmod_false = 5, gmod_max_value };   
+	explicit DomainHeuristic(double d = 0.95, const HeuParams& params = HeuParams());
+	~DomainHeuristic();
+	virtual void startInit(const Solver& s);
+	using BaseType::endInit;
+protected:
+	// base interface
+	virtual Literal     doSelect(Solver& s);
+	virtual void        initScores(Solver& s, bool moms);
+	// Constraint interface
+	virtual Constraint* cloneAttach(Solver&) { return 0; }
+	virtual void        reason(Solver&, Literal, LitVec&){}
+	virtual PropResult  propagate(Solver&, Literal, uint32&);
+	virtual void        undoLevel(Solver& s);
+	// own interface
+	void detach();
+private:
+	// parsing & init
+	typedef SymbolTable::symbol_type    SymbolType;
+	typedef PodVector<SymbolType>::type SymbolMap;
+	struct CmpSymbol;
+	struct DomEntry {
+		enum Modifier { mod_factor = 0, mod_level = 1, mod_sign = 2, mod_tf = 3, mod_init = 4 };
+		DomEntry(const SymbolType* s, Modifier m = mod_init) : sym(s), mod(m), val(0), prio(0), sign(0) {}
+		bool     parse(const char*& mod);
+		ValueRep signValue() const {
+			if (mod == mod_sign) { return ValueRep(val); }
+			if (mod == mod_tf)   { return ValueRep(value_true + sign); }
+			return value_free;
+		}
+		const SymbolType* sym;    // domain symbol
+		Modifier          mod;    // type of modification
+		int16             val;    // value of modification
+		uint16            prio:15;// (optional) priority of modification
+		uint16            sign: 1;// sign in case of mod_tf
+	};
+	struct DomAction {
+		uint32 var:29; // dom var to be modified
+		uint32 mod: 2; // modification to apply
+		uint32 comp:1; // compound action?
+		uint32 undo;   // next action in undo list
+		int16  val;    // value to apply 
+		uint16 prio;   // prio of modification
+	};
+	struct DomPrio {
+		void    clear() { prio[0] = prio[1] = prio[2] = 0; }
+		uint16  operator[](unsigned i) const { return prio[i]; }
+		uint16& operator[](unsigned i)       { return prio[i]; }
+		uint16  prio[3];
+	};
+	struct Frame {
+		Frame(uint32 lev, uint32 h) : dl(lev), head(h) {}
+		uint32 dl;
+		uint32 head;
+	};
+	typedef PodVector<DomAction>::type ActionVec;
+	typedef PodVector<DomPrio>::type   PrioVec;
+	typedef PodVector<Frame>::type     FrameVec;
+	static const char* const domKey_s;
+	static const char* const domEnd_s;
+	static bool match(const char*& in, const char* match);
+	static bool matchDom(const char*& in, std::string& out);
+	static bool matchInt(const char*& in, int& out);
+	void        addAction(Solver& s, const SymbolType& pre, const DomEntry& e);
+	void        addAction(Solver& s, Literal x, uint32 modf, int16 levVal);
+	void        endInit(const DomEntry& e, const DomPrio& prio) { 
+		Var v = e.sym->lit.var();
+		if (e.val)                            { score_[v].value += e.val; }
+		if (score_[v].domKey < dPrios_.size()){ dPrios_[score_[v].domKey] = prio; }
+	}
+	void        pushUndo(Solver& s, uint32 actionId);
+	void        applyAction(Solver& s, DomAction& act, uint16& oldPrio);
+	Solver*   solver_;  // solver to which we are attached as a constraint
+	ActionVec actions_; // dynamic modifications
+	PrioVec   dPrios_;  // dynamic priorities
+	FrameVec  frames_;  // dynamic undo information
+};
 }
 #endif
