@@ -290,6 +290,7 @@ public:
 	void  clear(Var v, uint8 f)           { if (v < state_.size()) { state_[v] &= ~f; } }
 	void  clear(Var v)                    { clear(v, static_cast<uint8>(-1)); }
 	void  clearHead(PrgEdge t)            { clear(t.node(), headFlag(t)); }
+	void  clearBody(Literal p)            { clear(p.var(), pos_flag+p.sign()); }
 	void  clearAll()                      { StateVec().swap(state_); }
 
 	bool  allMarked(const VarVec& vec, uint8 f) const {
@@ -552,7 +553,7 @@ public:
 	 *    make the body false (i.e. the body is a selfblocker). Otherwise, true.
 	 */
 	bool     simplifyHeads(LogicProgram& prg, bool strong);
-	bool     mergeHeads(LogicProgram& prg, PrgBody& heads, bool strong);
+	bool     mergeHeads(LogicProgram& prg, PrgBody& heads, bool strong, bool simplify = true);
 	void     removeHead(PrgHead* h, EdgeType t);
 	//! Simplifies the body, i.e. its predecessors-lists.
 	/*!
