@@ -36,7 +36,7 @@
  */
 namespace Clasp {
 class Solver;
-class ClauseInfo;	
+class ClauseInfo;
 class Assignment;
 class SharedContext;
 class SharedLiterals;
@@ -81,7 +81,7 @@ public:
 		Literal lits_[1];     // literals of the clause: [lits_[0], lits_[size_])
 	};
 	
-	SatPreprocessor() : ctx_(0), elimTop_(0) {}
+	SatPreprocessor() : ctx_(0), elimTop_(0), seen_(1,1) {}
 	virtual ~SatPreprocessor();
 
 	//! Creates a clone of this preprocessor.
@@ -126,6 +126,7 @@ protected:
 	virtual void  doCleanUp() = 0;
 	Clause*       clause(uint32 clId)       { return clauses_[clId]; }
 	const Clause* clause(uint32 clId) const { return clauses_[clId]; }
+	void          freezeSeen();
 	void          discardClauses(bool discardEliminated);
 	void          setClause(uint32 clId, const LitVec& cl) {
 		clauses_[clId] = Clause::newClause(&cl[0], (uint32)cl.size());
@@ -145,8 +146,9 @@ protected:
 private:
 	SatPreprocessor(const SatPreprocessor&);
 	SatPreprocessor& operator=(const SatPreprocessor&);
-	ClauseList      clauses_; // initial non-unit clauses
-	LitVec          units_;   // initial unit clauses
+	ClauseList clauses_; // initial non-unit clauses
+	LitVec     units_;   // initial unit clauses
+	Range32    seen_;    // vars seen in previous step
 };
 //@}
 
@@ -198,11 +200,27 @@ struct VarInfo {
 		FROZEN = 0x80u // is the variable frozen?
 	};
 	VarInfo() : rep(0) {}
-	bool  has(FLAG f)  const { return (rep & flag(f)) != 0; }
-	bool  has(uint32 f)const { return (rep & f) != 0;       }
-	void  set(FLAG f)        { rep |= flag(f); }
-	void  toggle(FLAG f)     { rep ^= flag(f); }
-	static uint8 flag(FLAG x){ return uint8(x); }
+
+	//! Returns the type of the variable (or Var_t::atom_body_var if variable was created with parameter eq=true).
+	VarType type()          const { return has(VarInfo::EQ) ? Var_t::atom_body_var : VarType(Var_t::atom_var + has(VarInfo::BODY)); }
+	//! Returns true if var is contained in a negative loop or head of a choice rule.
+	bool    nant()          const { return has(VarInfo::NANT); }
+	//! Returns true if var is a projection variable.
+	bool    project()       const { return has(VarInfo::PROJECT);}
+	bool    inDisj()        const { return has(VarInfo::DISJ);}
+	//! Returns true if var is excluded from variable elimination.
+	bool    frozen()        const { return has(VarInfo::FROZEN); }
+	//! Returns the preferred sign of this variable w.r.t its type.
+	/*!
+	 * \return false (i.e no sign) if var originated from body, otherwise true.
+	 */
+	bool    preferredSign() const { return !has(VarInfo::BODY); }
+	
+	bool    has(FLAG f)     const { return (rep & flag(f)) != 0; }
+	bool    has(uint32 f)   const { return (rep & f) != 0;       }
+	void    set(FLAG f)           { rep |= flag(f); }
+	void    toggle(FLAG f)        { rep ^= flag(f); }
+	static uint8 flag(FLAG x)     { return uint8(x); }
 	uint8 rep;
 };
 
@@ -211,6 +229,8 @@ class ShortImplicationsGraph {
 public:
 	ShortImplicationsGraph();
 	~ShortImplicationsGraph();
+	enum ImpType { binary_imp = 2, ternary_imp = 3 };
+	
 	//! Makes room for nodes number of nodes.
 	void resize(uint32 nodes);
 	//! Mark the instance as shared/unshared.
@@ -219,16 +239,11 @@ public:
 	 * to specialized shared blocks of memory.
 	 */
 	void markShared(bool b) { shared_ = b; }
-	//! Adds the binary constraint (p, q) to the implication graph.
+	//! Adds the given constraint to the implication graph.
 	/*!
 	 * \return true iff a new implication was added.
 	 */
-	bool addBinary(Literal p, Literal q, bool learnt);
-	//! Adds the ternary constraint (p, q, r) to the implication graph.
-	/*!
-	 * \return true iff a new implication was added.
-	 */
-	bool addTernary(Literal p, Literal q, Literal r, bool learnt);
+	bool add(ImpType t, bool learnt, const Literal* lits);
 	
 	//! Removes p and its implications.
 	/*!
@@ -250,6 +265,7 @@ public:
 	uint32 numTernary()const { return tern_[0]; }
 	uint32 numLearnt() const { return bin_[1] + tern_[1]; }
 	uint32 numEdges(Literal p) const;
+	uint32 size() const      { return graph_.size(); }
 	//! Applies op on all unary- and binary implications following from p.
 	/*!
 	 * OP must provide two functions:
@@ -384,9 +400,11 @@ public:
 	typedef SingleOwnerPtr<SDG>            SccGraph;
 	typedef Configuration*                 ConfigPtr;
 	typedef std::auto_ptr<Distributor>     DistrPtr;
-	typedef ProblemStats                   Stats;
+	typedef const ProblemStats&            StatsRef;
+	typedef SymbolTable&                   SymbolsRef;
 	typedef LitVec::size_type              size_type;
-	typedef ShortImplicationsGraph         BTIG;
+	typedef ShortImplicationsGraph         ImpGraph;
+	typedef const ImpGraph&                ImpGraphRef;
 	typedef EventHandler*                  LogPtr;
 	typedef std::auto_ptr<SatPreprocessor> SatPrePtr;
 	enum InitMode   { init_share_symbols, init_copy_symbols };
@@ -397,21 +415,18 @@ public:
 	//! Creates a new object for sharing variables and the binary and ternary implication graph.
 	explicit SharedContext();
 	~SharedContext();
+	//! Resets this object to the state after default construction.
+	void       reset();
 	//! Enables event reporting via the given event handler.
-	void       eventHandler(EventHandler* r) { progress_ = r; }
-	LogPtr     eventHandler() const          { return progress_; }
+	void       setEventHandler(EventHandler* r) { progress_ = r; }
 	//! Sets how to handle physical sharing of constraints.
 	void       setShareMode(ContextParams::ShareMode m);
 	//! Sets whether the short implication graph should be used for storing short learnt constraints.
 	void       setShortMode(ContextParams::ShortMode m);
-	//! Resets this object to the state after default construction.
-	void       reset();
 	//! Sets maximal number of solvers sharing this object.
-	void       concurrency(uint32 numSolver);
+	void       setConcurrency(uint32 numSolver);
 	//! Adds an additional solver to this object and returns it.
 	Solver&    addSolver();
-	bool       hasSolver(uint32 id) const { return id < solvers_.size(); }
-	bool       seedSolvers()        const { return share_.seed != 0; }
 	//! Configures the statistic object of attached solvers.
 	/*!
 	 * The level determines the amount of extra statistics.
@@ -423,40 +438,93 @@ public:
    */
 	void       enableStats(uint32 level);
 	void       accuStats();
+	//! If b is true, sets preprocessing mode to model-preserving operations only.
+	void       setPreserveModels(bool b = true) { share_.satPreM = b ? SatPreParams::prepro_preserve_models : SatPreParams::prepro_preserve_sat; }
 	//! Sets the configuration for this object and its attached solvers.
 	/*!
 	 * \note If own is true, ownership of c is transferred.
 	 */
 	void       setConfiguration(Configuration* c, bool own);
-
-	SatPrePtr  satPrepro;  /*!< Preprocessor for simplifying problem.                  */
-	DistrPtr   distributor;/*!< Distributor object to use for distribution of learnt constraints. */
-	SccGraph   sccGraph;   /*!< Program dependency graph - only used for ASP-problems. */
-	//! Returns true unless the master has an unresolvable top-level conflict.
-	bool       ok()                            const;
-	//! Returns the master solver associated with this object.
-	Solver*    master()                        const { return solver(0);    }	
-	//! Returns the solver with the given id.
-	Solver*    solver(uint32 id)               const { return solvers_[id]; }
-	uint32     concurrency()                   const { return share_.count; }
-	bool       frozen()                        const { return share_.frozen;}
-	bool       isShared()                      const { return frozen() && concurrency() > 1; }
-	bool       physicalShare(ConstraintType t) const { return (share_.shareM & (1 + (t != Constraint_t::static_constraint))) != 0; }
-	bool       physicalShareProblem()          const { return (share_.shareM & ContextParams::share_problem) != 0; }
-	bool       allowImplicit(ConstraintType t) const { return t != Constraint_t::static_constraint ? share_.shortM != ContextParams::short_explicit : !isShared(); }
-	bool       preserveModels()                const { return static_cast<SatPreParams::Mode>(share_.satPreM) == SatPreParams::prepro_preserve_models; }
-	bool       isExtended()                    const { return problem_.vars_frozen || symTab().type() == SymbolTable::map_indirect; }
-	//! Returns the current configuration used in this object.
-	/*!
-	 * If no configuration was set, a default configuration is used.
-	 * \see DefaultConfiguration
-	 */
-	ConfigPtr  configuration()                 const { return config_.get(); }
-	//@}
+	SatPrePtr  satPrepro;  /*!< Preprocessor for simplifying problem.                            */
+	SccGraph   sccGraph;   /*!< Program dependency graph - only used for ASP-problems.           */
 	
+	//! Returns the current configuration used in this object.
+	ConfigPtr  configuration()      const { return config_.get(); }
+	//! Returns the active event handler or 0 if none was set.
+	LogPtr     eventHandler()       const { return progress_; }
+	//! Returns whether this object seeds the RNG of new solvers.
+	bool       seedSolvers()        const { return share_.seed != 0; }
+	//! Returns the number of solvers that can share this object.
+	uint32     concurrency()        const { return share_.count; }
+	bool       preserveModels()     const { return static_cast<SatPreParams::Mode>(share_.satPreM) == SatPreParams::prepro_preserve_models; }
+	//! Returns whether physical sharing is enabled for constraints of type t.
+	bool       physicalShare(ConstraintType t) const { return (share_.shareM & (1 + (t != Constraint_t::static_constraint))) != 0; }
+	//! Returns whether pyhiscal sharing of problem constraints is enabled.
+	bool       physicalShareProblem()          const { return (share_.shareM & ContextParams::share_problem) != 0; }
+	//! Returns whether short constraints of type t can be stored in the short implication graph.
+	bool       allowImplicit(ConstraintType t) const { return t != Constraint_t::static_constraint ? share_.shortM != ContextParams::short_explicit : !isShared(); }
+	//@}
+
 	/*!
-	 * \name Problem specification
-	 * Functions for adding a problem to the master solver.
+	 * \name Problem introspection
+	 * Functions for querying information about the problem.
+	 */
+	//@{
+	//! Returns true unless the master has an unresolvable top-level conflict.
+	bool       ok()                 const;
+	//! Returns whether the problem is currently frozen and therefore ready for being solved.
+	bool       frozen()             const { return share_.frozen;}
+	//! Returns whether more than one solver is actively working on the problem.
+	bool       isShared()           const { return frozen() && concurrency() > 1; } 
+	bool       isExtended()         const { return problem_.vars_frozen || symbolTable().type() == SymbolTable::map_indirect; }
+	//! Returns whether this object has a solver associcated with the given id.
+	bool       hasSolver(uint32 id) const { return id < solvers_.size(); }
+	//! Returns the master solver associated with this object.
+	Solver*    master()             const { return solver(0);    }	
+	//! Returns the solver with the given id.
+	Solver*    solver(uint32 id)    const { return solvers_[id]; }
+	
+	//! Returns the number of problem variables.
+	/*!
+	 * \note The special sentinel-var 0 is not counted, i.e. numVars() returns
+	 * the number of problem-variables.
+	 * To iterate over all problem variables use a loop like:
+	 * \code
+	 * for (Var i = 1; i <= numVars(); ++i) {...}
+	 * \endcode
+	 */
+	uint32     numVars()            const { return static_cast<uint32>(varInfo_.size() - 1); }
+	//! Returns the number of eliminated vars.
+	uint32     numEliminatedVars()  const { return problem_.vars_eliminated; }
+	//! Returns true if var represents a valid variable in this problem.
+	/*!
+	 * \note The range of valid variables is [1..numVars()]. The variable 0
+	 * is a special sentinel variable. 
+	 */
+	bool       validVar(Var var)    const { return var < static_cast<uint32>(varInfo_.size()); }
+	//! Returns information about the given variable.
+	VarInfo    varInfo(Var v)       const { assert(validVar(v)); return varInfo_[v]; }
+	//! Returns true if v is currently eliminated, i.e. no longer part of the problem.
+	bool       eliminated(Var v)    const;
+	bool       marked(Literal p)    const { return varInfo(p.var()).has(VarInfo::MARK_P + p.sign()); }
+	Literal    stepLiteral()        const { return step_; }
+	//! Returns the number of problem constraints.
+	uint32     numConstraints()     const;
+	//! Returns the number of binary constraints.
+	uint32     numBinary()          const { return btig_.numBinary();  }
+	//! Returns the number of ternary constraints.
+	uint32     numTernary()         const { return btig_.numTernary(); }
+	//! Returns the number of unary constraints.
+	uint32     numUnary()           const { return lastTopLevel_; }
+	//! Returns an estimate of the problem complexity based on the number and type of constraints.
+	uint32     problemComplexity()  const;
+	StatsRef   stats()              const { return problem_; }
+	SymbolsRef symbolTable()        const { return symTabPtr_->symTab; }
+	//@}
+
+	/*!
+	 * \name Problem setup
+	 * Functions for specifying the problem.
 	 * Problem specification is a four-stage process:
 	 * -# Add variables to the SharedContext object.
 	 * -# Call startAddConstraints().
@@ -472,19 +540,18 @@ public:
 	 * \note !frozen() is a precondition for all functions in this group!
 	 */
 	//@{
-	bool    unfreeze();
-	//! Reserves space for at least varGuess variables.
+	//! Unfreezes a frozen program and prepares it for updates.
 	/*!
-	 * \param varGuess Number of vars to reserve space for.
-	 * \note If the number of variables is known upfront, passing the correct value
-	 * for varGuess avoids repeated regrowing of the state data structures.
+	 * The function also triggers forgetting of volatile knowledge and removes
+	 * any auxiliary variables.
+	 * \see requestStepVar()
+	 * \see Solver::popAuxVar()
 	 */
-	void    reserveVars(uint32 varGuess);
+	bool    unfreeze();
 	//! Clones vars and symbol table from other.
 	void    cloneVars(const SharedContext& other, InitMode m = init_copy_symbols);
-	//! Removes all vars [startVar, numVars()].
-	void    popVars(uint32 startVar) { varInfo_.resize(std::max(Var(1), startVar)); problem_.vars = numVars(); }
-
+	//! Sets the range of problem variables to [1, maxVar)
+	void    resizeVars(uint32 maxVar) { varInfo_.resize(std::max(Var(1), maxVar)); problem_.vars = numVars(); }
 	//! Adds a new variable of type t.
 	/*!
 	 * \param t  Type of the new variable (either Var_t::atom_var or Var_t::body_var).
@@ -494,113 +561,44 @@ public:
 	 * \note Problem variables are numbered from 1 onwards!
 	 */
 	Var     addVar(VarType t, bool eq = false);
+	//! Requests a special variable for tagging volatile knowledge in incremental setting.
+	void    requestStepVar();
 	//! Request additional reason data slot for variable v.
 	void    requestData(Var v);
-	//! Freezes/defreezes a variable (a frozen var is exempt from SatELite preprocessing).
+	//! Freezes/defreezes a variable (a frozen var is exempt from Sat-preprocessing).
 	void    setFrozen(Var v, bool b);
-	//! Adds v to resp. removes v from the set of projection variables.
-	void    setProject(Var v, bool b)    { assert(validVar(v)); if (b != varInfo_[v].has(VarInfo::PROJECT)) varInfo_[v].toggle(VarInfo::PROJECT); }
 	//! Marks/unmarks v as contained in a negative loop or head of a choice rule.
-	void    setNant(Var v, bool b)       { assert(validVar(v)); if (b != varInfo_[v].has(VarInfo::NANT))    varInfo_[v].toggle(VarInfo::NANT);    }
-	//! Mark/unmark v as contained in a non-hcf disjunction.
-	void    setInDisj(Var v, bool b)     { assert(validVar(v)); if (b != varInfo_[v].has(VarInfo::DISJ))    varInfo_[v].toggle(VarInfo::DISJ);    }
-	//! Eliminates the variable v.
+	void    setNant(Var v, bool b)       { if (b != varInfo(v).has(VarInfo::NANT))    varInfo_[v].toggle(VarInfo::NANT);    }
+	//! Marks/unmarks v as contained in a non-hcf disjunction.
+	void    setInDisj(Var v, bool b)     { if (b != varInfo(v).has(VarInfo::DISJ))    varInfo_[v].toggle(VarInfo::DISJ);    }
+	//! Marks/unmarks v as projection variable.
+	void    setProject(Var v, bool b)    { if (b != varInfo(v).has(VarInfo::PROJECT)) varInfo_[v].toggle(VarInfo::PROJECT); }
+	void    setVarEq(Var v, bool b)      { if (b != varInfo(v).has(VarInfo::EQ))      varInfo_[v].toggle(VarInfo::EQ);      }
+	void    mark(Literal p)              { assert(validVar(p.var())); varInfo_[p.var()].rep |= (VarInfo::MARK_P + p.sign()); }
+	void    unmark(Var v)                { assert(validVar(v)); varInfo_[v].rep &= ~(VarInfo::MARK_P|VarInfo::MARK_N); }
+	//! Eliminates the variable v from the problem.
 	/*!
 	 * \pre v must not occur in any constraint and frozen(v) == false and value(v) == value_free
 	 */
 	void    eliminate(Var v);
-	void    setVarEq(Var v, bool b)      { assert(validVar(v)); if (b != varInfo_[v].has(VarInfo::EQ))      varInfo_[v].toggle(VarInfo::EQ);      }
-	void    mark(Literal p)              { assert(validVar(p.var())); varInfo_[p.var()].rep |= (VarInfo::MARK_P + p.sign()); }
-	void    unmark(Var v)                { assert(validVar(v)); varInfo_[v].rep &= ~(VarInfo::MARK_P|VarInfo::MARK_N); }
-	void    setPreserveModels()          { share_.satPreM = SatPreParams::prepro_preserve_models; }
-	//! Requests a special tag literal for tagging conditional knowledge.
-	/*!
-	 * Once a tag literal p is set, learnt clauses containing ~p are
-	 * tagged as "conditional". Conditional clauses can be removed from a solver
-	 * by calling Solver::removeConditional(). Furthermore, calling 
-	 * Solver::strengthenConditional() removes ~p from conditional clauses and
-	 * transforms them to unconditional knowledge.
-	 *
-	 * \note Typically, the tag literal is an initial assumption and hence true during 
-	 *       the whole search. 
-	 */
-	void    requestTagLiteral();
-	//! Requests a special literal for tagging enumeration/step-related knowledge.
-	void    requestStepLiteral();
-	
-	//! Returns the number of problem variables.
-	/*!
-	 * \note The special sentinel-var 0 is not counted, i.e. numVars() returns
-	 * the number of problem-variables.
-	 * To iterate over all problem variables use a loop like:
-	 * \code
-	 * for (Var i = 1; i <= numVars(); ++i) {...}
-	 * \endcode
-	 */
-	uint32  numVars()          const { return static_cast<uint32>(varInfo_.size() - 1); }
-	//! Returns the number of eliminated vars.
-	uint32  numEliminatedVars()const { return problem_.vars_eliminated; }
-	
-	//! Returns true if var represents a valid variable in this object.
-	/*!
-	 * \note The range of valid variables is [1..numVars()]. The variable 0
-	 * is a special sentinel variable. 
-	 */
-	bool    validVar(Var var)  const { return var <= numVars(); }
-	//! Returns the type of variable v.
-	/*!
-	 * If v was added with parameter eq=true, the return value
-	 * is Var_t::atom_body_var.
-	 */
-	VarType type(Var v)        const {
-		assert(validVar(v));
-		return varInfo_[v].has(VarInfo::EQ)
-			? Var_t::atom_body_var
-			: VarType(Var_t::atom_var + varInfo_[v].has(VarInfo::BODY));
-	}
-	//! Returns true if v is currently eliminated, i.e. no longer part of the problem.
-	bool    eliminated(Var v)  const;
-	//! Returns true if v is excluded from variable elimination.
-	bool    frozen(Var v)      const { assert(validVar(v)); return varInfo_[v].has(VarInfo::FROZEN); }
-	//! Returns true if v is a projection variable.
-	bool    project(Var v)     const { assert(validVar(v)); return varInfo_[v].has(VarInfo::PROJECT);}
-	//! Returns true if v is contained in a negative loop or head of a choice rule.
-	bool    nant(Var v)        const { assert(validVar(v)); return varInfo_[v].has(VarInfo::NANT);}
-	bool    inDisj(Var v)      const { assert(validVar(v)); return varInfo_[v].has(VarInfo::DISJ);}
-	Literal tagLiteral()       const { return tag_; }
-	Literal stepLiteral()      const { return step_; }
-	bool    marked(Literal p)  const { assert(validVar(p.var())); return varInfo_[p.var()].has(VarInfo::MARK_P + p.sign()); }
-	VarInfo info(Var v)        const { assert(validVar(v)); return varInfo_[v]; }
-	//! Returns the preferred decision literal of variable v w.r.t its type.
-	/*!
-	 * \return 
-	 *  - posLit(v) if type(v) == body_var
-	 *  - negLit(v) if type(v) == atom_var
-	 * \note If type(v) is atom_body_var, the preferred literal is determined
-	 *       by v's primary type, i.e. the one that was initially passed to addVar().
-	 */
-	Literal preferredLiteralByType(Var v) const {
-		assert(validVar(v));
-		return Literal(v, !varInfo_[v].has(VarInfo::BODY));
-	}
-	
+
 	//! Prepares the master solver so that constraints can be added.
 	/*!
 	 * Must be called to publish previously added variables to master solver
 	 * and before constraints over these variables can be added.
-	 * \post !frozen()
 	 * \return The master solver associated with this object.
 	 */
 	Solver& startAddConstraints(uint32 constraintGuess = 100);
-	//! Same as master()->addUnary(p, Constraint_t::static_constraint)
-	bool    addUnary(Literal p);
-	//! Same as master()->add(c)
+	
+	//! A convenience method for adding facts to the master.
+	bool    addUnary(Literal x);
+	//! A convenience method for adding binary clauses.
+	bool    addBinary(Literal x, Literal y);
+	//! A convenience method for adding ternary clauses.
+	bool    addTernary(Literal x, Literal y, Literal z);
+	//! A convenience method for adding constraints to the master.
 	void    add(Constraint* c);
-	//! Same as shortImplications->addBinary(p, q, learnt).
-	bool    addBinary(Literal p, Literal q, bool learnt = false)             { return btig_.addBinary(p, q, learnt); }
-	//! Same as shortImplications->addTernary(p, q, learnt).
-	bool    addTernary(Literal p, Literal q, Literal r, bool learnt = false) { return btig_.addTernary(p, q, r, learnt); }
-  
+
 	//! Finishes initialization of the master solver.
 	/*!
 	 * The function must be called once before search is started. After endInit()
@@ -614,18 +612,27 @@ public:
 	 * \post frozen()
 	 */
 	bool    endInit(bool attachAll = false);
+	//@}
 	
+	/*!
+	 * \name (Parallel) solving
+	 * Functions to be called during (parallel) solving.
+	 * 
+	 * \note If not otherwise noted, the functions in this group can be safely called 
+	 * from multiple threads.
+	 */
+	//@{
 	//! Attaches the solver with the given id to this object.
 	/*!
 	 * \note It is safe to attach multiple solvers concurrently
-	 * but the master solver shall not change during the whole
-	 * operation.
+	 * but the master solver shall not change during the whole operation.
 	 *
-	 * \pre id < concurrency()
+	 * \pre hasSolver(id)
 	 */
-	bool    attach(uint32 id) { return attach(*solver(id)); }
-	bool    attach(Solver& s);
-
+	bool     attach(uint32 id) { return attach(*solver(id)); }
+	bool     attach(Solver& s);
+	
+	
 	//! Detaches the solver with the given id from this object.
 	/*!
 	 * The function removes any tentative constraints from s.
@@ -634,75 +641,40 @@ public:
 	 *       i.e. in a parallel search different solvers may call detach()
 	 *       concurrently.
 	 */
-	void    detach(uint32 id, bool reset = false) { return detach(*solver(id), reset); }
-	void    detach(Solver& s, bool reset = false);
+	void     detach(uint32 id, bool reset = false) { return detach(*solver(id), reset); }
+	void     detach(Solver& s, bool reset = false);
 	
-	//! Returns the number of problem constraints.
-	uint32      numConstraints()    const;
-	//! Estimates the problem complexity.
+	DistrPtr distributor;/*!< Distributor object to use for distribution of learnt constraints.*/
+	
+	uint32   winner()             const   { return share_.winner; }
+	void     setWinner(uint32 sId)        { share_.winner = std::min(sId, concurrency()); }
+	
+	//! Simplifies the problem constraints w.r.t the master's assignment.
+	void     simplify(bool shuffle);
+	
+	//! Adds the given short implication to the short implication graph if possible.
 	/*!
-	 * \return sum of c->estimateComplexity(*master()), where c is a problem constraint.
+	 * \return 
+	 *   - > 0 if implication was added.
+	 *   - < 0 if implication can't be added because allowImplicit() is false for ct.
+	 *   - = 0 if implication is subsumed by some constraint in the short implication graph.
 	 */
-	uint32      problemComplexity() const;
-	//! Returns the size of top-level after last call to endInit()
-	size_type   topLevelSize()      const { return lastTopLevel_; }
+	int         addImp(ImpGraph::ImpType t, const Literal* lits, ConstraintType ct);
+	//! Returns the number of learnt short implications.
+	uint32      numLearntShort()     const { return btig_.numLearnt(); }
+	ImpGraphRef shortImplications()  const { return btig_; }
+	void        simplifyShort(const Solver& s, Literal p);
+	void               report(const Event& ev)                 const { if (progress_) progress_->dispatch(ev);  }
+	bool               report(const Solver& s, const Model& m) const { return !progress_ || progress_->onModel(s, m); }
+	void               initStats(Solver& s)                    const;
+	const SolverStats& stats(const Solver& s, bool accu)       const;
 	//@}
-
-	/*!
-	 * \name Learning and distribution
-	 * Functions for distributing knowledge.
-	 * 
-	 * \note The functions in this group can be safely called 
-	 * from multiple threads.
-	 */
-	//@{
-	//! Distributes the clause in lits via the distributor.
-	/*!
-	 * The function first calls the distribution strategy 
-	 * to decides whether the clause is a valid candidate for distribution.
-	 * If so and a distributor was set, it distributes the clause and returns a handle to the
-	 * now shared literals of the clause. Otherwise, it returns 0.
-	 *
-	 * \param owner The solver that created the clause.
-	 * \param lits  The literals of the clause.
-	 * \param size  The number of literals in the clause.
-	 * \param extra Additional information about the clause.
-	 * \note 
-	 *   If the return value is not null, it is the caller's 
-	 *   responsibility to release the returned handle (i.e. by calling release()).
-	 */
-	SharedLiterals* distribute(const Solver& owner, const Literal* lits, uint32 size, const ClauseInfo& extra) const;
-	//! Tries to receive at most maxOut clauses.
-	/*!
-	 * The function queries the distributor object for new clauses to be delivered to
-	 * the solver target. Clauses are stored in out.
-	 * \return The number of clauses received.
-	 */
-	uint32          receive(const Solver& target, SharedLiterals** out, uint32 maxOut) const;
-	//! Returns the number of learnt binary and ternary clauses
-	uint32          numLearntShort() const { return btig_.numLearnt(); }
-	//@}
-	const Stats& stats()     const   { return problem_; }
-	uint32       winner()    const   { return share_.winner; }
-	uint32       numBinary() const   { return btig_.numBinary();  }
-	uint32       numTernary()const   { return btig_.numTernary(); }
-	SymbolTable& symTab()    const   { return symTabPtr_->symTab; }
-	const BTIG&  shortImplications()                     const { return btig_; }
-	uint32       numShortImplications(Literal p)         const { return btig_.numEdges(p); }	
-	void         report(const Event& ev)                 const { if (progress_) progress_->dispatch(ev);  }
-	bool         report(const Solver& s, const Model& m) const { return !progress_ || progress_->onModel(s, m); }
-	void         initStats(Solver& s)                    const;
-	void         setWinner(uint32 x) { share_.winner = std::min(x, concurrency()); }
-	void         simplifyDB(bool shuffle);
-	void         simplify(const Solver& s, Literal p) {
-		if (!isShared()) { btig_.removeTrue(s, p); }
-	}
-	const SolverStats& stats(const Solver& s, bool accu)  const;
 private:
 	SharedContext(const SharedContext&);
 	SharedContext& operator=(const SharedContext&);
-	void init();
-	bool unfreezeStep(Solver& s);
+	void    init();
+	bool    unfreezeStep();
+	Literal addAuxLit();
 	typedef SingleOwnerPtr<Configuration> Config;
 	typedef PodVector<VarInfo>::type      VarVec;
 	typedef PodVector<SolverStats*>::type StatsVec;
@@ -713,13 +685,12 @@ private:
 	}*           symTabPtr_;     // pointer to shared symbol table
 	ProblemStats   problem_;     // problem statistics
 	VarVec         varInfo_;     // varInfo[v] stores info about variable v
-	BTIG           btig_;        // binary-/ternary implication graph
+	ImpGraph       btig_;        // binary-/ternary implication graph
 	Config         config_;      // active configuration
 	SolverVec      solvers_;     // solvers associated with this context
 	LogPtr         progress_;    // event handler or 0 if not used
-	Literal        tag_;         // literal for tagging learnt constraints
 	Literal        step_;        // literal for tagging enumeration/step constraints
-	size_type      lastTopLevel_;// size of master's top-level after last init
+	uint32         lastTopLevel_;// size of master's top-level after last init
 	struct Share {               // Additional data
 		uint32 count   :12;        //   max number of objects sharing this object
 		uint32 winner  :12;        //   id of solver that terminated the search
