@@ -104,65 +104,55 @@ private:
 	//@}
 public:
 	//! Returns a pointer to the shared context object of this solver.
-	const SharedContext*    sharedContext() const  { return shared_; }
+	const SharedContext*    sharedContext() const { return shared_; }
 	//! Returns a pointer to the sat-preprocessor used by this solver.
 	SatPreprocessor*        satPrepro()     const;
 	//! Returns the configuration for this object.
 	const SolverParams&     configuration() const;
-	const Heuristic&        heuristic()     const      { return heuristic_;}
-	uint32                  id()            const      { return strategy.id; }
-	Heuristic&              heuristic()                { return heuristic_;}
-	/*!
-	 * \name Adding constraints.
-	 */
-	//@{
-	//! Adds the unary constraint p to the solver.
-	/*!
-	 * \note Unary constraints are immediately asserted.
-	 * \return false if asserting p leads to a conflict.
-	 */
-	bool addUnary(Literal p, ConstraintType);	
+	//! Returns the solve parameters for this object.
+	const SolveParams&      searchConfig()  const;
+	const Heuristic&        heuristic()     const { return heuristic_;}
+	uint32                  id()            const { return strategy.id; }
+	Heuristic&              heuristic()           { return heuristic_;}
+	VarInfo                 varInfo(Var v)  const { return shared_->validVar(v) ? shared_->varInfo(v) : VarInfo(); }
+	const SymbolTable&      symbolTable()   const { return shared_->symbolTable(); }
+	Literal                 tagLiteral()    const { return tag_; }
 	//! Adds the problem constraint c to the solver.
-	void add(Constraint* c);
-	//! Adds the short constraint con to the short implication graph if possible.
-	/*! 
-	 * \note If con can't be stored in the short implication graph (because updates are
-	 *       not permitted or con is not a short implication), the function
-	 *       returns false without adding con. Otherwise, it also publishes con
-	 *       via a call to SharedContext::distribute().
-	 *
-	 * \see  SharedContext::allowImplicit(ConstraintType)
-	 */
-	bool addShort(const ClauseRep& con);
-	//! Adds the learnt constraint c to the solver.
 	/*!
-	 * \pre endInit() was called.
+	 * Problem constraints shall only be added to the master solver of 
+	 * a SharedContext object and only during the setup phase.
+	 * \pre this == sharedContext()->master() && !sharedContext()->frozen().
 	 */
-	void addLearnt(LearntConstraint* c, uint32 size, ConstraintType type) {
-		learnts_.push_back(c);
-		stats.addLearnt(size, type); 
+	void                    add(Constraint* c);
+	//! Adds a suitable representation of the given clause to the solver.
+	/*!
+	 * Depending on the type and size of the given clause, the function
+	 * either adds a (learnt) constraint to this solver or an implication
+	 * to the shared implication graph.
+	 * \note If c is a problem clause, the precondition of add(Constraint* c) applies.
+	 */
+	bool                    add(const ClauseRep& c, bool isNew = true);
+	//! Returns whether c can be stored in the shared short implication graph.
+	bool                    allowImplicit(const ClauseRep& c) const {
+		return c.isImp() 
+		  ? shared_->allowImplicit(c.info.type()) && !c.info.aux() && (c.prep == 1 || (!auxVar(c.lits[0].var()) && !auxVar(c.lits[1].var()) && (c.size == 2 || !auxVar(c.lits[2].var()))))
+		  : c.size <= 1;
 	}
-	void addLearnt(LearntConstraint* c, uint32 size) { addLearnt(c, size, c->type()); }
-	
+	//! Returns the post propagator with the given priority or 0 if no such post propagator exists.
+	PostPropagator*         getPost(uint32 prio) const;
 	//! Adds p as post propagator to this solver.
 	/*!
 	 * \pre p was not added previously and is not part of any other solver.
 	 * \note Post propagators are stored and called in priority order.
 	 * \see PostPropagator::priority()
 	 */
-	bool addPost(PostPropagator* p)    { return addPost(p, initPost_ != 0); }
+	bool                    addPost(PostPropagator* p) { return addPost(p, initPost_ != 0); }
 	//! Removes p from the solver's list of post propagators.
 	/*!
 	 * \note removePost(p) shall only be called during propagation
 	 *       of p or if no propagation is currently active.
 	 */
-	void removePost(PostPropagator* p) { post_.remove(p); }
-	typedef PostPropagator PP;
-	//! Returns the post propagator with the given priority or 0 if no such post propagator exists.
-	PP*  getPost(uint32 prio) const;
-	void setEnumerationConstraint(Constraint* c);
-	//@}
-	
+	void                    removePost(PostPropagator* p){ post_.remove(p); }
 	/*!
 	 * \name CDNL-functions.
 	 */
@@ -195,6 +185,8 @@ public:
 	//! Adds path to the current root-path and adjusts the root-level accordingly.
 	bool pushRoot(const LitVec& path);
 	bool pushRoot(Literal p);
+	
+	void setEnumerationConstraint(Constraint* c);
 
 	//! Moves the root-level i levels down (i.e. away from the top-level).
 	/*!
@@ -209,15 +201,17 @@ public:
 	
 	//! Moves the root-level i levels up (i.e. towards the top-level).
 	/*!
-	 * The function undos all levels between the new root level and the current decision level,
+	 * The function removes all levels between the new root level and the current decision level,
 	 * resets the current backtrack-level, and re-assigns any implied literals.
-	 * \param i Number of levels to pop.
+	 * \param      i      Number of root decisions to pop.
+	 * \param[out] popped Optional storage for popped root decisions.
+	 * \param      filter Whether or not aux variables should be added to popped.
 	 * \post decisionLevel() == rootLevel()
 	 * \note The function first calls clearStopConflict() to remove any stop conflicts.
 	 * \note The function *does not* propagate any asserted literals. It is
 	 *       the caller's responsibility to call propagate() after the function returned.
 	 */
-	bool popRootLevel(uint32 i = 1);
+	bool popRootLevel(uint32 i = 1, LitVec* popped = 0, bool aux = true);
 
 	//! Removes a previously set stop conflict and restores the root level.
 	void clearStopConflict();
@@ -233,6 +227,38 @@ public:
 	 */
 	bool clearAssumptions();
 
+	//! Adds the learnt constraint c to the solver.
+	void addLearnt(LearntConstraint* c, uint32 size, ConstraintType type) {
+		learnts_.push_back(c);
+		stats.addLearnt(size, type); 
+	}
+	void addLearnt(LearntConstraint* c, uint32 size) { addLearnt(c, size, c->type()); }
+	//! Tries to receive at most maxOut clauses.
+	/*!
+	 * The function queries the distributor object for new clauses to be delivered to
+	 * this solver. Clauses are stored in out.
+	 * \return The number of clauses received.
+	 */
+	uint32          receive(SharedLiterals** out, uint32 maxOut) const;
+	//! Distributes the clause in lits via the distributor.
+	/*!
+	 * The function first calls the distribution strategy 
+	 * to decides whether the clause is a valid candidate for distribution.
+	 * If so and a distributor was set, it distributes the clause and returns a handle to the
+	 * now shared literals of the clause. Otherwise, it returns 0.
+	 *
+	 * \param owner The solver that created the clause.
+	 * \param lits  The literals of the clause.
+	 * \param size  The number of literals in the clause.
+	 * \param extra Additional information about the clause.
+	 * \note 
+	 *   If the return value is not null, it is the caller's 
+	 *   responsibility to release the returned handle (i.e. by calling release()).
+	 * \note If the clause contains aux vars, it is not distributed.
+	 */
+	SharedLiterals* distribute(const Literal* lits, uint32 size, const ClauseInfo& extra);
+	
+
 	//! Returns to the maximum of rootLevel() and backtrackLevel() and increases the number of restarts.
 	void restart() {
 		undoUntil(0);
@@ -248,7 +274,13 @@ public:
 	uint32 backtrackLevel() const { return levels_.backtrack; }
 
 	//! Returns whether the solver can split-off work.
-	bool splittable() const { return decisionLevel() > rootLevel() && !frozenLevel(rootLevel()+1); }
+	bool splittable() const;
+
+	//! Tries to split-off disjoint work from the solver's currrent guiding path and returns it in out.
+	/*!
+	 * \return splittable()
+	 */
+	bool split(LitVec& out);
 
 	//! Copies the solver's currrent guiding path to gp.
 	/*!
@@ -257,11 +289,9 @@ public:
 	 *   - any literals that are implied on a level <= rootLevel() because of newly learnt
 	 *     information. This particularly includes literals that were flipped during model enumeration.
 	 * 
-	 * \param[in,out] gp          where to store the guiding path
-	 * \param[in,out] startPos    position in the trail from which copying should start
-	 * \param[in,out] numImplied  number of newly implied literals in gp
+	 * \param[out] gp where to store the guiding path
 	 */
-	void updateGuidingPath(LitVec& gp, LitVec::size_type& startPos, uint32& numImplied);
+	void copyGuidingPath(LitVec& out);
 
 	//! If called on top-level, removes SAT-clauses + Constraints for which Constraint::simplify returned true.
 	/*!
@@ -273,15 +303,27 @@ public:
 	//! Shuffle constraints upon next simplification.
 	void shuffleOnNextSimplify(){ shufSimp_ = 1; }
 
+	//! Requests a special aux variable for tagging conditional knowledge.
+	/*!
+	 * Once a tag variable t is set, learnt clauses containing ~t are
+	 * tagged as "conditional". Conditional clauses are removed once t becomes
+	 * unassigned or Solver::removeConditional() is called. Furthermore, calling 
+	 * Solver::strengthenConditional() removes ~t from conditional clauses and
+	 * transforms them to unconditional knowledge.
+	 *
+	 * \note Typically, the tag variable is a root assumption and hence true during 
+	 *       the whole search.
+	 */
+	Var  pushTagVar(bool pushToRoot);
 	//! Removes all conditional knowledge, i.e. all previously tagged learnt clauses.
 	/*!
-	 * \see SharedContext::requestTagLiteral()
+	 * \see Solver::pushTagVar()
 	 */
 	void removeConditional();
-
+	
 	//! Resolves all tagged clauses with the tag literal and thereby strengthens the learnt db.
 	/*!
-	 * \see SharedContext::requestTagLiteral()
+	 * \see Solver::pushTagVar()
 	 */
 	void strengthenConditional();
 
@@ -348,6 +390,8 @@ public:
 		impliedLits_.add(decisionLevel(), ImpliedLiteral(p, dl, r, d));
 		return (isTrue(p) && setReason(p, r, d)) || force(p, r, d);
 	}
+	//! Assigns p as a fact at decision level 0.
+	bool force(Literal p) { return force(p, 0, Antecedent(posLit(0))); }
 
 	//! Assumes the literal p if possible.
 	/*!
@@ -434,6 +478,7 @@ public:
 	 */
 	uint32 inDegree(WeightLitVec& out);
 	
+	struct DBInfo { uint32 size; uint32 locked; uint32 pinned; };
 	//! Removes upto remMax percent of the learnt nogoods.
 	/*!
 	 * \param remMax  Fraction of nogoods to remove ([0.0,1.0]).
@@ -443,7 +488,6 @@ public:
 	 *   Nogoods that are the reason for a literal to be in the assignment
 	 *   are said to be locked and won't be removed.
 	 */
-	struct DBInfo { uint32 size; uint32 locked; uint32 pinned; };
 	DBInfo reduceLearnts(float remMax, const ReduceStrategy& rs = ReduceStrategy());
 
 	//! Resolves the active conflict using the selected strategy.
@@ -474,13 +518,23 @@ public:
 	 * \post decisionLevel == max(min(decisionLevel(), dl), max(rootLevel(), btLevel))
 	 */
 	void undoUntil(uint32 dl);
-
+	
 	/*!
 	 * Similar to undoUntil but optionally also pops the backtrack-level
 	 * to dl if possible.
 	 * \return The current decision level.
 	 */
 	uint32 undoUntil(uint32 dl, bool popBtLevel);
+	
+	//! Adds a new auxiliary variable to this solver.
+	/*!
+	 * Auxiliary variables are local to one solver and are not considered
+	 * as part of the problem. They shall be added/used only during solving, i.e.
+	 * after problem setup is completed.
+	 */
+	Var  pushAuxVar();
+	//! Pops the num most recently added auxiliary variables from this solver.
+	void popAuxVar(uint32 num = UINT32_MAX);
 	//@}  
 	
 	/*!
@@ -490,18 +544,17 @@ public:
 	 * parameter.
 	 */
 	//@{
-	
 	//! Returns the number of problem variables.
-	/*!
-	 * \see SharedContext::numVars()
-	 */
+	uint32   numProblemVars()       const { return shared_->numVars(); }
+	//! Returns the number of active solver-local aux variables.
+	uint32   numAuxVars()           const { return numVars() - numProblemVars(); }
+	//! Returns the number of solver variables, i.e. numProblemVars() + numAuxVars()
 	uint32   numVars()              const { return assign_.numVars() - 1; }
 	//! Returns true if var represents a valid variable in this solver.
-	/*!
-	 * \see SharedContext::validVar(Var v)
-	 */
 	bool     validVar(Var var)      const { return var <= numVars(); }
-	//! Returns the number of assigned problem variables.
+	//! Returns true if var is a solver-local aux var.
+	bool     auxVar(Var var)        const { return shared_->numVars() < var; }
+	//! Returns the number of assigned variables.
 	uint32   numAssignedVars()      const { return assign_.assigned(); }
 	//! Returns the number of free variables.
 	/*!
@@ -511,6 +564,8 @@ public:
 	uint32   numFreeVars()          const { return assign_.free()-1; }
 	//! Returns the value of v w.r.t the current assignment.
 	ValueRep value(Var v)           const { assert(validVar(v)); return assign_.value(v); }
+	//! Returns the value of v w.r.t the top level.
+	ValueRep topValue(Var v)        const { return level(v) == 0 ? value(v) : value_free; }
 	//! Returns the set of preferred values of v.
 	ValueSet pref(Var v)            const { assert(validVar(v)); return assign_.pref(v);}
 	//! Returns true if p is true w.r.t the current assignment.
@@ -655,7 +710,7 @@ public:
 	 * know what you are doing!
 	 */
 	//@{
-	bool addPost(PostPropagator* p, bool init)    { post_.add(p, p->priority()); return !init || p->init(*this); }
+	bool addPost(PostPropagator* p, bool init) { post_.add(p, p->priority()); return !init || p->init(*this); }
 	//! Updates the reason for p being tue.
 	/*!
 	 * \pre p is true and x is a valid reason for p
@@ -715,9 +770,9 @@ public:
 	void  freeLearntBytes(uint64 bytes) { memUse_ -= (bytes < memUse_) ? bytes : memUse_; }
 	
 	//! simplifies cc and returns finalizeConflictClause(cc, info);
-	uint32  simplifyConflictClause(LitVec& cc, ClauseInfo& info, ClauseHead* rhs);
-	uint32  finalizeConflictClause(LitVec& cc, ClauseInfo& info);
-	uint32  countLevels(const Literal* first, const Literal* last, uint32 maxLevels);
+	uint32 simplifyConflictClause(LitVec& cc, ClauseInfo& info, ClauseHead* rhs);
+	uint32 finalizeConflictClause(LitVec& cc, ClauseInfo& info, uint32 ccRepMode = 0);
+	uint32 countLevels(const Literal* first, const Literal* last, uint32 maxLevels);
 	bool hasLevel(uint32 dl)    const { assert(validLevel(dl)); return levels_[dl-1].marked != 0; }
 	bool frozenLevel(uint32 dl) const { assert(validLevel(dl)); return levels_[dl-1].freeze != 0; }
 	void markLevel(uint32 dl)    { assert(validLevel(dl)); levels_[dl-1].marked = 1; }
@@ -783,7 +838,6 @@ private:
 	bool    validWatch(Literal p) const { return p.index() < (uint32)watches_.size(); }
 	void    freeMem();
 	bool    simplifySAT();
-	void    simplifyShort(Literal p);
 	bool    unitPropagate();
 	void    cancelPropagation() { assign_.qReset(); post_.cancel(); }
 	void    undoLevel(bool sp);
@@ -822,6 +876,7 @@ private:
 	VarVec            cflStamp_;    // temporary vector for computing number of conflicts in branch
 	ImpliedList       impliedLits_; // lits that were asserted on current dl but are logically implied earlier
 	ClauseInfo        ccInfo_;      // temporary: information about conflict clause cc_
+	Literal           tag_;         // aux literal for tagging learnt constraints
 	uint32            lbdTime_;     // temporary counter for computing lbd
 	uint32            dbIdx_;       // position of first new problem constraint in master db 
 	uint32            lastSimp_ :30;// number of top-level assignments on last call to simplify
@@ -833,15 +888,6 @@ inline bool isRevLit(const Solver& s, Literal p, uint32 maxL) {
 	return s.isFalse(p) && (s.seen(p) || s.level(p.var()) < maxL);
 }
 
-template <class T>
-inline bool LegacyPostPropagator<T>::propagateFixpoint(Solver& s, PostPropagator* p) {
-	for (;;) {
-		if (!_propagate(s,p,derived())){ return false; }
-		if (s.queueSize() == 0)        { return true;  }
-		if (!s.propagateUntil(this))   { return false; }
-	}
-}
-
 template <class C>
 void simplifyDB(Solver& s, C& db, bool shuffle) {
 	typename C::size_type j = 0;
@@ -886,12 +932,18 @@ public:
 	virtual void endInit(Solver& /* s */) { }  
 	
 	/*!
-	 * Called for each var that changes its state from eliminated back to normal.
-	 * The default-implementation is a noop.
-	 * \param s Solver in which v is resurrected.
-	 * \param v The variable that is resurrected.
+	 * Called if the state of one or more variables changed. 
+	 * A state change is one of: 
+	 *   - A previously eliminated variable is resurrected.
+	 *   - A new aux variable was added.
+	 *   - An aux variable was removed.
+	 *   .
+	 * \param s Solver in which the state change occurred.
+	 * \param v The first variable affected by the change.
+	 * \param n The range of variables affected, i.e. [v, v+n).
+	 * \note Use s.validVar(v) and s.auxVar(v) to determine the reason for the update.
 	 */
-	virtual void resurrect(const Solver& /* s */, Var /* v */) {}
+	virtual void updateVar(const Solver& /* s */, Var /* v */, uint32 /* n */) = 0;
 	
 	/*!
 	 * Called on decision level 0. Variables that are assigned on this level
@@ -988,11 +1040,11 @@ public:
 	static Literal defaultLiteral(Solver& s, Var v) {
 		switch(s.strategy.signDef) {
 			default: // 
-			case SolverStrategies::sign_atom: return Literal(v, !s.sharedContext()->info(v).has(VarInfo::BODY));
+			case SolverStrategies::sign_atom: return Literal(v, !s.varInfo(v).has(VarInfo::BODY));
 			case SolverStrategies::sign_no  : return posLit(v);
 			case SolverStrategies::sign_yes : return negLit(v);
 			case SolverStrategies::sign_rnd : return Literal(v, s.rng.drand() < 0.5);
-			case SolverStrategies::sign_disj: return Literal(v, !s.sharedContext()->info(v).has(VarInfo::BODY|VarInfo::DISJ));
+			case SolverStrategies::sign_disj: return Literal(v, !s.varInfo(v).has(VarInfo::BODY|VarInfo::DISJ));
 		}
 	}
 private:
@@ -1002,6 +1054,8 @@ private:
 
 //! Selects the first free literal w.r.t to the initial variable order.
 class SelectFirst : public DecisionHeuristic {
+public:
+	void updateVar(const Solver& /* s */, Var /* v */, uint32 /* n */) {}
 private:
 	Literal doSelect(Solver& s);
 };
