@@ -61,6 +61,8 @@ inline bool matchStatPath(const char*& x, const char* path) {
 #define CLASP_STAT_DEFINE(m, k, a, accu) m
 #define CLASP_STAT_GET(m, k, a, accu) if (std::strcmp(key, k) == 0) { return double( (a) ); }
 #define CLASP_STAT_KEY(m, k, a, accu) k "\0"
+#define NO_ARG
+
 //! A struct for holding core statistics used by a solver.
 /*!
  * Core statistics are always present in a solver and hence
@@ -80,17 +82,17 @@ struct CoreStats {
 		CLASP_CORE_STATS(CLASP_STAT_ACCU, (*this), o)
 	}
 	double operator[](const char* key) const {
-		CLASP_CORE_STATS(CLASP_STAT_GET, (*this), )
+		CLASP_CORE_STATS(CLASP_STAT_GET, (*this), NO_ARG)
 		return -1.0;
 	}
 	static const char* keys(const char* path) {
-		if (!path || !*path) return CLASP_CORE_STATS(CLASP_STAT_KEY,,);
+		if (!path || !*path) return CLASP_CORE_STATS(CLASP_STAT_KEY,NO_ARG,NO_ARG);
 		return 0;
 	}
 	uint64 backtracks() const { return conflicts-analyzed; }
 	uint64 backjumps()  const { return analyzed; }
 	double avgRestart() const { return ratio(analyzed, restarts); }
-	CLASP_CORE_STATS(CLASP_STAT_DEFINE, , )
+	CLASP_CORE_STATS(CLASP_STAT_DEFINE, NO_ARG, NO_ARG)
 };
 
 //! A struct for holding (optional) extra statistics.
@@ -98,6 +100,7 @@ struct ExtendedStats {
 	typedef ConstraintType type_t;
 	typedef uint64 Array[Constraint_t::max_value];
 #define CLASP_EXTENDED_STATS(STAT, SELF, OTHER)     \
+	STAT(uint64 domChoices; /**< "domain" choices   */, STRING(domChoices) , SELF.domChoices , SELF.domChoices += OTHER.domChoices) \
 	STAT(uint64 models;     /**< number of models   */, STRING(models)     , SELF.models     , SELF.models     += OTHER.models)     \
 	STAT(uint64 modelLits;  /**< DLs in models      */, STRING(modelLits)  , SELF.modelLits  , SELF.modelLits  += OTHER.modelLits)  \
 	STAT(uint64 hccTests;   /**< stability tests    */, STRING(hccTests)   , SELF.hccTests   , SELF.hccTests   += OTHER.hccTests)   \
@@ -106,8 +109,8 @@ struct ExtendedStats {
 	STAT(uint64 distributed;/**< lemmas distributed */, STRING(distributed), SELF.distributed, SELF.distributed+= OTHER.distributed)\
 	STAT(uint64 sumDistLbd; /**< sum of lemma lbds  */, STRING(sumDistLbd) , SELF.sumDistLbd , SELF.sumDistLbd += OTHER.sumDistLbd) \
 	STAT(uint64 integrated; /**< lemmas integrated  */, STRING(integrated) , SELF.integrated , SELF.integrated += OTHER.integrated) \
-	STAT(Array learnts;  /**< lemmas of type t-1    */, "lemmas"           , SELF.lemmas()   , )   \
-	STAT(Array lits;     /**< lits of type t-1      */, "learntLits"       , SELF.learntLits(), )   \
+	STAT(Array learnts;  /**< lemmas of type t-1    */, "lemmas"           , SELF.lemmas()   , NO_ARG)   \
+	STAT(Array lits;     /**< lits of type t-1      */, "learntLits"       , SELF.learntLits(), NO_ARG)  \
 	STAT(uint32 binary;  /**< binary lemmas         */, STRING(binary)     , SELF.binary  , SELF.binary  += OTHER.binary)  \
 	STAT(uint32 ternary; /**< ternary lemmas        */, STRING(ternary)    , SELF.ternary , SELF.ternary += OTHER.ternary) \
 	STAT(double cpuTime; /**< cpu time used         */, STRING(cpuTime)    , SELF.cpuTime , SELF.cpuTime += OTHER.cpuTime) \
@@ -116,12 +119,12 @@ struct ExtendedStats {
 	STAT(uint64 gpLits;  /**< lits in received gps  */, STRING(gpLits)     , SELF.gpLits  , SELF.gpLits  += OTHER.gpLits)  \
 	STAT(uint32 gps;     /**< guiding paths received*/, STRING(gps)        , SELF.gps     , SELF.gps     += OTHER.gps)     \
 	STAT(uint32 splits;  /**< split requests handled*/, STRING(splits)     , SELF.splits  , SELF.splits  += OTHER.splits)  \
-	STAT(             , "conflictLemmas", SELF.lemmas(Constraint_t::learnt_conflict), SELF.learnts[0] += OTHER.learnts[0]) \
-	STAT(             , "loopLemmas"    , SELF.lemmas(Constraint_t::learnt_loop)    , SELF.learnts[1] += OTHER.learnts[1]) \
-	STAT(             , "otherLemmas"   , SELF.lemmas(Constraint_t::learnt_other)   , SELF.learnts[2] += OTHER.learnts[2]) \
-	STAT(             , "conflictLits"  , SELF.lits[Constraint_t::learnt_conflict-1], SELF.lits[0]    += OTHER.lits[0])    \
-	STAT(             , "loopLits"      , SELF.lits[Constraint_t::learnt_loop-1]    , SELF.lits[1]    += OTHER.lits[1])    \
-	STAT(             , "otherLits"     , SELF.lits[Constraint_t::learnt_other-1]   , SELF.lits[2]    += OTHER.lits[2])
+	STAT(NO_ARG       , "conflictLemmas", SELF.lemmas(Constraint_t::learnt_conflict), SELF.learnts[0] += OTHER.learnts[0]) \
+	STAT(NO_ARG       , "loopLemmas"    , SELF.lemmas(Constraint_t::learnt_loop)    , SELF.learnts[1] += OTHER.learnts[1]) \
+	STAT(NO_ARG       , "otherLemmas"   , SELF.lemmas(Constraint_t::learnt_other)   , SELF.learnts[2] += OTHER.learnts[2]) \
+	STAT(NO_ARG       , "conflictLits"  , SELF.lits[Constraint_t::learnt_conflict-1], SELF.lits[0]    += OTHER.lits[0])    \
+	STAT(NO_ARG       , "loopLits"      , SELF.lits[Constraint_t::learnt_loop-1]    , SELF.lits[1]    += OTHER.lits[1])    \
+	STAT(NO_ARG       , "otherLits"     , SELF.lits[Constraint_t::learnt_other-1]   , SELF.lits[2]    += OTHER.lits[2])
 	
 	ExtendedStats() { reset(); }
 	void reset() { std::memset(this, 0, sizeof(ExtendedStats)); }
@@ -129,11 +132,11 @@ struct ExtendedStats {
 		CLASP_EXTENDED_STATS(CLASP_STAT_ACCU, (*this), o)
 	}
 	double operator[](const char* key) const {
-		CLASP_EXTENDED_STATS(CLASP_STAT_GET, (*this), )
+		CLASP_EXTENDED_STATS(CLASP_STAT_GET, (*this), NO_ARG)
 		return -1.0;
 	}
 	static const char* keys(const char* path) {
-		if (!path || !*path) { return CLASP_EXTENDED_STATS(CLASP_STAT_KEY,,); }
+		if (!path || !*path) { return CLASP_EXTENDED_STATS(CLASP_STAT_KEY,NO_ARG,NO_ARG); }
 		return 0;
 	}
 	void addLearnt(uint32 size, type_t t) {
@@ -153,16 +156,16 @@ struct ExtendedStats {
 	double avgIntJump()    const { return ratio(intJumps, intImps); }
 	double avgGp()         const { return ratio(gpLits, gps); }
 	double intRatio()      const { return ratio(integrated, distributed); }
-	CLASP_EXTENDED_STATS(CLASP_STAT_DEFINE,,)
+	CLASP_EXTENDED_STATS(CLASP_STAT_DEFINE,NO_ARG,NO_ARG)
 };
 
 //! A struct for holding (optional) jump statistics.
 struct JumpStats {
 #define CLASP_JUMP_STATS(STAT, SELF, OTHER)     \
 	STAT(uint64 jumps;    , STRING(jumps)    , SELF.jumps    , SELF.jumps    += OTHER.jumps)   \
-	STAT(uint64 bounded;  , STRING(bounded)  , SELF.bounded  , SELF.bounded  += SELF.bounded)  \
-	STAT(uint64 jumpSum;  , STRING(jumpSum)  , SELF.jumpSum  , SELF.jumpSum  += SELF.jumpSum)  \
-	STAT(uint64 boundSum; , STRING(boundSum) , SELF.boundSum , SELF.boundSum += SELF.boundSum) \
+	STAT(uint64 bounded;  , STRING(bounded)  , SELF.bounded  , SELF.bounded  += OTHER.bounded)  \
+	STAT(uint64 jumpSum;  , STRING(jumpSum)  , SELF.jumpSum  , SELF.jumpSum  += OTHER.jumpSum)  \
+	STAT(uint64 boundSum; , STRING(boundSum) , SELF.boundSum , SELF.boundSum += OTHER.boundSum) \
 	STAT(uint32 maxJump;  , STRING(maxJump)  , SELF.maxJump  , MAX_MEM(SELF.maxJump, OTHER.maxJump))   \
 	STAT(uint32 maxJumpEx;, STRING(maxJumpEx), SELF.maxJumpEx, MAX_MEM(SELF.maxJumpEx,OTHER.maxJumpEx)) \
 	STAT(uint32 maxBound; , STRING(maxBound) , SELF.maxBound , MAX_MEM(SELF.maxBound, OTHER.maxBound)) 
@@ -175,11 +178,11 @@ struct JumpStats {
 #undef MAX_MEM
 	}
 	double operator[](const char* key) const {
-		CLASP_JUMP_STATS(CLASP_STAT_GET, (*this), )
+		CLASP_JUMP_STATS(CLASP_STAT_GET, (*this), NO_ARG)
 		return -1.0;
 	}
 	static const char* keys(const char* path) {
-		if (!path || !*path) { return CLASP_JUMP_STATS(CLASP_STAT_KEY,,); }
+		if (!path || !*path) { return CLASP_JUMP_STATS(CLASP_STAT_KEY,NO_ARG,NO_ARG); }
 		return 0;
 	}
 	void update(uint32 dl, uint32 uipLevel, uint32 bLevel) {
@@ -199,7 +202,7 @@ struct JumpStats {
 	double avgBound()   const { return ratio(boundSum, bounded); }
 	double avgJump()    const { return ratio(jumpSum, jumps); }
 	double avgJumpEx()  const { return ratio(jumped(), jumps); }
-	CLASP_JUMP_STATS(CLASP_STAT_DEFINE,,)
+	CLASP_JUMP_STATS(CLASP_STAT_DEFINE,NO_ARG,NO_ARG)
 };
 
 struct QueueImpl {
@@ -300,9 +303,9 @@ struct SolverStats : public CoreStats {
 	const char* subKeys(const char* p) const;
 	const char* keys(const char* path) const {
 		if (!path || !*path) {
-			if (jumps && extra) { return "jumps.\0extra.\0" CLASP_CORE_STATS(CLASP_STAT_KEY,,); }
-			if (extra)          { return "extra.\0" CLASP_CORE_STATS(CLASP_STAT_KEY,,); }
-			if (jumps)          { return "jumps.\0" CLASP_CORE_STATS(CLASP_STAT_KEY,,); }
+			if (jumps && extra) { return "jumps.\0extra.\0" CLASP_CORE_STATS(CLASP_STAT_KEY,NO_ARG,NO_ARG); }
+			if (extra)          { return "extra.\0" CLASP_CORE_STATS(CLASP_STAT_KEY,NO_ARG,NO_ARG); }
+			if (jumps)          { return "jumps.\0" CLASP_CORE_STATS(CLASP_STAT_KEY,NO_ARG,NO_ARG); }
 			return CoreStats::keys(path);
 		}
 		return subKeys(path);
@@ -315,6 +318,7 @@ struct SolverStats : public CoreStats {
 	inline void addModel(uint32 decisionLevel);
 	inline void addCpuTime(double t);
 	inline void addSplit(uint32 num = 1);
+	inline void addDomChoice(uint32 num = 1);
 	inline void addIntegratedAsserting(uint32 receivedDL, uint32 jumpDL);
 	inline void addIntegrated(uint32 num = 1);
 	inline void removeIntegrated(uint32 num = 1);
@@ -334,6 +338,7 @@ inline void SolverStats::addSplit(uint32 num)                      { if (extra)
 inline void SolverStats::addPath(const LitVec::size_type& sz)      { if (extra) { ++extra->gps; extra->gpLits += sz; } }
 inline void SolverStats::addTest(bool partial)                     { if (extra) { ++extra->hccTests; extra->hccPartial += (uint32)partial; } }
 inline void SolverStats::addModel(uint32 DL)                       { if (extra) { ++extra->models; extra->modelLits += DL; } }
+inline void SolverStats::addDomChoice(uint32 n)                    { if (extra) { extra->domChoices += n; } }
 inline void SolverStats::addIntegratedAsserting(uint32 rDL, uint32 jDL) {
 	if (extra) { ++extra->intImps; extra->intJumps += (rDL - jDL); }
 }
@@ -349,6 +354,7 @@ inline void SolverStats::updateJumps(uint32 dl, uint32 uipLevel, uint32 bLevel,
 #undef CLASP_CORE_STATS
 #undef CLASP_EXTENDED_STATS
 #undef CLASP_JUMP_STATS
+#undef NO_ARG
 ///////////////////////////////////////////////////////////////////////////////
 // Clauses
 ///////////////////////////////////////////////////////////////////////////////
@@ -356,8 +362,8 @@ inline void SolverStats::updateJumps(uint32 dl, uint32 uipLevel, uint32 bLevel,
 class ClauseInfo {
 public:
 	typedef ClauseInfo self_type;
-	enum { MAX_LBD = Activity::MAX_LBD, MAX_ACTIVITY = (1<<22)-1 }; 
-	ClauseInfo(ConstraintType t = Constraint_t::static_constraint) : act_(0), lbd_(MAX_LBD), type_(t), tag_(0) {
+	enum { MAX_LBD = Activity::MAX_LBD, MAX_ACTIVITY = (1<<21)-1 }; 
+	ClauseInfo(ConstraintType t = Constraint_t::static_constraint) : act_(0), lbd_(MAX_LBD), type_(t), tag_(0), aux_(0) {
 		static_assert(sizeof(self_type) == sizeof(uint32), "Unsupported padding");
 	}
 	bool           learnt()   const { return type() != Constraint_t::static_constraint; }
@@ -365,38 +371,31 @@ public:
 	uint32         activity() const { return static_cast<uint32>(act_); }
 	uint32         lbd()      const { return static_cast<uint32>(lbd_); }
 	bool           tagged()   const { return tag_ != 0; }
+	bool           aux()      const { return tagged() || aux_ != 0; }
 	self_type&     setType(ConstraintType t) { type_  = static_cast<uint32>(t); return *this; }
 	self_type&     setActivity(uint32 act)   { act_   = std::min(act, (uint32)MAX_ACTIVITY); return *this; }
 	self_type&     setTagged(bool b)         { tag_   = static_cast<uint32>(b); return *this; }
 	self_type&     setLbd(uint32 a_lbd)      { lbd_   = std::min(a_lbd, (uint32)MAX_LBD); return *this; }
+	self_type&     setAux(bool b)            { aux_   = static_cast<uint32>(b); return *this; }
 private:
-	uint32 act_ : 22; // Activity of clause
+	uint32 act_ : 21; // Activity of clause
 	uint32 lbd_ :  7; // Literal block distance in the range [0, MAX_LBD]
 	uint32 type_:  2; // One of ConstraintType
 	uint32 tag_ :  1; // Conditional constraint?
+	uint32 aux_ :  1; // Contains solver-local aux vars?
 };
 //! Primitive representation of a clause.
 struct ClauseRep {
-	Literal*   lits; /*!< Pointer to array of literals (not owned!). */
-	uint32     size; /*!< Size of array of literals. */
-	ClauseInfo info; /*!< Additional clause info.    */
+	static ClauseRep create(Literal* cl, uint32 sz, const ClauseInfo& i = ClauseInfo())  { return ClauseRep(cl, sz, false, i);}
+	static ClauseRep prepared(Literal* cl, uint32 sz, const ClauseInfo& i = ClauseInfo()){ return ClauseRep(cl, sz, true, i); }
+	ClauseRep(Literal* cl = 0, uint32 sz = 0, bool p = false, const ClauseInfo& i = ClauseInfo()) : info(i), size(sz), prep(uint32(p)), lits(cl) {}
+	ClauseInfo info;    /*!< Additional clause info.    */
+	uint32     size:31; /*!< Size of array of literals. */
+	uint32     prep: 1; /*!< Whether lits is already prepared. */
+	Literal*   lits;    /*!< Pointer to array of literals (not owned!). */
 	bool isImp() const { return size > 1 && size < 4; }
 };
-//! Primitive representation of an implicit short clause.
-struct ShortRep : ClauseRep { 
-	ShortRep(const Literal* c, uint32 n, const ClauseInfo& i = ClauseInfo()) { 
-		assert(3 >= n);  std::memcpy(buf, c, n*sizeof(Literal)); 
-		lits = buf; size = n; info = i;
-	}
-	ShortRep(Literal x, Literal y = negLit(0), Literal z = negLit(0)) {
-		lits = buf; size = 0;
-		if (x != negLit(0)) { buf[size++] = x; }
-		if (y != negLit(0)) { buf[size++] = y; }
-		if (z != negLit(0)) { buf[size++] = z; }
-	}
-	ShortRep& type(ConstraintType t) { info.setType(t); return *this; }
-	Literal buf[3]; 
-};
+
 //! (Abstract) base class for clause types.
 /*!
  * ClauseHead is used to enforce a common memory-layout for all clauses.
@@ -765,20 +764,21 @@ public:
 	 * know what you are doing!
 	 */
 	//@{
-	bool seen(Var v, uint8 m) const { return (assign_[v] & (m<<2)) != 0; }
-	void setSeen(Var v, uint8 m)    { assign_[v] |= (m<<2); }
-	void clearSeen(Var v)           { assign_[v] &= ~uint32(12); }
-	void clearValue(Var v)          { assign_[v] &= ~uint32(3); }
-	void setValue(Var v, ValueRep val) {
+	uint32 units()              const { return units_; }
+	bool   seen(Var v, uint8 m) const { return (assign_[v] & (m<<2)) != 0; }
+	void   setSeen(Var v, uint8 m)    { assign_[v] |= (m<<2); }
+	void   clearSeen(Var v)           { assign_[v] &= ~uint32(12); }
+	void   clearValue(Var v)          { assign_[v] &= ~uint32(3); }
+	void   setValue(Var v, ValueRep val) {
 		assert(value(v) == val || value(v) == value_free);
 		assign_[v] |= val;
 	}
-	void setReason(Var v, const Antecedent& a) { reason_[v] = a;  }
-	void setData(Var v, uint32 data) { reason_.setData(v, data); }
-	void setPref(Var v, ValueSet::Value which, ValueRep to) { pref_[v].set(which, to); }
-	void copyAssignment(Assignment& o) const { o.assign_ = assign_; }
-	bool markUnits()                 { while (units_ != front) { setSeen(trail[units_++].var(), 3u); }  return true; }
-	void setUnits(uint32 ts)         { units_ = ts; }
+	void   setReason(Var v, const Antecedent& a) { reason_[v] = a;  }
+	void   setData(Var v, uint32 data) { reason_.setData(v, data); }
+	void   setPref(Var v, ValueSet::Value which, ValueRep to) { pref_[v].set(which, to); }
+	void   copyAssignment(Assignment& o) const { o.assign_ = assign_; }
+	bool   markUnits()                 { while (units_ != front) { setSeen(trail[units_++].var(), 3u); }  return true; }
+	void   setUnits(uint32 ts)         { units_ = ts; }
 	//@}
 private:
 	static const uint32 elim_mask = uint32(0xFFFFFFF0u);
