@@ -89,7 +89,7 @@ bool Preprocessor::preprocessEq(uint32 maxIters) {
 				(*it)->clearLiteral(false);
 				(*it)->setInUpper(false);
 			}
-			prg_->ctx()->popVars(startVar);
+			prg_->ctx()->resizeVars(startVar);
 			litToNode_.clear();
 		}
 		VarVec& supported = prg_->getSupportedBodies(true);
