@@ -54,12 +54,12 @@ public:
 	typedef Solver::ConstraintDB              ConstraintDB;
 	typedef SharedLiterals                    SharedLits;
 	CBFinder(Solver& s, MinimizeConstraint* min, SharedCon* sh) 
-		: EnumerationConstraint(s, min), shared(sh), last(0) { setIgnoreSymmetric(true); }
+		: EnumerationConstraint(s, min), shared(sh), last(0) {}
 	
 	Constraint* cloneAttach(Solver& s) { return new CBFinder(s, cloneMinimizer(s), shared); } 
 	void doCommitModel(Enumerator& ctx, Solver& s) { static_cast<CBConsequences&>(ctx).addCurrent(s, current, s.model); }
 	void destroy(Solver* s, bool detach);
-	bool doUpdate(Solver& s, bool disjoint);
+	bool doUpdate(Solver& s);
 	void pushLocked(Solver& s, ClauseHead* h);
 	LitVec       current;
 	SharedCon*   shared;
@@ -82,7 +82,7 @@ CBConsequences::~CBConsequences() {
 }
 EnumerationConstraint* CBConsequences::doInit(SharedContext& ctx, MinimizeConstraint* min, int) {
 	cons_.clear();
-	const SymbolTable& index = ctx.symTab();
+	const SymbolTable& index = ctx.symbolTable();
 	if (index.type() == SymbolTable::map_direct) {
 		for (Var v = 1, end = index.size(); v < end; ++v) { 
 			if (!ctx.marked(posLit(v))) {
@@ -107,16 +107,20 @@ EnumerationConstraint* CBConsequences::doInit(SharedContext& ctx, MinimizeConstr
 	}
 	delete shared_;
 	shared_ = ctx.concurrency() > 1 ? new SharedConstraint() : 0;
+	setIgnoreSymmetric(true);
 	return new CBFinder(*ctx.master(), min, shared_);
 }
 void CBConsequences::addCurrent(Solver& s, LitVec& con, ValueVec& m) {
 	con.clear();
 	con.push_back(~s.sharedContext()->stepLiteral());
+	for (LitVec::iterator it = cons_.begin(), end = cons_.end(); it != end; ++it) {
+		m[it->var()] = 0;
+	}
 	if (type_ == brave_consequences) {
 		for (LitVec::iterator it = cons_.begin(), end = cons_.end(); it != end; ++it) {
 			Literal& p = *it;
 			if (s.isTrue(p) || p.watched())  { 
-				m[p.var()] = trueValue(p); 
+				m[p.var()] |= trueValue(p); 
 				p.watch();
 			}
 			else if (s.level(p.var())) {
@@ -128,11 +132,12 @@ void CBConsequences::addCurrent(Solver& s, LitVec& con, ValueVec& m) {
 		for (LitVec::iterator it = cons_.begin(), end = cons_.end(); it != end; ++it) {
 			Literal& p = *it;
 			if (!s.isTrue(p) || !p.watched()) {
-				m[p.var()] = falseValue(p);
+				m[p.var()] &= ~trueValue(p);
 				p.clearWatch();
 			}
-			else if (s.level(p.var())) {
-				con.push_back(~p);
+			else {
+				if (s.level(p.var())) { con.push_back(~p); }
+				m[p.var()] |= trueValue(p);
 			}
 		}
 	}
@@ -158,7 +163,7 @@ void CBConsequences::CBFinder::pushLocked(Solver& s, ClauseHead* c) {
 	}
 	locked.push_back(c);
 }
-bool CBConsequences::CBFinder::doUpdate(Solver& s, bool) {
+bool CBConsequences::CBFinder::doUpdate(Solver& s) {
 	ClauseCreator::Result ret;
 	uint32 flags = ClauseCreator::clause_explicit|ClauseCreator::clause_no_add;
 	if (!shared) {
