@@ -87,7 +87,7 @@ void format(const Clasp::mt::MessageEvent& ev, char* out, uint32 outSize) {
 /////////////////////////////////////////////////////////////////////////////////////////
 // Output
 /////////////////////////////////////////////////////////////////////////////////////////
-Output::Output(uint32 verb) : summary_(0), verbose_(0) {
+Output::Output(uint32 verb) : summary_(0), verbose_(0), hidePref_(0) {
 	std::memset(quiet_, 0, sizeof(quiet_));
 	setCallQuiet(print_no);
 	setVerbosity(verb);
@@ -104,11 +104,12 @@ void Output::setVerbosity(uint32 verb) {
 void Output::setModelQuiet(PrintLevel model){ quiet_[0] = static_cast<uint8>(model); }
 void Output::setOptQuiet(PrintLevel opt)    { quiet_[1] = static_cast<uint8>(opt);   }
 void Output::setCallQuiet(PrintLevel call)  { quiet_[2] = static_cast<uint8>(call);  }
+void Output::setHide(char c)                { hidePref_ = c; }
 
 void Output::saveModel(const Model& m) {
-	vals_.assign(m.model, m.model + m.size);
-	saved_       = m;
-	saved_.model = &vals_[0];
+	vals_         = *m.values;
+	saved_        = m;
+	saved_.values = &vals_;
 }
 
 void Output::onEvent(const Event& ev) {
@@ -123,17 +124,18 @@ void Output::onEvent(const Event& ev) {
 }
 bool Output::onModel(const Solver& s, const Model& m) {
 	if (modelQ() == print_all || (optQ() == print_all && m.costs)) {
-		printModel(s.sharedContext()->symTab(), m, print_all);
+		printModel(s.symbolTable(), m, print_all);
 	}
 	if (modelQ() == print_best || (optQ() == print_best && m.costs)) {
-		saveModel(m);
+		if (m.opt == 1 && !m.consequences()) { printModel(s.symbolTable(), m, print_best); clearModel(); }
+		else                                 { saveModel(m); }
 	}
 	return true;
 }
 void Output::startStep(const ClaspFacade&) { clearModel(); summary_ = 0; }
 void Output::stopStep(const ClaspFacade::Summary& s){
 	if (getModel()) {
-		printModel(s.ctx().symTab(), *getModel(), print_best);
+		printModel(s.ctx().symbolTable(), *getModel(), print_best);
 		clearModel();
 	}
 	if (callQ() == print_all) { 
@@ -268,6 +270,9 @@ void JsonOutput::visitCoreSolverStats(double cpuTime, uint64 models, const Solve
 
 void JsonOutput::visitExtSolverStats(const ExtendedStats& stx, bool accu) {
 	pushObject("More");
+	if (stx.domChoices) {
+		printKeyValue("DomChoices", stx.domChoices);
+	}
 	if (stx.hccTests) {
 		pushObject("StabTests");
 		printKeyValue("Sum", stx.hccTests);
@@ -470,7 +475,7 @@ void JsonOutput::printModel(const SymbolTable& index, const Model& m, PrintLevel
 		printf("%-*s", indent(), " ");
 		if (index.type() == SymbolTable::map_indirect) {
 			for (SymbolTable::const_iterator it = index.begin(); it != index.end(); ++it) {
-				if (m.value(it->second.lit.var()) == trueValue(it->second.lit) && !it->second.name.empty()) {
+				if (m.value(it->second.lit.var()) == trueValue(it->second.lit) && doPrint(it->second)) {
 					printString(it->second.name.c_str(), sep);
 					sep = ", ";
 				}	
@@ -491,7 +496,7 @@ void JsonOutput::printModel(const SymbolTable& index, const Model& m, PrintLevel
 	if (hasModel) { popObject(); }
 }
 void JsonOutput::printCosts(const SharedMinimizeData& opt) { 
-	pushObject("Opt", type_array);
+	pushObject("Costs", type_array);
 	printf("%-*s", indent(), " ");
 	const char* sep = "";
 	for (uint32 i = 0, end = opt.numRules(); i != end; ++i) {
@@ -510,14 +515,14 @@ void JsonOutput::printSummary(const ClaspFacade::Summary& run, bool final) {
 	if (verbosity()) {
 		if (run.result.interrupted()){ printKeyValue(run.result.signal != SIGALRM ? "INTERRUPTED" : "TIME LIMIT", uint32(1));  }
 		pushObject("Models");
-		printKeyValue("Number", run.models());
+		printKeyValue("Number", run.enumerated());
 		printKeyValue("More"  , run.complete() ? "no" : "yes");
 		if (run.sat()) {
-			if (run.models() != run.enumerated()){ printKeyValue("Enumerated", run.enumerated()); }
-			if (run.consequences())              { printKeyValue(run.consequences(), run.complete() ? "yes":"unknown"); }
-			if (run.costs())                     { 
-				printKeyValue("Optimum", run.complete()?"yes":"unknown");
-				if (run.optimize()) { printCosts(*run.costs()); }
+			if (run.consequences()){ printKeyValue(run.consequences(), run.complete() ? "yes":"unknown"); }
+			if (run.optimize())    { 
+				printKeyValue("Optimum", run.optimum()?"yes":"unknown"); 
+				printKeyValue("Optimal", run.optimal());
+				printCosts(*run.costs());
 			}
 		}
 		popObject();
@@ -545,7 +550,8 @@ void JsonOutput::printStatistics(const ClaspFacade::Summary& summary, bool final
 // TextOutput
 /////////////////////////////////////////////////////////////////////////////////////////
 #define printKeyValue(k, fmt, value) printf("%s%-*s: " fmt, format[cat_comment], width_, (k), (value))
-#define printLN(cat, fmt, ...)       printf("%s" fmt "\n", format[cat], ##__VA_ARGS__)
+#define printLN(cat, fmt, ...)       printf("%s" fmt "\n", format[cat], __VA_ARGS__)
+#define printBR(cat)                 printf("%s\n", format[cat])
 #define printKey(k)                  printf("%s%-*s: ", format[cat_comment], width_, (k))
 const char* const rowSep   = "----------------------------------------------------------------------------|";
 const char* const finalSep = "=============================== Accumulation ===============================|";
@@ -556,7 +562,7 @@ static inline std::string prettify(const std::string& str) {
 	t.append(str.end()-38, str.end());
 	return t;
 }
-TextOutput::TextOutput(uint32 verbosity, Format f, char ifs) : Output(verbosity), stTime_(0.0), state_(0) {
+TextOutput::TextOutput(uint32 verbosity, Format f, const char* catAtom, char ifs) : Output(verbosity), stTime_(0.0), state_(0) {
 	result[res_unknonw]    = "UNKNOWN";
 	result[res_sat]        = "SATISFIABLE";
 	result[res_unsat]      = "UNSATISFIABLE";
@@ -566,13 +572,12 @@ TextOutput::TextOutput(uint32 verbosity, Format f, char ifs) : Output(verbosity)
 	format[cat_objective]  = "Optimization: ";
 	format[cat_result]     = "";
 	format[cat_value_term] = "";
-	format[cat_atom_pre]   = "";
-	format[cat_atom_post]  = "";
+	format[cat_atom]       = "%s";
 	if (f == format_aspcomp) {
 		format[cat_comment]   = "% ";
 		format[cat_value]     = "ANSWER\n";
 		format[cat_objective] = "COST ";
-		format[cat_atom_post] = ".";
+		format[cat_atom]      = "%s.";
 		result[res_sat]       = "";
 		result[res_unsat]     = "INCONSISTENT";
 		result[res_opt]       = "OPTIMUM";
@@ -585,12 +590,29 @@ TextOutput::TextOutput(uint32 verbosity, Format f, char ifs) : Output(verbosity)
 		format[cat_objective] = "o ";
 		format[cat_result]    = "s ";
 		format[cat_value_term]= "0";
+		format[cat_atom]      = "%d";
 		if (f == format_pb09) {
 			format[cat_value_term]= "";
-			format[cat_atom_pre]  = "x";
+			format[cat_atom]      = "x%d";
 			setModelQuiet(print_best);
 		}
 	}
+	if (catAtom && *catAtom) {
+		format[cat_atom] = catAtom;
+		char ca = f == format_sat09 || f == format_pb09 ? 'd' : 's';
+		while (*catAtom) {
+			if      (*catAtom == '\n'){ break; }
+			else if (*catAtom == '%') {
+				if      (!*++catAtom)    { ca = '%'; break; }
+				else if (*catAtom == ca) { ca = 0; }
+				else if (*catAtom != '%'){ break; }
+			}
+			++catAtom;
+		}
+		CLASP_FAIL_IF(ca != 0         , "cat_atom: Invalid format string - format '%%%c' expected!", ca);
+		CLASP_FAIL_IF(*catAtom == '\n', "cat_atom: Invalid format string - new line not allowed!");
+		CLASP_FAIL_IF(*catAtom != 0   , "cat_atom: Invalid format string - '%%%c' too many arguments!", *catAtom);
+	}
 	ifs_[0] = ifs;
 	ifs_[1] = 0;
 	width_  = 13+(int)strlen(format[cat_comment]);
@@ -626,30 +648,21 @@ void TextOutput::printSummary(const ClaspFacade::Summary& run, bool final) {
 	else if (run.sat())   { res = !run.optimum() ? result[res_sat] : result[res_opt]; }
 	if (std::strlen(res)) { printLN(cat_result, "%s", res); }
 	if (verbosity() || run.stats()) {
-		printLN(cat_comment, "");
+		printBR(cat_comment);
 		if (run.result.interrupted()){ printKeyValue((run.result.signal != SIGALRM ? "INTERRUPTED" : "TIME LIMIT"), "%u\n", uint32(1));  }
 		printKey("Models");
-		if (!run.complete()) {
-			char buf[64];
-			int wr    = sprintf(buf, "%" PRIu64, run.models());
-			buf[wr]   = '+';
-			buf[wr+1] = 0;
-			printf("%-6s\n", buf);
-		}
-		else { printf("%-6" PRIu64"\n", run.models()); }
+		char buf[64];
+		int wr = sprintf(buf, "%" PRIu64, run.enumerated());
+		if (!run.complete()) { buf[wr++] = '+'; }
+		buf[wr]=0;
+		printf("%-6s\n", buf);
 		if (run.sat()) {
-			if (run.enumerated() != run.models()) { 
-				printKeyValue("  Enumerated", "%" PRIu64"\n", run.enumerated());
-			}
-			if (run.consequences()) {
-				printLN(cat_comment, "  %-*s: %s", width_-2, run.consequences(), (run.complete()?"yes":"unknown"));
-			}
-			if (run.costs()) {
-				printKeyValue("  Optimum", "%s\n", run.complete()?"yes":"unknown");
-				if (run.optimize()) {
-					printKey("Optimization");
-					printCosts(*run.costs());
-				}
+			if (run.consequences()) { printLN(cat_comment, "  %-*s: %s", width_-2, run.consequences(), (run.complete()?"yes":"unknown")); }
+			if (run.costs())        { printKeyValue("  Optimum", "%s\n", run.optimum()?"yes":"unknown"); }
+			if (run.optimize())     {
+				if (run.optimal() > 1){ printKeyValue("  Optimal", "%" PRIu64"\n", run.optimal()); }
+				printKey("Optimization");
+				printCosts(*run.costs());
 				printf("\n");
 			}
 		}
@@ -662,13 +675,13 @@ void TextOutput::printSummary(const ClaspFacade::Summary& run, bool final) {
 			, run.unsatTime);
 		printKeyValue("CPU Time", "%.3fs\n", run.cpuTime);
 		if (run.ctx().concurrency() > 1) {
-			printKeyValue("Threads", "%-6u", run.ctx().concurrency());
+			printKeyValue("Threads", "%-8u", run.ctx().concurrency());
 			printf(" (Winner: %u)\n", run.ctx().winner());
 		}
 	}
 }
 void TextOutput::printStatistics(const ClaspFacade::Summary& run, bool final) {
-	printLN(cat_comment, "");
+	printBR(cat_comment);
 	StatsVisitor::visitStats(run.ctx(), run.lpStats(), final && run.step);
 }
 void TextOutput::startStep(const ClaspFacade& f) {
@@ -743,8 +756,8 @@ void TextOutput::printSolveProgress(const Event& ev) {
 	fflush(stdout);
 }
 
-void TextOutput::printSep(CategoryKey k) const {
-	printf("%s%s", ifs_, *ifs_ != '\n' ? "" : format[k]);
+int TextOutput::printSep(CategoryKey k) const {
+	return printf("%s%s", ifs_, *ifs_ != '\n' ? "" : format[k]);
 }
 void TextOutput::printModel(const SymbolTable& sym, const Model& m, PrintLevel x) {
 	if (x == modelQ()) {
@@ -757,8 +770,8 @@ void TextOutput::printModel(const SymbolTable& sym, const Model& m, PrintLevel x
 			bool   const onlyTrue   = m.consequences();
 			for (Var v = 1; v < sym.size(); ++v) {
 				if (!onlyTrue || m.isTrue(posLit(v))) {
-					if (printed) { printSep(cat_value); }
-					printed += printf("%s%s%u%s", m.value(v) == value_false ? "-":"", format[cat_atom_pre], v, format[cat_atom_post]);
+					if (printed) { printed += printSep(cat_value); }
+					printed += printf(format[cat_atom], m.value(v) == value_false ? -int(v) : int(v));
 					if (printed >= maxLineLen) {
 						printed = 0;
 						printf("\n%s", format[cat_value]);
@@ -783,9 +796,9 @@ void TextOutput::printModel(const SymbolTable& sym, const Model& m, PrintLevel x
 void TextOutput::printNames(const Clasp::SymbolTable& sym, const Clasp::Model& m) {
 	bool first = true;
 	for (SymbolTable::const_iterator it = sym.begin(); it != sym.end(); ++it) {
-		if (m.isTrue(it->second.lit) && !it->second.name.empty()) {
+		if (m.isTrue(it->second.lit) && doPrint(it->second)) {
 			if (!first) { printSep(cat_value); }
-			printf("%s%s%s", format[cat_atom_pre], it->second.name.c_str(), format[cat_atom_post]);
+			printf(format[cat_atom], it->second.name.c_str());
 			first = false;
 		}
 	}
@@ -800,29 +813,29 @@ void TextOutput::printCosts(const SharedMinimizeData& costs) const {
 }
 void TextOutput::startSection(const char* n) const {
 	printLN(cat_comment, "============ %s Stats ============", n);
-	printLN(cat_comment, "");
+	printBR(cat_comment);
 }
 void TextOutput::startObject(const char* n, uint32 i) const {
 	printLN(cat_comment, "[%s %u]", n, i);
-	printLN(cat_comment, "");
+	printBR(cat_comment);
 }
 void TextOutput::visitProblemStats(const ProblemStats& stats, const Asp::LpStats* lp) {
 	StatsVisitor::visitProblemStats(stats, lp);
-	printLN(cat_comment, "");
+	printBR(cat_comment);
 }
 void TextOutput::visitSolverStats(const Clasp::SolverStats& s, bool accu) {
 	StatsVisitor::visitSolverStats(s, accu);
-	printLN(cat_comment, "");
+	printBR(cat_comment);
 }
 	
 void TextOutput::visitLogicProgramStats(const Asp::LpStats& lp) {
 	using namespace Asp;
-	printKeyValue("Atoms", "%-6u", lp.atoms);
+	printKeyValue("Atoms", "%-8u", lp.atoms);
 	if (lp.auxAtoms) {
 		printf(" (Original: %u Auxiliary: %u)", lp.atoms-lp.auxAtoms, lp.auxAtoms);
 	}
 	printf("\n");
-	printKeyValue("Rules", "%-6u", lp.rules());
+	printKeyValue("Rules", "%-8u", lp.rules());
 	printf(" ");
 	char space = '(', close = ' ';
 	for (Asp::RuleType i = BASICRULE; i != ENDRULE; ++i) {
@@ -835,9 +848,9 @@ void TextOutput::visitLogicProgramStats(const Asp::LpStats& lp) {
 		}
 	}
 	printf("%c\n", close);
-	printKeyValue("Bodies", "%-6u\n", lp.bodies);
+	printKeyValue("Bodies", "%-8u\n", lp.bodies);
 	if (lp.eqs() > 0) {
-		printKeyValue("Equivalences", "%-6u", lp.eqs());
+		printKeyValue("Equivalences", "%-8u", lp.eqs());
 		printf(" (Atom=Atom: %u Body=Body: %u Other: %u)\n" 
 			, lp.eqs(Var_t::atom_var)
 			, lp.eqs(Var_t::body_var)
@@ -845,15 +858,15 @@ void TextOutput::visitLogicProgramStats(const Asp::LpStats& lp) {
 	}
 	printKey("Tight");
 	if      (lp.sccs == 0)              { printf("Yes"); }
-	else if (lp.sccs != PrgNode::noScc) { printf("%-6s (SCCs: %u Non-Hcfs: %u Nodes: %u Gammas: %u)", "No", lp.sccs, lp.nonHcfs, lp.ufsNodes, lp.gammas); }
+	else if (lp.sccs != PrgNode::noScc) { printf("%-8s (SCCs: %u Non-Hcfs: %u Nodes: %u Gammas: %u)", "No", lp.sccs, lp.nonHcfs, lp.ufsNodes, lp.gammas); }
 	else                                 { printf("N/A"); }
 	printf("\n");
 }
 void TextOutput::visitProblemStats(const ProblemStats& ps) {
 	uint32 sum = ps.constraints + ps.constraints_binary + ps.constraints_ternary;
-	printKeyValue("Variables", "%-6u", ps.vars);
+	printKeyValue("Variables", "%-8u", ps.vars);
 	printf(" (Eliminated: %4u Frozen: %4u)\n", ps.vars_eliminated, ps.vars_frozen);
-	printKeyValue("Constraints", "%-6u", sum);
+	printKeyValue("Constraints", "%-8u", sum);
 	printf(" (Binary:%5.1f%% Ternary:%5.1f%% Other:%5.1f%%)\n"
 		, percent(ps.constraints_binary, sum)
 		, percent(ps.constraints_ternary, sum)
@@ -864,10 +877,12 @@ void TextOutput::visitCoreSolverStats(double cpuTime, uint64 models, const Solve
 		printKeyValue("CPU Time",  "%.3fs\n", cpuTime);
 		printKeyValue("Models", "%" PRIu64"\n", models);
 	}
-	printKeyValue("Choices", "%" PRIu64"\n", st.choices);
-	printKeyValue("Conflicts", "%-6" PRIu64"", st.conflicts);
+	printKeyValue("Choices", "%-8" PRIu64, st.choices);
+	if (st.extra && st.extra->domChoices) { printf(" (Domain: %" PRIu64")", st.extra->domChoices); }
+	printf("\n");
+	printKeyValue("Conflicts", "%-8" PRIu64"", st.conflicts);
 	printf(" (Analyzed: %" PRIu64")\n", st.backjumps());
-	printKeyValue("Restarts", "%-6" PRIu64"", st.restarts);
+	printKeyValue("Restarts", "%-8" PRIu64"", st.restarts);
 	if (st.restarts) {
 		printf(" (Average: %.2f Last: %" PRIu64")", st.avgRestart(), st.lastRestart);
 	}
@@ -875,49 +890,50 @@ void TextOutput::visitCoreSolverStats(double cpuTime, uint64 models, const Solve
 }
 void TextOutput::visitExtSolverStats(const ExtendedStats& stx, bool accu) {
 	if (stx.hccTests) {
-		printKeyValue("Stab. Tests", "%-6" PRIu64, stx.hccTests);
+		printKeyValue("Stab. Tests", "%-8" PRIu64, stx.hccTests);
 		printf(" (Full: %" PRIu64" Partial: %" PRIu64")\n", stx.hccTests - stx.hccPartial, stx.hccPartial);
 	}
 	if (stx.models) {
-		printKeyValue("Model-Level", "%-6.1f\n", stx.avgModel());
+		printKeyValue("Model-Level", "%-8.1f\n", stx.avgModel());
 	}
-	printKeyValue("Problems", "%-6" PRIu64,  (uint64)stx.gps);
+	printKeyValue("Problems", "%-8" PRIu64,  (uint64)stx.gps);
 	printf(" (Average Length: %.2f Splits: %" PRIu64")\n", stx.avgGp(), (uint64)stx.splits);
 	uint64 sum          = stx.lemmas();
-	printKeyValue("Lemmas", "%-6" PRIu64, sum);
+	printKeyValue("Lemmas", "%-8" PRIu64, sum);
 	printf(" (Deleted: %" PRIu64")\n",  stx.deleted);
 	const char* names[] = {"  Conflict", "  Loop", "  Other"};
 	for (int i = 0; i <= Constraint_t::learnt_other; ++i) {
 		if (i == 0) {
-			printKeyValue("  Binary", "%-6" PRIu64, uint64(stx.binary));
+			printKeyValue("  Binary", "%-8" PRIu64, uint64(stx.binary));
 			printf(" (Ratio: %6.2f%%)\n", percent(stx.binary, sum));
-			printKeyValue("  Ternary", "%-6" PRIu64, uint64(stx.ternary));
+			printKeyValue("  Ternary", "%-8" PRIu64, uint64(stx.ternary));
 			printf(" (Ratio: %6.2f%%)\n", percent(stx.ternary, sum));
 		}
 		else {
-			printKeyValue(names[i-1], "%-6" PRIu64, stx.lemmas(ConstraintType(i)));
+			printKeyValue(names[i-1], "%-8" PRIu64, stx.lemmas(ConstraintType(i)));
 			printf(" (Average Length: %6.1f Ratio: %6.2f%%) \n", stx.avgLen(ConstraintType(i)), percent(stx.lemmas(ConstraintType(i)), sum));
 		}
 	}
 	if (stx.distributed || stx.integrated) {
-		printKeyValue("  Distributed", "%-6" PRIu64, stx.distributed);
+		printKeyValue("  Distributed", "%-8" PRIu64, stx.distributed);
 		printf(" (Ratio: %6.2f%% Average LBD: %.2f) \n", stx.distRatio()*100.0, stx.avgDistLbd());
-		printKeyValue("  Integrated", "%-6" PRIu64, stx.integrated);
+		printKeyValue("  Integrated", "%-8" PRIu64, stx.integrated);
 		if (accu){ printf(" (Ratio: %6.2f%% ", stx.intRatio()*100.0); }
 		else     { printf(" ("); }
 		printf("Unit: %" PRIu64" Average Jumps: %.2f)\n", stx.intImps, stx.avgIntJump());
 	}	
 }
 void TextOutput::visitJumpStats(const JumpStats& st, bool) {
-	printKeyValue("Backjumps", "%-6" PRIu64, st.jumps);
+	printKeyValue("Backjumps", "%-8" PRIu64, st.jumps);
 	printf(" (Average: %5.2f Max: %3u Sum: %6" PRIu64")\n", st.avgJump(), st.maxJump, st.jumpSum);
-	printKeyValue("  Executed", "%-6" PRIu64, st.jumps-st.bounded);
+	printKeyValue("  Executed", "%-8" PRIu64, st.jumps-st.bounded);
 	printf(" (Average: %5.2f Max: %3u Sum: %6" PRIu64" Ratio: %6.2f%%)\n", st.avgJumpEx(), st.maxJumpEx, st.jumped(), st.jumpedRatio()*100.0);
-	printKeyValue("  Bounded", "%-6" PRIu64, st.bounded);
+	printKeyValue("  Bounded", "%-8" PRIu64, st.bounded);
 	printf(" (Average: %5.2f Max: %3u Sum: %6" PRIu64" Ratio: %6.2f%%)\n", st.avgBound(), st.maxBound, st.boundSum, 100.0 - (st.jumpedRatio()*100.0));
 }
 
 #undef printKeyValue
 #undef printKey
 #undef printLN
+#undef printBR
 }}
