@@ -41,7 +41,7 @@ struct Sink {
 };
 
 void* alloc(uint32 size) {
-	#pragma message TODO("replace with CACHE_LINE_ALIGNED alloc")
+	CLASP_PRAGMA_TODO("replace with CACHE_LINE_ALIGNED alloc")
 	return ::operator new(size);
 }
 void free(void* mem) {
@@ -123,17 +123,19 @@ uint32 ClauseCreator::watchOrder(const Solver& s, Literal p) {
 }
 
 ClauseRep ClauseCreator::prepare(Solver& s, const Literal* in, uint32 inSize, const ClauseInfo& e, uint32 flags, Literal* out, uint32 outMax) {
-	ClauseRep ret = { out, 0, e }; 
 	assert(out && outMax > 2);
+	ClauseRep ret  = ClauseRep::prepared(out, 0, e);
 	uint32 abst_w1 = 0, abst_w2 = 0;
 	bool simplify  = ((flags & clause_force_simplify) != 0) && inSize > 2 && outMax >= inSize;
-	Literal tag    = ~s.sharedContext()->tagLiteral();
+	Literal tag    = ~s.tagLiteral();
+	Var     vMax   = 0;
 	for (uint32 i = 0, j = 0, MAX_OUT = outMax - 1; i != inSize; ++i) {
 		Literal p     = in[i];
 		uint32 abst_p = watchOrder(s, p);
 		if ((abst_p + 1) > 1 && (!simplify || !s.seen(p.var()))) {
 			out[j] = p;
 			if (p == tag)         { ret.info.setTagged(true); }
+			if (p.var() > vMax)   { vMax = p.var();}
 			if (simplify)         { s.markSeen(p); }
 			if (abst_p > abst_w1) { std::swap(abst_p, abst_w1); std::swap(out[0], out[j]); }
 			if (abst_p > abst_w2) { std::swap(abst_p, abst_w2); std::swap(out[1], out[j]); }
@@ -152,6 +154,7 @@ ClauseRep ClauseCreator::prepare(Solver& s, const Literal* in, uint32 inSize, co
 		out[0]   = abst_w1 == UINT32_MAX || out[0] == ~out[1] ? posLit(0) : out[0]; 
 		ret.size = 1;
 	}
+	ret.info.setAux(s.auxVar(vMax));
 	return ret;
 }
 
@@ -163,8 +166,7 @@ ClauseRep ClauseCreator::prepare(Solver& s, LitVec& lits, uint32 flags, const Cl
 		shrinkVecTo(lits, x.size);
 		return x;
 	}
-	ClauseRep x = { &lits[0], (uint32)lits.size(), info };
-	return x;
+	return ClauseRep::prepared(&lits[0], (uint32)lits.size(), info);
 }
 
 ClauseRep ClauseCreator::prepare(bool forceSimplify) {
@@ -191,10 +193,10 @@ ClauseCreator::Status ClauseCreator::status(const Solver& s, const ClauseRep& c)
 }
 
 bool ClauseCreator::ignoreClause(const Solver& s, const ClauseRep& c, Status st, uint32 flags) {
-	if (st == status_sat) {
-		return (flags & clause_not_sat) != 0 || ((flags & clause_not_root_sat) != 0 && s.level(c.lits[0].var()) <= s.rootLevel());
-	}
-	return st == status_subsumed || ((st & status_unsat) != 0 && (flags & clause_not_conflict) != 0);
+	uint32 x = (st & (status_sat|status_unsat));
+	if (x == status_open)  { return false; }
+	if (x == status_unsat) { return st != status_empty && (flags & clause_not_conflict) != 0; }
+	return st == status_subsumed || (st == status_sat && ( (flags & clause_not_sat) != 0 || ((flags & clause_not_root_sat) != 0 && s.level(c.lits[0].var()) <= s.rootLevel())));
 }
 
 ClauseCreator::Result ClauseCreator::end(uint32 flags) {
@@ -232,6 +234,7 @@ ClauseHead* ClauseCreator::newProblemClause(Solver& s, const ClauseRep& clause,
 		ret = Clause::newShared(s, SharedLiterals::newShareable(clause.lits, clause.size, clause.info.type(), 1), clause.info, clause.lits, false);
 	}
 	if ( (flags & clause_no_add) == 0 ) {
+		assert(!clause.info.aux());
 		s.add(ret);
 	}
 	return ret;
@@ -240,13 +243,12 @@ ClauseHead* ClauseCreator::newProblemClause(Solver& s, const ClauseRep& clause,
 ClauseHead* ClauseCreator::newLearntClause(Solver& s, const ClauseRep& clause, uint32 flags) {
 	ClauseHead* ret;
 	Detail::Sink sharedPtr(0);
-	sharedPtr.clause = s.sharedContext()->distribute(s, clause.lits, clause.size, clause.info);
+	sharedPtr.clause = s.distribute(clause.lits, clause.size, clause.info);
 	if (clause.size <= Clause::MAX_SHORT_LEN || sharedPtr.clause == 0) {
 		if (!s.isFalse(clause.lits[1]) || !s.strategy.compress || clause.size < s.strategy.compress) {
 			ret = Clause::newClause(s, clause);
 		}
 		else {
-			std::stable_sort(clause.lits+2, clause.lits+clause.size, Detail::GreaterLevel(s));
 			ret = Clause::newContractedClause(s, clause, 2, true);
 		}
 	}
@@ -268,25 +270,28 @@ ClauseHead* ClauseCreator::newUnshared(Solver& s, SharedLiterals* clause, const
 			temp.push_back(*x);
 		}
 	}
-	ClauseRep rep = { &temp[0], (uint32)temp.size(), e };
-	return Clause::newClause(s, rep);
+	return Clause::newClause(s, ClauseRep::prepared(&temp[0], (uint32)temp.size(), e));
 }
 
 ClauseCreator::Result ClauseCreator::create_prepared(Solver& s, const ClauseRep& clause, uint32 flags) {
-	assert(s.decisionLevel() == 0 || clause.info.learnt());
+	assert(s.decisionLevel() == 0 || (clause.info.learnt() && clause.prep));
+	Status x = status(s, clause);
+	if (ignoreClause(s, clause, x, flags)){ 
+		return Result(0, x);
+	}
 	if (clause.size > 1) {
-		Status x = status(s, clause);
-		if (ignoreClause(s, clause, x, flags)){ 
-			return Result(0, x); 
-		}
-		if (!clause.info.learnt() && s.satPrepro()) {
+		Result ret(0, x);
+		if (!clause.info.learnt() && s.satPrepro() && !s.sharedContext()->frozen()) {
 			return Result(0, s.satPrepro()->addClause(clause.lits, clause.size) ? x : status_unsat);
 		}
-		s.heuristic()->newConstraint(s, clause.lits, clause.size, clause.info.type());
-		Result ret(0, x);
-		if (clause.size > 3 || clause.info.tagged() || (flags&clause_explicit) != 0 || !s.addShort(clause)) {
+		if ((flags & clause_no_heuristic) == 0) { s.heuristic()->newConstraint(s, clause.lits, clause.size, clause.info.type()); }
+		if (clause.size > 3 || (flags&clause_explicit) != 0 || !s.allowImplicit(clause)) {
 			ret.local = clause.info.learnt() ? newLearntClause(s, clause, flags) : newProblemClause(s, clause, flags);
 		}
+		else {
+			// add implicit short rep
+			s.add(clause);
+		}
 		if ((x & (status_unit|status_unsat)) != 0) {
 			Antecedent ante(ret.local);
 			if (!ret.local){ ante = clause.size == 3 ? Antecedent(~clause.lits[1], ~clause.lits[2]) : Antecedent(~clause.lits[1]); }
@@ -294,8 +299,8 @@ ClauseCreator::Result ClauseCreator::create_prepared(Solver& s, const ClauseRep&
 		}
 		return ret;
 	}
-	Literal imp = clause.size == 1	? clause.lits[0] : negLit(0);
-	return Result(0, s.addUnary(imp, clause.info.type()) ? status_unit : status_unsat);
+	s.add(clause);
+	return Result(0, !s.hasConflict() ? status_unit : status_unsat);
 }
 
 ClauseCreator::Result ClauseCreator::create(Solver& s, LitVec& lits, uint32 flags, const ClauseInfo& extra) {
@@ -303,9 +308,9 @@ ClauseCreator::Result ClauseCreator::create(Solver& s, LitVec& lits, uint32 flag
 }
 
 ClauseCreator::Result ClauseCreator::create(Solver& s, const ClauseRep& rep, uint32 flags) {
-	return create_prepared(s
-	  , (flags & clause_no_prepare) != 0 ? rep : prepare(s, rep.lits, rep.size, rep.info, flags, rep.lits)
-	  , flags);
+	return create_prepared(s, (rep.prep == 0 && (flags & clause_no_prepare) == 0 
+		? prepare(s, rep.lits, rep.size, rep.info, flags, rep.lits)
+		: ClauseRep::prepared(rep.lits, rep.size, rep.info)), flags);
 }
 
 ClauseCreator::Result ClauseCreator::integrate(Solver& s, SharedLiterals* clause, uint32 modeFlags, ConstraintType t) {
@@ -314,13 +319,13 @@ ClauseCreator::Result ClauseCreator::integrate(Solver& s, SharedLiterals* clause
 	// determine state of clause
 	Literal temp[Clause::MAX_SHORT_LEN]; temp[0] = temp[1] = negLit(0);
 	ClauseRep x    = prepare(s, clause->begin(), clause->size(), ClauseInfo(t), 0, temp, Clause::MAX_SHORT_LEN);
-	uint32 impSize = (modeFlags & clause_explicit) != 0 || !s.sharedContext()->allowImplicit(t) ? 1 : 3;
+	uint32 impSize = (modeFlags & clause_explicit) != 0 || !s.allowImplicit(x) ? 1 : 3;
 	Status xs      = status(s, x);
 	if (ignoreClause(s, x, xs, modeFlags)) {
 		return Result(0, xs);
 	}
 	Result result(0, xs);
-	s.heuristic()->newConstraint(s, clause->begin(), clause->size(), t);
+	if ((modeFlags & clause_no_heuristic) == 0) { s.heuristic()->newConstraint(s, clause->begin(), clause->size(), t); }
 	if (x.size > Clause::MAX_SHORT_LEN && s.sharedContext()->physicalShare(t)) {
 		result.local  = Clause::newShared(s, clause, x.info, temp, shared.clause == 0);
 		shared.clause = 0;
@@ -374,6 +379,7 @@ ClauseHead* Clause::newShared(Solver& s, SharedLiterals* shared_lits, const Clau
 
 ClauseHead* Clause::newContractedClause(Solver& s, const ClauseRep& rep, uint32 tailStart, bool extend) {
 	assert(rep.size >= 2);
+	if (extend) { std::stable_sort(rep.lits+tailStart, rep.lits+rep.size, Detail::GreaterLevel(s)); }
 	return new (alloc(s, rep.size, rep.info.learnt())) Clause(s, rep, tailStart, extend);
 }
 
@@ -658,7 +664,7 @@ ClauseHead::BoolPair Clause::strengthen(Solver& s, Literal p, bool toShort) {
 		ret.first = *it == p;
 		eot       = *it == p ? removeFromTail(s, it, eot) : it + 1;
 	}
-	if (ret.first && ~p == s.sharedContext()->tagLiteral()) {
+	if (ret.first && ~p == s.tagLiteral()) {
 		clearTagged();
 	}
 	ret.second = toShort && eot == t.first && toImplication(s);
@@ -795,8 +801,7 @@ bool SharedLitsClause::simplify(Solver& s, bool reinit) {
 		detach(s);
 		SharedLitsClause::destroy(0, false);
 		// construct short clause in "this"
-		ClauseRep raw = { lits, static_cast<uint32>(j-lits),  ClauseInfo() };
-		ClauseHead* h = Clause::newClause(this, s, raw);
+		ClauseHead* h = Clause::newClause(this, s, ClauseRep::prepared(lits, static_cast<uint32>(j-lits)));
 		// restore extra data - safe because memory layout has not changed!
 		info_.rep = rep;
 		return h->simplify(s, reinit);
@@ -1087,7 +1092,7 @@ bool LoopFormula::simplify(Solver& s, bool) {
 	// - remove all false bodies
 	// - find the watched bodies
 	for (i = 1; i != end_-1; ++i) {
-		assert( !s.isFalse(lits_[i]) || !lits_[i].watched() ); // otherwise should be asserting 
+		assert( !s.isFalse(lits_[i]) || !lits_[i].watched() || s.isTrue(lits_[other_]) );
 		if (!s.isFalse(lits_[i])) {
 			sat |= s.isTrue(lits_[i]);
 			if (i != j) { lits_[j] = lits_[i]; }
