@@ -17,7 +17,6 @@
 // along with Clasp; if not, write to the Free Software
 // Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 //
-
 #ifdef _MSC_VER
 #pragma warning (disable : 4146) // unary minus operator applied to unsigned type, result still unsigned
 #pragma warning (disable : 4996) // 'std::_Fill_n' was declared deprecated
@@ -744,7 +743,7 @@ bool PrgAtom::addConstraints(const LogicProgram& prg, ClauseCreator& gc) {
 			*j++ = *it;
 			nant = nant || it->isChoice();
 			if (!it->isDisj()) { gc.add(B); }
-			if (it->isNormal() && !ClauseCreator::create(*ctx.master(), ShortRep(literal(), ~B), 0)) { // FTA/BFA
+			if (it->isNormal() && !ctx.addBinary(literal(), ~B)) { // FTA/BFA
 				return false;
 			}
 		}
@@ -900,16 +899,18 @@ bool PrgBody::simplifyBody(LogicProgram& prg, bool strong, uint32* eqId) {
 	weight_t w         = 1, *jw = hasWeights() ? data_.ext[0]->weights : 0;
 	Literal* lits      = goals_begin();
 	Literal* j         = lits;
+	RuleState& todo    = prg.ruleState();
 	Var a;
-	bool mark, addEq;
+	bool mark, isEq;
+	int todos = 0;
 	for (Literal* it = j, *end = j + size(); it != end; ++it) {
 		a       = it->var();
-		addEq   = a != prg.getEqAtom(a);
+		isEq    = a != prg.getEqAtom(a);
 		oldHash+= hashId(it->sign()?-a:a);
-		if (addEq) {
+		if (isEq) {
 			prg.getAtom(a)->removeDep(id(), !it->sign()); // remove old edge
 			*it   = prg.getAtom(a)->eqGoal(it->sign());   // replace with eq goal
-			a     = it->var();                            // and remember to add the new edge later
+			a     = it->var();                            // and check it
 		}
 		Literal aLit = it->sign() ? ~prg.getAtom(a)->literal() : prg.getAtom(a)->literal();
 		ValueRep v   = prg.getAtom(a)->value();
@@ -928,8 +929,11 @@ bool PrgBody::simplifyBody(LogicProgram& prg, bool strong, uint32* eqId) {
 			prg.getAtom(a)->removeDep(id(), !it->sign());
 		}
 		else if (!mark || !ctx.marked(aLit)) {
-			if (mark) ctx.mark(aLit);
-			if (addEq)prg.getAtom(a)->addDep(id(), !it->sign()); // add new edge
+			if (mark) { ctx.mark(aLit); }
+			if (isEq) { // remember to add edge for new goal later
+				todo.addToBody(Literal(*it));
+				++todos;
+			}
 			*j++    = *it;       // copy literal
 			if (hasWeights()) {  // and weight
 				*jw++ = weight(it - lits);
@@ -951,10 +955,12 @@ bool PrgBody::simplifyBody(LogicProgram& prg, bool strong, uint32* eqId) {
 				uint32 pos        = findLit(prg, aLit);
 				data_.ext[0]->weights[pos] += w;
 			}
-			else { 
-				// ignore if normal
+			else { // ignore if normal
 				--bound;
-				if (!addEq) prg.getAtom(it->var())->removeDep(id(), !it->sign());
+				if (!isEq) { // remove edge
+					if (todo.inBody(*it)) { todo.clearBody(*it); --todos; }
+					else                  { prg.getAtom(it->var())->removeDep(id(), !it->sign()); }
+				}
 			} 
 		}
 	}
@@ -973,6 +979,12 @@ bool PrgBody::simplifyBody(LogicProgram& prg, bool strong, uint32* eqId) {
 			if (jw) { std::swap(jw[p], jw[n-1]); }
 			continue;
 		}
+		assert(a == lits[i].var());
+		if (todos && todo.inBody(lits[i])) {
+			prg.getAtom(a)->addDep(id(), !lits[i].sign());
+			todo.clearBody(lits[i]);
+			--todos;
+		}
 		Var v   = prg.getAtom(a)->var();
 		w       = !jw ? 1 : jw[i];
 		sumW   += w;
@@ -1092,34 +1104,43 @@ bool PrgBody::simplifyHeads(LogicProgram& prg, bool strong) {
 	return PrgBody::mergeHeads(prg, *this, strong);
 }
 
-bool PrgBody::mergeHeads(LogicProgram& prg, PrgBody& heads, bool strong) {
+bool PrgBody::mergeHeads(LogicProgram& prg, PrgBody& heads, bool strong, bool simplify) {
 	RuleState& rs = prg.ruleState();
 	bool       ok = true;
-	// mark the body literals so that we can easily detect superfluous atoms
-	// and selfblocking situations.	
-	for (const Literal* it = goals_begin(), *end = it + size(); it != end; ++it) {
-		rs.addToBody(*it);
+	assert((this == &heads || heads.sHead_ == 0) && "Heads to merge not simplified!");
+	if (simplify || &heads == this) {
+		// mark the body literals so that we can easily detect superfluous atoms
+		// and selfblocking situations.	
+		for (const Literal* it = goals_begin(), *end = it + size(); it != end; ++it) {
+			rs.addToBody(*it);
+		}
+		// remove duplicate/superfluous heads & check for blocked atoms
+		prepareSimplifyHeads(prg, rs);
+		if (this == &heads) {
+			ok = simplifyHeadsImpl(prg, *this, rs, strong);
+		}
+		else {
+			assert(heads.sHead_ == 0 && "Heads to merge not simplified!");
+			heads.prepareSimplifyHeads(prg, rs);
+			ok = simplifyHeadsImpl(prg, *this, rs, strong) && heads.simplifyHeadsImpl(prg, *this, rs, strong);
+			assert(ok || heads.heads_begin() == heads.heads_end());
+		}
+		// clear temporary flags & reestablish ordering
+		std::sort(const_cast<PrgEdge*>(heads_begin()), const_cast<PrgEdge*>(heads_end()));
+		for (head_iterator it = heads_begin(), end = heads_end(); it != end; ++it) {	
+			rs.clear(it->node());
+		}
+		for (const Literal* it = goals_begin(), *end = it + size(); it != end; ++it) {
+			rs.clear(it->var());
+		}	
+		sHead_ = 0;
 	}
-	// remove duplicate/superfluous heads & check for blocked atoms
-	prepareSimplifyHeads(prg, rs);
-	if (this == &heads) {
-		ok = simplifyHeadsImpl(prg, *this, rs, strong);
+	else if (relevant()) {
+		for (head_iterator it = heads.heads_begin(), end = heads.heads_end(); it != end; ++it) {
+			PrgHead* h = prg.getHead(*it);
+			if (h->relevant()) { addHead(h, it->type()); }
+		}
 	}
-	else {
-		assert(heads.sHead_ == 0 && "Heads to merge not simplified!");
-		heads.prepareSimplifyHeads(prg, rs);
-		ok = simplifyHeadsImpl(prg, *this, rs, strong) && heads.simplifyHeadsImpl(prg, *this, rs, strong);
-		assert(ok || heads.heads_begin() == heads.heads_end());
-	}
-	// clear temporary flags & reestablish ordering
-	std::sort(const_cast<PrgEdge*>(heads_begin()), const_cast<PrgEdge*>(heads_end()));
-	for (head_iterator it = heads_begin(), end = heads_end(); it != end; ++it) {	
-		rs.clear(it->node());
-	}
-	for (const Literal* it = goals_begin(), *end = it + size(); it != end; ++it) {
-		rs.clear(it->var());
-	}	
-	sHead_ = 0;
 	return ok || (assignValue(value_false) && propagateValue(prg, prg.options().backprop));
 }
 
@@ -1318,13 +1339,12 @@ bool PrgBody::addConstraints(const LogicProgram& prg, ClauseCreator& gc) {
 	if (type() == BodyInfo::NORMAL_BODY) {
 		bool    taut= false;
 		Literal negB= ~literal();
-		Solver& s   = *prg.ctx()->master();
 		gc.start().add(literal()); 
 		for (const Literal* it = goals_begin(), *end = goals_end(); it != end; ++it) {
 			assert(it->var() != 0);
 			Literal li = prg.getAtom(it->var())->literal() ^ it->sign();
 			if (li == literal()) { taut = true; continue; }
-			if (!ClauseCreator::create(s, ShortRep(negB, li), 0)) { // [~B li]
+			if (!prg.ctx()->addBinary(negB, li)) { // [~B li]
 				return false;
 			}
 			if (li.var() != negB.var()) { gc.add(~li); }  // [B v ~l1 v ... v ~ln]
@@ -1336,7 +1356,7 @@ bool PrgBody::addConstraints(const LogicProgram& prg, ClauseCreator& gc) {
 		Literal eq = prg.getAtom(goal(i).var())->literal() ^ goal(i).sign();
 		lits.push_back(WeightLiteral(eq, weight(i)));
 	}
-	return WeightConstraint::newWeightConstraint(*prg.ctx(), literal(), lits, bound());
+	return WeightConstraint::create(*prg.ctx()->master(), literal(), lits, bound()).ok();
 }
 
 // Returns the SCC of body B, i.e.
