@@ -53,8 +53,8 @@ public:
 	Constraint* cloneAttach(Solver& s) { return new RecordFinder(s, cloneMinimizer(s), queue); }
 	void doCommitModel(Enumerator& ctx, Solver& s);
 	void destroy(Solver* s, bool detach);
-	bool doUpdate(Solver& s, bool disjoint);
-	bool simplify(Solver& s, bool) { simplifyDB(s, db, false); return false; }
+	bool doUpdate(Solver& s);
+	bool simplify(Solver& s, bool) { EnumerationConstraint::simplify(s, false); simplifyDB(s, db, false); return false; }
 	QPtr          queue;
 	ThreadId      id;
 	ConstraintDB  db;
@@ -70,7 +70,7 @@ void ModelEnumerator::RecordFinder::destroy(Solver* s, bool detach) {
 	EnumerationConstraint::destroy(s, detach);
 }
 
-bool ModelEnumerator::RecordFinder::doUpdate(Solver& s, bool) {
+bool ModelEnumerator::RecordFinder::doUpdate(Solver& s) {
 	if (queue) {
 		uint32 f = ClauseCreator::clause_no_add | ClauseCreator::clause_no_release | ClauseCreator::clause_explicit;
 		for (SL* clause; queue->tryConsume(id, clause); ) {
@@ -96,7 +96,16 @@ void ModelEnumerator::RecordFinder::doCommitModel(Enumerator& x, Solver& s) {
 	solution.clear();
 	if (!ctx.projectionEnabled()) {
 		for (uint32 x = s.decisionLevel(); x != 0; --x) {
-			solution.push_back(~s.decision(x));
+			Literal d = s.decision(x);
+			if      (!s.auxVar(d.var()))  { solution.push_back(~d); }
+			else if (d != s.tagLiteral()) {
+				// Todo: set of vars could be reduced to those having the aux var in their reason set.
+				const LitVec& tr = s.trail();
+				const uint32  end= x != s.decisionLevel() ? s.levelStart(x+1) : (uint32)tr.size();
+				for (uint32 n = s.levelStart(x)+1; n != end; ++n) {
+					if (!s.auxVar(tr[n].var())) { solution.push_back(~tr[n]); }
+				}
+			}
 		}
 	}
 	else {
@@ -121,49 +130,53 @@ void ModelEnumerator::RecordFinder::doCommitModel(Enumerator& x, Solver& s) {
 /////////////////////////////////////////////////////////////////////////////////////////
 class ModelEnumerator::BacktrackFinder : public EnumerationConstraint {
 public:
-	typedef std::pair<Constraint*, uint32> NogoodPair;
-	typedef PodVector<NogoodPair>::type    Nogoods;
 	BacktrackFinder(Solver& s, MinimizeConstraint* min, uint32 projOpts) : EnumerationConstraint(s, min), front(UINT32_MAX), back(0), btLev(0), opts(projOpts) {}
 	bool hasModel() const { return front != UINT32_MAX; }
 	
 	// EnumerationConstraint interface
 	void doCommitModel(Enumerator& ctx, Solver& s);
-	bool doUpdate(Solver& s, bool disjoint);
+	bool doUpdate(Solver& s);
 	// Constraint interface
 	Constraint* cloneAttach(Solver& s){ return new BacktrackFinder(s, cloneMinimizer(s), opts); }
 	void        destroy(Solver*, bool);
-	void        undoLevel(Solver& s);
+	PropResult  propagate(Solver&, Literal, uint32&);
 	void        reason(Solver& s, Literal p, LitVec& x){
 		for (uint32 i = 1, end = s.level(p.var()); i <= end; ++i) { 
 			x.push_back(s.decision(i)); 
 		}
 	}
-	Nogoods nogoods;
-	LitVec  solution;
-	uint32  front;
-	uint32  back;
-	uint32  btLev;
-	uint32  opts;
+	typedef Solver::ConstraintDB ConstraintDB;
+	ConstraintDB nogoods;
+	LitVec       solution;
+	uint32       front;
+	uint32       back;
+	uint32       btLev;
+	uint32       opts;
 };
 
 void ModelEnumerator::BacktrackFinder::destroy(Solver* s, bool detach) {
 	while (!nogoods.empty()) {
-		Constraint* c = nogoods.back().first;
+		Constraint* c = nogoods.back();
 		nogoods.pop_back();
-		c->destroy(s, detach);
+		if (c) c->destroy(s, detach);
 	}
 	EnumerationConstraint::destroy(s, detach);
 }
 
-void ModelEnumerator::BacktrackFinder::undoLevel(Solver& s) {
-	while (!nogoods.empty() && nogoods.back().second >= s.decisionLevel()) {
-		Constraint* c = nogoods.back().first;
-		nogoods.pop_back();
+Constraint::PropResult ModelEnumerator::BacktrackFinder::propagate(Solver& s, Literal, uint32& pos) {
+	assert(pos < nogoods.size() && nogoods[pos] != 0);
+	ClauseHead* c = static_cast<ClauseHead*>(nogoods[pos]);
+	if (!c->locked(s)) {
 		c->destroy(&s, true);
+		nogoods[pos] = (c = 0);
+		while (!nogoods.empty() && !nogoods.back()) {
+			nogoods.pop_back();
+		}
 	}
+	return PropResult(true, c != 0);
 }
 
-bool ModelEnumerator::BacktrackFinder::doUpdate(Solver& s, bool disjoint) {
+bool ModelEnumerator::BacktrackFinder::doUpdate(Solver& s) {
 	if (hasModel()) {
 		bool ok = true;
 		assert(btLev <= s.decisionLevel());
@@ -182,23 +195,25 @@ bool ModelEnumerator::BacktrackFinder::doUpdate(Solver& s, bool disjoint) {
 			bool heu  = (opts & ModelEnumerator::project_use_heuristic) != 0;
 			Literal x = heu ? s.heuristic()->selectRange(s, &solution[0], &solution[0]+back) : solution[1];
 			ClauseRep rep = ClauseCreator::prepare(s, solution, ClauseCreator::clause_no_prepare, Constraint_t::learnt_conflict);
-			c = Clause::newContractedClause(s, rep, back, false);
+			c = Clause::newContractedClause(s, rep, back, true);
 			s.assume(~x);
 			// Remember that we must backtrack the current decision
 			// level in order to guarantee a different projected solution.
 			s.setBacktrackLevel(s.decisionLevel());
 			// Attach nogood to the current decision literal. 
-			// Once the solver removes that level, the nogood (which is then satisfied) is destroyed.
-			// Hence, the number of projection nogoods is linear in the number of (projection) atoms.
-			s.addUndoWatch(s.decisionLevel(), this);
-			nogoods.push_back( NogoodPair(c, s.decisionLevel()) );
+			// Once we backtrack to ~x, the then obsolete nogood is destroyed 
+			// keeping the number of projection nogoods linear in the number of (projection) atoms.
+			s.addWatch(x, this, (uint32)nogoods.size());
+			nogoods.push_back(c);
 			ok = true;
 		}
 		solution.clear();
 		front = UINT32_MAX;
 		return ok;
 	}
-	if (disjoint) return true;
+	if (optimize() || s.sharedContext()->concurrency() == 1 || disjointPath()) {
+		return true;
+	}
 	s.setStopConflict();
 	return false;
 }
@@ -217,7 +232,7 @@ void ModelEnumerator::BacktrackFinder::doCommitModel(Enumerator& ctx, Solver& s)
 		// all decision vars up to the backtrack level are projection vars.
 		// Find starting backtrack level.
 		btLev  = s.backtrackLevel();
-		for (uint32 r = btLev+1, end = s.decisionLevel(); r <= end && s.sharedContext()->project(s.decision(r).var()); ++r) {
+		for (uint32 r = btLev+1, end = s.decisionLevel(); r <= end && s.varInfo(s.decision(r).var()).project(); ++r) {
 			++btLev;
 		}
 		// Store the current projected assignment as a nogood.
@@ -288,11 +303,12 @@ EnumerationConstraint* ModelEnumerator::doInit(SharedContext& ctx, MinimizeConst
 	if (detectStrategy() || (ctx.concurrency() > 1 && !ModelEnumerator::supportsParallel())) {
 		st = 0;
 	}
-	bool trivial = optimize() || std::abs(numModels) == 1;
-	if (optimize() && project_) {
+	bool optOne  = minimizer() && minimizer()->mode() == MinimizeMode_t::optimize;
+	bool trivial = optOne || std::abs(numModels) == 1;
+	if (optOne && project_) {
 		const SharedMinimizeData* min = minimizer();
 		for (const WeightLiteral* it = min->lits; !isSentinel(it->first) && trivial; ++it) {
-			trivial = ctx.project(it->first.var());
+			trivial = ctx.varInfo(it->first.var()).project();
 		}
 		if (!trivial) { ctx.report(warning(Event::subsystem_prepare, "Projection: Optimization may depend on enumeration order.")); }
 	}
@@ -308,12 +324,12 @@ EnumerationConstraint* ModelEnumerator::doInit(SharedContext& ctx, MinimizeConst
 	EnumerationConstraint* c = st == strategy_backtrack 
 	  ? static_cast<ConPtr>(new BacktrackFinder(s, min, projectOpts()))
 	  : static_cast<ConPtr>(new RecordFinder(s, min, queue_));
-	if (projectionEnabled()) { c->setIgnoreSymmetric(true); }
+	if (projectionEnabled()) { setIgnoreSymmetric(true); }
 	return c;
 }
 
 void ModelEnumerator::initProjection(SharedContext& ctx) {
-	const SymbolTable& index = ctx.symTab();
+	const SymbolTable& index = ctx.symbolTable();
 	if (index.type() == SymbolTable::map_indirect) {
 		for (SymbolTable::const_iterator it = index.curBegin(); it != index.end(); ++it) {
 			if (!it->second.name.empty() && it->second.name[0] != '_') {
