@@ -81,7 +81,6 @@ void ShortImplicationsGraph::ImplicationList::clear(bool b) {
 	}
 	learnt = 0;
 }
-
 void ShortImplicationsGraph::ImplicationList::simplifyLearnt(const Solver& s) {
 	Block* x = learnt;
 	learnt   = 0;
@@ -97,7 +96,6 @@ void ShortImplicationsGraph::ImplicationList::simplifyLearnt(const Solver& s) {
 		delete t;
 	}
 }
-
 void ShortImplicationsGraph::ImplicationList::addLearnt(Literal p, Literal q) {
 	Literal nc[2] = {p, q};
 	uint32  ns    = 1 + !isSentinel(q);
@@ -177,46 +175,36 @@ void ShortImplicationsGraph::resize(uint32 nodes) {
 
 uint32 ShortImplicationsGraph::numEdges(Literal p) const { return graph_[p.index()].size(); }
 
-bool ShortImplicationsGraph::addBinary(Literal p, Literal q, bool learnt) {
-	uint32 added = 0;
+bool ShortImplicationsGraph::add(ImpType t, bool learnt, const Literal* lits) {
+	uint32& stats= (t == ternary_imp ? tern_ : bin_)[learnt];
+	Literal p = lits[0], q = lits[1], r = (t == ternary_imp ? lits[2] : negLit(0));
 	if (!shared_) {
-		if (!learnt) { p.clearWatch(); q.clearWatch(); }
-		else         { p.watch(); q.watch(); }
-		getList(~p).push_left(q);
-		getList(~q).push_left(p);
-		added = 1;
-	}
-#if WITH_THREADS
-	else if (learnt && !getList(~p).hasLearnt(q)) {
-		getList(~p).addLearnt(q);
-		getList(~q).addLearnt(p);
-		added = 1;
-	}
-#endif
-	bin_[learnt] += added;
-	return added > 0;
-}
-
-bool ShortImplicationsGraph::addTernary(Literal p, Literal q, Literal r, bool learnt) {
-	uint32 added = 0;
-	if (!shared_) {
-		if (!learnt) { p.clearWatch(); q.clearWatch(); r.clearWatch(); }
-		else         { p.watch(); q.watch(); r.watch(); }
-		getList(~p).push_right(std::make_pair(q, r));
-		getList(~q).push_right(std::make_pair(p, r));
-		getList(~r).push_right(std::make_pair(p, q));
-		added = 1;
+		if (!learnt){ p.clearWatch(), q.clearWatch(), r.clearWatch(); }
+		else        { p.watch(), q.watch(), r.watch(); }
+		if (t == binary_imp) {
+			getList(~p).push_left(q);
+			getList(~q).push_left(p);
+		}
+		else {
+			getList(~p).push_right(std::make_pair(q, r));
+			getList(~q).push_right(std::make_pair(p, r));
+			getList(~r).push_right(std::make_pair(p, q));
+		}
+		++stats;
+		return true;
 	}
 #if WITH_THREADS
 	else if (learnt && !getList(~p).hasLearnt(q, r)) {
 		getList(~p).addLearnt(q, r);
 		getList(~q).addLearnt(p, r);
-		getList(~r).addLearnt(p, q);
-		added = 1;
+		if (t == ternary_imp) {
+			getList(~r).addLearnt(p, q);
+		}
+		++stats;
+		return true;
 	}
 #endif
-	tern_[learnt] += added;
-	return added > 0;
+	return false;
 }
 
 void ShortImplicationsGraph::remove_bin(ImplicationList& w, Literal p) {
@@ -228,6 +216,15 @@ void ShortImplicationsGraph::remove_tern(ImplicationList& w, Literal p) {
 	w.try_shrink();	
 }
 
+// Removes all binary clauses containing p - those are now SAT.
+// Binary clauses containing ~p are unit and therefore likewise SAT. Those
+// are removed when their second literal is processed.
+//
+// Ternary clauses containing p are SAT and therefore removed.
+// Ternary clauses containing ~p are now either binary or SAT. Those that
+// are SAT are removed when the satisfied literal is processed.
+// All conditional binary-clauses are replaced with real binary clauses.
+// Note: clauses containing p watch ~p. Those containing ~p watch p.
 void ShortImplicationsGraph::removeTrue(const Solver& s, Literal p) {
 	assert(!shared_);
 	typedef ImplicationList SWL;
@@ -263,7 +260,8 @@ void ShortImplicationsGraph::removeTrue(const Solver& s, Literal p) {
 		remove_tern(graph_[(~r).index()], ~p);
 		if (s.value(q.var()) == value_free && s.value(r.var()) == value_free) {
 			// clause is binary on dl 0
-			addBinary(q, r, false);
+			Literal imp[2] = {q,r};
+			add(binary_imp, false, imp);
 		}
 		// else: clause is SAT and removed when the satisfied literal is processed
 	}
@@ -326,8 +324,10 @@ void SatPreprocessor::discardClauses(bool full) {
 		} while (r);
 		elimTop_ = 0;
 	}
+	if (full) { seen_ = Range32(1,1); }
 }
 void SatPreprocessor::cleanUp(bool full) {
+	if (ctx_) { seen_.hi = ctx_->numVars() + 1; }
 	doCleanUp();
 	discardClauses(full);
 }
@@ -344,6 +344,16 @@ bool SatPreprocessor::addClause(const Literal* lits, uint32 size) {
 	}
 	return true;
 }	
+
+void SatPreprocessor::freezeSeen() {
+	if (!ctx_->validVar(seen_.lo)) { seen_.lo = 1; }
+	if (!ctx_->validVar(seen_.hi)) { seen_.hi = ctx_->numVars() + 1; }
+	for (Var v = seen_.lo; v != seen_.hi; ++v) {
+		if (!ctx_->eliminated(v)) { ctx_->setFrozen(v, true); }
+	}
+	seen_.lo = seen_.hi;
+}
+
 bool SatPreprocessor::preprocess(SharedContext& ctx, Options& opts) {
 	ctx_ = &ctx;
 	Solver* s = ctx_->master();
@@ -359,9 +369,8 @@ bool SatPreprocessor::preprocess(SharedContext& ctx, Options& opts) {
 			if (rest) ctx->satPrepro.reset(rest);
 		}
 	} onExit(this, &ctx);
-	ConstraintType unit = Constraint_t::static_constraint;
 	for (LitVec::const_iterator it = units_.begin(), end = units_.end(); it != end; ++it) {
-		if (!s->addUnary(*it, unit)) return false;
+		if (!ctx.addUnary(*it)) return false;
 	}
 	units_.clear();
 	// skip preprocessing if other constraints are UNSAT
@@ -375,7 +384,7 @@ bool SatPreprocessor::preprocess(SharedContext& ctx, Options& opts) {
 	if (opts.limFrozen != 0 && ctx_->stats().vars_frozen) {
 		uint32 varFrozen = ctx_->stats().vars_frozen;
 		for (LitVec::const_iterator it = s->trail().begin(), end = s->trail().end(); it != end; ++it) {
- 			varFrozen -= (ctx_->frozen(it->var()));
+ 			varFrozen -= (ctx_->varInfo(it->var()).frozen());
  		}
 		limFrozen = ((varFrozen / double(s->numFreeVars())) * 100.0) > double(opts.limFrozen);
 	}
@@ -386,17 +395,18 @@ bool SatPreprocessor::preprocess(SharedContext& ctx, Options& opts) {
 			Clause* c   = clauses_[i]; assert(c);
 			clauses_[i] = 0;
 			c->simplify(*s);
-			if (s->value((*c)[0].var()) == value_free) {
+			Literal x   = (*c)[0];
+			if (s->value(x.var()) == value_free) {
 				clauses_[j++] = c; 
 			}
-			else if (c) {
-				Literal x = (*c)[0];
-				c->destroy(); 
-				if (!s->addUnary(x, unit)) { return false; }
+			else {
+				c->destroy();
+				if (!ctx.addUnary(x)) { return false; }	
 			}
 		}
 		clauses_.erase(clauses_.begin()+j, clauses_.end());
 		// 2. run preprocessing
+		freezeSeen();
 		if (!s->propagate() || !doPreprocess()) {
 			return false;
 		}
@@ -405,13 +415,12 @@ bool SatPreprocessor::preprocess(SharedContext& ctx, Options& opts) {
 	if (!s->simplify()) return false;
 	// 3. move preprocessed clauses to ctx
 	for (ClauseList::size_type i = 0; i != clauses_.size(); ++i) {
-		if (clauses_[i]) {
-			Clause& c = *clauses_[i];
-			ClauseRep rep = { &c[0], c.size(), ClauseInfo() };
-			if (!ClauseCreator::create(*s, rep, 0)) {
+		if (Clause* c = clauses_[i]) {
+			if (!ClauseCreator::create(*s, ClauseRep::create(&(*c)[0], c->size()), 0)) {
 				return false;
 			}
-			c.destroy();
+			clauses_[i] = 0;
+			c->destroy();
 		}
 	}
 	ClauseList().swap(clauses_);
@@ -497,8 +506,8 @@ void SharedContext::cloneVars(const SharedContext& other, InitMode m) {
 	problem_.vars_eliminated = other.problem_.vars_eliminated;
 	problem_.vars_frozen     = other.problem_.vars_frozen;
 	varInfo_                 = other.varInfo_;
-	if (&symTab() != &other.symTab()) {
-		if (m == init_copy_symbols) { other.symTab().copyTo(symTab()); }
+	if (&symbolTable() != &other.symbolTable()) {
+		if (m == init_copy_symbols) { other.symbolTable().copyTo(symbolTable()); }
 		else { 
 			++other.symTabPtr_->refs;
 			if (--symTabPtr_->refs == 0) { delete symTabPtr_; }
@@ -518,7 +527,7 @@ void SharedContext::reset() {
 	new (this) SharedContext();	
 }
 
-void SharedContext::concurrency(uint32 n) {
+void SharedContext::setConcurrency(uint32 n) {
 	if (n <= 1) { share_.count = 1; share_.shareM = ContextParams::share_auto; }
 	else        { share_.count = n; solvers_.reserve(n); }
 	while (solvers_.size() > share_.count) {
@@ -573,11 +582,6 @@ void SharedContext::setConfiguration(Configuration* c, bool own) {
 	}
 }
 
-
-void SharedContext::reserveVars(uint32 varGuess) {
-	varInfo_.reserve(varGuess + 1);
-}
-
 bool SharedContext::unfreeze() {
 	if (frozen()) {
 		share_.frozen = 0;
@@ -585,31 +589,36 @@ bool SharedContext::unfreeze() {
 		btig_.markShared(false);
 		return master()->popRootLevel(master()->rootLevel())
 		  &&   btig_.propagate(*master(), posLit(0)) // any newly learnt facts
-		  &&   unfreezeStep(*master());
+		  &&   unfreezeStep();
 	}
 	return true;
 }
 
-bool SharedContext::unfreezeStep(Solver& s) {
-	assert(!s.decisionLevel());
-	if (s.value(step_.var()) == value_free){ s.force(~step_, posLit(0));}
-	if (s.value(tag_.var()) == value_free) { s.force(~tag_, posLit(0)); }
-	uint32 rem = !isSentinel(step_) + !isSentinel(tag_);
-	if (rem && s.simplify()) {
-		LitVec::size_type i = s.assign_.trail.size();
-		while (i--) {
-			Var v = s.assign_.trail[i].var();
-			if (v == step_.var() || v == tag_.var()) {
-				s.assign_.trail[i]     = s.assign_.trail.back();
-				s.assign_.trail.back() = negLit(v);
-				s.assign_.undoLast();
-				if (--rem == 0) { break; }
+bool SharedContext::unfreezeStep() {
+	for (uint32 i = 0; i != solvers_.size(); ++i) {
+		Solver& s = *solvers_[i];
+		s.popRootLevel(s.rootLevel());
+		s.popAuxVar();
+		if (s.value(step_.var()) == value_free){ s.force(~step_, posLit(0));}
+		if (!isSentinel(step_) && s.simplify()) {
+			LitVec::size_type i = s.assign_.trail.size();
+			while (i--) {
+				Var v = s.assign_.trail[i].var();
+				if (v == step_.var()) {
+					s.assign_.trail[i]     = s.assign_.trail.back();
+					s.assign_.trail.back() = negLit(v);
+					s.assign_.undoLast();
+					break;
+				}
 			}
+			s.assign_.qReset();
+			s.assign_.setUnits(s.lastSimp_ = (uint32)s.assign_.trail.size()); 
+		}
+		if (s.configuration().heuReinit) {
+			s.heuristic_.reset(0);
 		}
-		s.assign_.qReset();
-		s.assign_.setUnits(s.lastSimp_ = (uint32)s.assign_.trail.size()); 
 	}
-	return !s.hasConflict();
+	return !master()->hasConflict();
 }
 
 Var SharedContext::addVar(VarType t, bool eq) {
@@ -621,9 +630,8 @@ Var SharedContext::addVar(VarType t, bool eq) {
 	return numVars();
 }
 
-void SharedContext::requestTagLiteral()  { if (tag_ == posLit(0))  { tag_ = negLit(0); } }
-void SharedContext::requestStepLiteral() { if (step_ == posLit(0)) { step_= negLit(0); } }
-void SharedContext::requestData(Var v)   { master()->requestData(v); }
+void SharedContext::requestStepVar()  { if (step_ == posLit(0)) { step_= negLit(0); } }
+void SharedContext::requestData(Var v){ master()->requestData(v); }
 
 void SharedContext::setFrozen(Var v, bool b) {
 	assert(validVar(v)); 
@@ -646,26 +654,46 @@ void SharedContext::eliminate(Var v) {
 	}
 }
 
+Literal SharedContext::addAuxLit() {
+	VarInfo nv; nv.set(VarInfo::FROZEN);
+	varInfo_.push_back(nv);
+	return posLit(numVars());
+}
 Solver& SharedContext::startAddConstraints(uint32 constraintGuess) {
-	if (!unfreeze()) { return *master(); }
-	if (master()->isFalse(tag_)) {
-		// add aux var for tag literal
-		tag_ = posLit(addVar(Var_t::atom_var));
-		setFrozen(tag_.var(), true);
-		--problem_.vars;
-	}
-	if (master()->isFalse(step_)) {
-		step_ = posLit(addVar(Var_t::atom_var));
-		setFrozen(step_.var(), true);
-		--problem_.vars;
-	}
+	if (!unfreeze())              { return *master(); }
+	if (master()->isFalse(step_)) { step_= addAuxLit(); }
 	btig_.resize((numVars()+1)<<1);
 	master()->startInit(constraintGuess);
 	return *master();
 }
+bool SharedContext::addUnary(Literal x) {
+	CLASP_ASSERT_CONTRACT(!frozen() || !isShared());
+	return master()->force(x);
+}
+bool SharedContext::addBinary(Literal x, Literal y) {
+	CLASP_ASSERT_CONTRACT(allowImplicit(Constraint_t::static_constraint));
+	Literal lits[2] = {x, y};
+	return ClauseCreator::create(*master(), ClauseRep(lits, 2), ClauseCreator::clause_force_simplify);
+}
+bool SharedContext::addTernary(Literal x, Literal y, Literal z) {
+	CLASP_ASSERT_CONTRACT(allowImplicit(Constraint_t::static_constraint));
+	Literal lits[3] = {x, y, z};
+	return ClauseCreator::create(*master(), ClauseRep(lits, 3), ClauseCreator::clause_force_simplify);
+}
+void SharedContext::add(Constraint* c) {
+	CLASP_ASSERT_CONTRACT(!frozen());
+	master()->add(c);
+}
+int SharedContext::addImp(ImpGraph::ImpType t, const Literal* lits, ConstraintType ct) {
+	if (!allowImplicit(ct)) { return -1; }
+	bool learnt = ct != Constraint_t::static_constraint;
+	if (!learnt && !frozen() && satPrepro.get()) {
+		satPrepro->addClause(lits, static_cast<uint32>(t));
+		return 1;
+	}
+	return int(btig_.add(t, learnt, lits));
+}
 
-bool SharedContext::addUnary(Literal p)      { return master()->addUnary(p, Constraint_t::static_constraint); }
-void SharedContext::add(Constraint* c)       { return master()->add(c); }
 uint32 SharedContext::numConstraints() const { return numBinary() + numTernary() + master()->constraints_.size(); }
 
 bool SharedContext::endInit(bool attachAll) {
@@ -677,7 +705,7 @@ bool SharedContext::endInit(bool attachAll) {
 	satPrepro.reset(temp.release());
 	btig_.markShared(concurrency() > 1);
 	master()->dbIdx_            = (uint32)master()->constraints_.size();
-	lastTopLevel_               = master()->assign_.front;
+	lastTopLevel_               = (uint32)master()->assign_.front;
 	problem_.constraints        = master()->constraints_.size();
 	problem_.constraints_binary = btig_.numBinary();
 	problem_.constraints_ternary= btig_.numTernary();
@@ -692,10 +720,9 @@ bool SharedContext::endInit(bool attachAll) {
 
 bool SharedContext::attach(Solver& other) {
 	assert(frozen() && other.shared_ == this);
-	if (other.validVar(tag_.var()) && other.validVar(step_.var())) {
+	if (other.validVar(step_.var())) {
 		if (!other.popRootLevel(other.rootLevel())){ return false; }
 		if (&other == master())                    { return true;  }
-		if (!unfreezeStep(other))                  { return false; }
 	}
 	initStats(other);
 	// 1. clone vars & assignment
@@ -725,8 +752,8 @@ bool SharedContext::attach(Solver& other) {
 void SharedContext::detach(Solver& s, bool reset) {
 	assert(s.shared_ == this);
 	if (reset) { s.reset(); }
-	s.removeConditional();
 	s.setEnumerationConstraint(0);
+	s.popAuxVar();
 }
 void SharedContext::initStats(Solver& s) const {
 	s.stats.enableStats(master()->stats);
@@ -745,7 +772,7 @@ void SharedContext::accuStats() {
 	if (sccGraph.get()) { sccGraph->accuStats(); }
 }
 
-void SharedContext::simplifyDB(bool shuffle) {
+void SharedContext::simplify(bool shuffle) {
 	Solver::ConstraintDB& db = master()->constraints_;
 	if (concurrency() == 1 || master()->dbIdx_ == 0) {
 		Clasp::simplifyDB(*master(), db, shuffle);
@@ -767,6 +794,9 @@ void SharedContext::simplifyDB(bool shuffle) {
 	}
 	master()->dbIdx_ = db.size();
 }
+void SharedContext::simplifyShort(const Solver& s, Literal p) {
+	if (!isShared() && p.index() < btig_.size()) { btig_.removeTrue(s, p); }
+}
 
 uint32 SharedContext::problemComplexity() const {
 	if (isExtended()) {
@@ -778,24 +808,6 @@ uint32 SharedContext::problemComplexity() const {
 	}
 	return numConstraints();
 }
-
-SharedLiterals* SharedContext::distribute(const Solver& s, const Literal* lits, uint32 size, const ClauseInfo& extra) const {
-	if (distributor.get() && !extra.tagged() && (size <= 3 || distributor->isCandidate(size, extra.lbd(), extra.type()))) {
-		uint32 initialRefs = share_.count - (size <= Clause::MAX_SHORT_LEN || !physicalShare(extra.type()));
-		SharedLiterals* x  = SharedLiterals::newShareable(lits, size, extra.type(), initialRefs);
-		distributor->publish(s, x);
-		const_cast<Solver&>(s).stats.addDistributed(extra.lbd(), extra.type());
-		return initialRefs == share_.count ? x : 0;
-	}
-	return 0;
-}
-
-uint32 SharedContext::receive(const Solver& target, SharedLiterals** out, uint32 maxOut) const {
-	if (distributor.get()) {
-		return distributor->receive(target, out, maxOut);
-	}
-	return 0;
-}
 /////////////////////////////////////////////////////////////////////////////////////////
 // Distributor
 /////////////////////////////////////////////////////////////////////////////////////////
