@@ -23,7 +23,6 @@
 #ifdef _MSC_VER
 #pragma warning (disable : 4996) // 'std::copy': Function call with parameters that may be unsafe
 #endif
-
 namespace Clasp { 
 
 DecisionHeuristic::~DecisionHeuristic() {}
@@ -103,9 +102,9 @@ void Solver::PPList::cancel()           const { for (PostPropagator* r = active(
 bool Solver::PPList::isModel(Solver& s) const {
 	if (s.hasConflict()) { return false; }
 	for (PostPropagator* r = active(); r; r = r->next) {
-		if (!r->isModel(s)) return false;
+		if (!r->isModel(s)){ return false; }
 	}
-	return true;
+	return !s.enumerationConstraint() || s.enumerationConstraint()->valid(s);
 }
 /////////////////////////////////////////////////////////////////////////////////////////
 // Solver: Construction/Destruction/Setup
@@ -139,6 +138,7 @@ void Solver::freeMem() {
 	constraints_.clear();
 	learnts_.clear();
 	setEnumerationConstraint(0);
+	heuristic_.reset(0);
 	PodVector<WatchList>::destruct(watches_);
 	// free undo lists
 	// first those still in use
@@ -160,6 +160,7 @@ void Solver::freeMem() {
 
 SatPreprocessor*    Solver::satPrepro()     const { return shared_->satPrepro.get(); }
 const SolverParams& Solver::configuration() const { return shared_->configuration()->solver(id()); }
+const SolveParams&  Solver::searchConfig()  const { return shared_->configuration()->search(id()); }
 
 void Solver::reset() {
 	SharedContext* myCtx = shared_;
@@ -187,19 +188,26 @@ void Solver::startInit(uint32 numConsGuess) {
 	const SolverParams& params = configuration();
 	if (strategy.loadCfg == 1) {
 		uint32 id       = this->id();
+		uint32 hId      = strategy.heuReserved;
 		strategy        = params;
 		strategy.id     = id; // keep id
 		strategy.loadCfg= 0;  // strategy is now "up to date"
 		if      (!params.ccMinRec)  { delete ccMin_; ccMin_ = 0; }
 		else if (!ccMin_)           { ccMin_ = new CCMinRecursive; }
-		if (!shared_->seedSolvers()){ rng.srand(params.seed); }
-		else                        {
+		if (id == params.id || !shared_->seedSolvers()) {
+			rng.srand(params.seed);
+		}
+		else {
 			RNG x(14182940); for (uint32 i = 0; i != id; ++i) { x.rand(); }
 			rng.srand(x.seed());
 		}
+		if (hId != params.heuId) { // heuristic has changed
+			heuristic_.reset(0);
+		}
 	}
-	if (params.heuReinit || heuristic_.get() == 0) {
+	if (heuristic_.get() == 0) {
 		heuristic_.reset(shared_->configuration()->heuristic(id()));
+		strategy.heuReserved = params.heuId;
 	}
 	if (!popRootLevel(rootLevel())) { return; }
 	post_.disable(); // disable post propagators during setup
@@ -247,78 +255,159 @@ bool Solver::endInit() {
 	return propagate() && simplify();
 }
 
+void Solver::add(Constraint* c) {
+	constraints_.push_back(c);
+}
+bool Solver::add(const ClauseRep& c, bool isNew) {
+	typedef ShortImplicationsGraph::ImpType ImpType;
+	if (c.prep == 0) {
+		return ClauseCreator::create(*this, c, ClauseCreator::clause_force_simplify).ok();
+	}
+	int added = 0;
+	if (c.size > 1) {
+		if (allowImplicit(c)) { added = shared_->addImp(static_cast<ImpType>(c.size), c.lits, c.info.type()); }
+		else                  { return ClauseCreator::create(*this, c, ClauseCreator::clause_explicit).ok(); }
+	}
+	else {
+		Literal u = c.size ? c.lits[0] : negLit(0);
+		uint32  ts= trail().size();
+		force(u);
+		added     = int(ts != trail().size());
+	}
+	if (added > 0 && isNew && c.info.learnt()) {
+		stats.addLearnt(c.size, c.info.type());
+		distribute(c.lits, c.size, c.info);
+	}
+	return !hasConflict();
+}
+uint32 Solver::receive(SharedLiterals** out, uint32 maxOut) const {
+	if (shared_->distributor.get()) {
+		return shared_->distributor->receive(*this, out, maxOut);
+	}
+	return 0;
+}
+SharedLiterals* Solver::distribute(const Literal* lits, uint32 size, const ClauseInfo& extra) {
+	if (shared_->distributor.get() && !extra.aux() && (size <= 3 || shared_->distributor->isCandidate(size, extra.lbd(), extra.type()))) {
+		uint32 initialRefs = shared_->concurrency() - (size <= Clause::MAX_SHORT_LEN || !shared_->physicalShare(extra.type()));
+		SharedLiterals* x  = SharedLiterals::newShareable(lits, size, extra.type(), initialRefs);
+		shared_->distributor->publish(*this, x);
+		stats.addDistributed(extra.lbd(), extra.type());
+		return initialRefs == shared_->concurrency() ? x : 0;
+	}
+	return 0;
+}
 void Solver::setEnumerationConstraint(Constraint* c) {
 	if (enum_) enum_->destroy(this, true);
 	enum_ = c;
 }
 
-void Solver::add(Constraint* c) {
-	assert(!shared_->frozen() && this == shared_->master());
-	constraints_.push_back(c);
-}
-
-bool Solver::addUnary(Literal p, ConstraintType t) {
-	assert(t != Constraint_t::static_constraint || (!shared_->frozen() && this == shared_->master()));
-	if (t != Constraint_t::static_constraint) {
-		bool subsumed = (isTrue(p) && level(p.var()) == 0) || (shared_->isShared() && shared_->allowImplicit(t) && !shared_->addBinary(p, negLit(0), true));
-		if (!subsumed) {
-			stats.addLearnt(1, t);
-			shared_->distribute(*this, &p, 1, ClauseInfo(t).setLbd(1));
-		}
-	}
-	return force(p, 0, Antecedent(posLit(0)));
-}
-
-bool Solver::addShort(const ClauseRep& rep) {
-	bool learnt  = rep.info.learnt();
-	if (!rep.isImp() || (learnt && !shared_->allowImplicit(rep.info.type()))) { 
-		return false;
-	}
-	const Literal* c = rep.lits;
-	learnt = (rep.size == 3 ? shared_->addTernary(c[0], c[1], c[2], learnt) : shared_->addBinary(c[0], c[1], learnt)) && learnt;
-	if (learnt) {
-		stats.addLearnt(rep.size, rep.info.type());
-		shared_->distribute(*this, rep.lits, rep.size, rep.info);
-	}
-	return true;
-}
-
 uint32 Solver::numConstraints() const {
 	return static_cast<uint32>(constraints_.size())
 		+ (shared_ ? shared_->numBinary()+shared_->numTernary() : 0);
 }
 
+Var Solver::pushAuxVar() {
+	Var aux = assign_.addVar();
+	setPref(aux, ValueSet::def_value, value_false);
+	watches_.insert(watches_.end(), 2, WatchList()); 
+	if (heuristic_.get()) { heuristic_->updateVar(*this, aux, 1); }
+	return aux;
+}
+void Solver::popAuxVar(uint32 num) {
+	num = numVars() >= shared_->numVars() ? std::min(numVars() - shared_->numVars(), num) : 0;
+	if (!num) { return; }
+	// 1. find first dl containing one of the aux vars
+	Literal aux = posLit(assign_.numVars() - num);
+	uint32  dl  = decisionLevel() + 1;
+	for (ImpliedList::iterator it = impliedLits_.begin(); it != impliedLits_.end(); ++it) {
+		if (!(it->lit < aux)) { dl = std::min(dl, it->level); }
+	}
+	for (Var v = aux.var(), end = aux.var()+num; v != end; ++v) {
+		if (value(v) != value_free){ dl = std::min(dl, level(v)); }
+	}
+	// 2. remove aux vars from assignment
+	if (dl > rootLevel()) {
+		if (backtrackLevel() >= dl) { setBacktrackLevel(dl-1); }
+		undoUntil(dl-1, true);
+	}
+	else {
+		popRootLevel((rootLevel() - dl) + 1);
+		if (dl == 0) { // top-level has aux vars - cleanup manually
+			uint32 j = shared_->numUnary(), units = assign_.units();
+			for (uint32 i = j, end = assign_.trail.size(); i != end; ++i) {
+				if (assign_.trail[i] < aux) { assign_.trail[j++] = assign_.trail[i]; }
+				else                        {
+					units         -= (i < units);
+					assign_.front -= (i < assign_.front);
+					lastSimp_     -= (i < lastSimp_);
+				}
+			}
+			shrinkVecTo(assign_.trail, j);
+			assign_.setUnits(units);
+		}
+	}
+	// 3. remove constraints over aux
+	ConstraintDB::size_type i, j, end = learnts_.size();
+	LitVec cc;
+	for (i = j = 0; i != end; ++i) {
+		cc.clear();
+		if (ClauseHead* c = static_cast<LearntConstraint*>(learnts_[i])->clause()) { c->toLits(cc); }
+		if (std::find_if(cc.begin(), cc.end(), std::not1(std::bind2nd(std::less<Literal>(), aux))) == cc.end()) {
+			learnts_[j++] = learnts_[i];
+		}
+		else { learnts_[i]->destroy(this, true); }
+	}
+	learnts_.erase(learnts_.begin()+j, learnts_.end());
+	// 4. remove aux vars and their watches
+	assign_.resize(assign_.numVars()-num);
+	for (uint32 n = num; n--;) { 
+		watches_.back().clear(true);
+		watches_.pop_back();
+		watches_.back().clear(true);
+		watches_.pop_back();
+	}
+	if (!validVar(tag_.var())) { tag_ = posLit(0); }
+	heuristic_->updateVar(*this, aux.var(), num);
+}
+
 bool Solver::pushRoot(const LitVec& path) {
 	// make sure we are on the current root level
-	undoUntil(0, true);
-	if (hasConflict() || !simplify() || !propagate() || !pushRoot(shared_->stepLiteral())) { return false; }
+	if (!popRootLevel(0) || !simplify() || !propagate()) { return false; }
+	// push path
 	stats.addPath(path.size());
 	for (LitVec::const_iterator it = path.begin(), end = path.end(); it != end; ++it) {
 		if (!pushRoot(*it)) { return false; }
 	}
 	ccInfo_.setActivity(1);
-	return pushRoot(shared_->tagLiteral());
+	return true;
 }
 
 bool Solver::pushRoot(Literal x) {
-	assert(queueSize() == 0);
-	if (value(x.var()) == value_free) {
-		assume(x); --stats.choices;
-		pushRootLevel();
-		return propagate();
-	}
-	return isTrue(x);
+	if (hasConflict())                 { return false; }
+	if (decisionLevel()!= rootLevel()) { popRootLevel();  }
+	if (queueSize() && !propagate())   { return false;    }
+	if (value(x.var()) != value_free)  { return isTrue(x);}
+	assume(x); --stats.choices;
+	pushRootLevel();
+	return propagate();
 }
 
-bool Solver::popRootLevel(uint32 i)  {
+bool Solver::popRootLevel(uint32 i, LitVec* popped, bool aux)  {
 	clearStopConflict();
-	levels_.root      -= std::min(i, rootLevel());
+	uint32 newRoot = levels_.root - std::min(i, rootLevel());
+	if (popped && newRoot < rootLevel()) {
+		for (uint32 i = newRoot+1; i <= rootLevel(); ++i) {
+			Literal x = decision(i);
+			if (aux || !auxVar(x.var())) { popped->push_back(x); }
+		}
+	}
+	levels_.root       = newRoot;
 	levels_.backtrack  = rootLevel();
 	impliedLits_.front = 0;
-	bool tagActive     = isTrue(sharedContext()->tagLiteral());
+	bool tagActive     = isTrue(tagLiteral());
 	// go back to new root level and re-assert still implied literals
 	undoUntil(rootLevel(), true);
-	if (tagActive && !isTrue(sharedContext()->tagLiteral())) {
+	if (tagActive && !isTrue(tagLiteral())) {
 		removeConditional();
 	}
 	return !hasConflict();
@@ -353,27 +442,41 @@ void Solver::setStopConflict() {
 	pushRootLevel(decisionLevel());
 }
 
-void Solver::updateGuidingPath(LitVec& gpOut, LitVec::size_type& start, uint32& implied) {
-	if (start < shared_->topLevelSize()) {
-		start = shared_->topLevelSize();
+void Solver::copyGuidingPath(LitVec& gpOut) {
+	uint32 aux = rootLevel()+1;
+	gpOut.clear();
+	for (uint32 i = 1, end = rootLevel()+1; i != end; ++i) {
+		Literal x = decision(i);
+		if      (!auxVar(x.var())) { gpOut.push_back(x); }
+		else if (i < aux)          { aux = i; }
 	}
-	const LitVec& tr = assign_.trail;
-	LitVec::size_type end = rootLevel() == decisionLevel() ? tr.size() : levels_[rootLevel()].trailPos;
-	for (LitVec::size_type i = start; i < end; ++i) {
-		if (reason(tr[i]).isNull()) {
-			gpOut.push_back(tr[i]);
-		}
-	}
-	start = end;
-	uint32 implLits = 0;
 	for (ImpliedList::iterator it = impliedLits_.begin(); it != impliedLits_.end(); ++it) {
-		if (it->level <= rootLevel() && ++implLits > implied) {
+		if (it->level <= rootLevel() && (it->ante.ante().isNull() || it->level < aux) && !auxVar(it->lit.var())) {
 			gpOut.push_back(it->lit);
 		}
 	}
-	implied = implLits;
 }
-
+bool Solver::splittable() const {
+	if (decisionLevel() == rootLevel() || frozenLevel(rootLevel()+1)) { return false; }
+	if (numAuxVars()) { // check if gp would contain solver local aux var
+		uint32 minAux = rootLevel() + 2;
+		for (uint32 i = 1; i != minAux; ++i) { 
+			if (auxVar(decision(i).var()) && decision(i) != tag_) { return false; } 
+		}
+		for (ImpliedList::iterator it = impliedLits_.begin(); it != impliedLits_.end(); ++it) {
+			if (it->ante.ante().isNull() && it->level < minAux && auxVar(it->lit.var()) && it->lit != tag_) { return false; }
+		}
+	}
+	return true;
+}
+bool Solver::split(LitVec& out) {
+	if (!splittable()) { return false; }
+	copyGuidingPath(out);
+	pushRootLevel();
+	out.push_back(~decision(rootLevel()));
+	stats.addSplit();
+	return true;
+}
 /////////////////////////////////////////////////////////////////////////////////////////
 // Solver: Watch management
 ////////////////////////////////////////////////////////////////////////////////////////
@@ -381,7 +484,7 @@ uint32 Solver::numWatches(Literal p) const {
 	assert( validVar(p.var()) );
 	if (!validWatch(p)) return 0;
 	return static_cast<uint32>(watches_[p.index()].size()) 
-		+ shared_->numShortImplications(p);
+		+ shared_->shortImplications().numEdges(p);
 }
 	
 bool Solver::hasWatch(Literal p, Constraint* c) const {
@@ -449,15 +552,19 @@ bool Solver::simplify() {
 	if (shufSimp_) { simplifySAT(); }
 	return true;
 }
-
+Var  Solver::pushTagVar(bool pushToRoot) {
+	if (isSentinel(tag_)) { tag_ = posLit(pushAuxVar()); }
+	if (pushToRoot)       { pushRoot(tag_); }
+	return tag_.var();
+}
 void Solver::removeConditional() { 
-	Literal p = ~sharedContext()->tagLiteral();
+	Literal p = ~tagLiteral();
 	if (!isSentinel(p)) {
 		ConstraintDB::size_type i, j, end = learnts_.size();
 		for (i = j = 0; i != end; ++i) {
 			ClauseHead* c = static_cast<LearntConstraint*>(learnts_[i])->clause();
 			if (!c || !c->tagged()) {
-				learnts_[j++] = c;
+				learnts_[j++] = learnts_[i];
 			}
 			else {
 				c->destroy(this, true);
@@ -468,13 +575,13 @@ void Solver::removeConditional() {
 }
 
 void Solver::strengthenConditional() { 
-	Literal p = ~sharedContext()->tagLiteral();
+	Literal p = ~tagLiteral();
 	if (!isSentinel(p)) {
 		ConstraintDB::size_type i, j, end = learnts_.size();
 		for (i = j = 0; i != end; ++i) {
 			ClauseHead* c = static_cast<LearntConstraint*>(learnts_[i])->clause();
 			if (!c || !c->tagged() || !c->strengthen(*this, p, true).second) {
-				learnts_[j++] = c;
+				learnts_[j++] = learnts_[i];
 			}
 			else {
 				assert((decisionLevel() == rootLevel() || !c->locked(*this)) && "Solver::strengthenConditional(): must not remove locked constraint!");
@@ -491,45 +598,30 @@ bool Solver::simplifySAT() {
 	}
 	assert(assign_.qEmpty());
 	assign_.front = lastSimp_;
-	while (!assign_.qEmpty()) {
-		simplifyShort(assign_.qPop()); // remove satisfied binary- and ternary clauses
+	lastSimp_     = (uint32)assign_.trail.size();
+	for (Literal p; !assign_.qEmpty(); ) {
+		p = assign_.qPop();
+		releaseVec(watches_[p.index()]);
+		releaseVec(watches_[(~p).index()]);
+		shared_->simplifyShort(*this, p);
 	}
-	lastSimp_ = (uint32)assign_.front;
-	if (shufSimp_) {
+	bool shuffle = shufSimp_ != 0;
+	shufSimp_    = 0;
+	if (shuffle) {
 		std::random_shuffle(constraints_.begin(), constraints_.end(), rng);
 		std::random_shuffle(learnts_.begin(), learnts_.end(), rng);
 	}
-	bool shuffle = shufSimp_ != 0;
-	if (this == sharedContext()->master()) { shared_->simplifyDB(shuffle); }
-	else                                   { simplifyDB(*this, constraints_, shuffle); }
+	if (this == shared_->master()) { shared_->simplify(shuffle); }
+	else                           { simplifyDB(*this, constraints_, shuffle); }
 	simplifyDB(*this, learnts_, shuffle);
 	post_.simplify(*this, shuffle);
 	if (enum_ && enum_->simplify(*this, shuffle)) {
 		enum_->destroy(this, false);
 		enum_ = 0;
 	}
-	shufSimp_ = 0;
 	return true;
 }
 
-// removes all binary clauses containing p - those are now SAT
-// binary clauses containing ~p are unit and therefore likewise SAT. Those
-// are removed when their second literal is processed.
-// Note: Binary clauses containing p are those that watch ~p.
-//
-// Simplifies ternary clauses.
-// Ternary clauses containing p are SAT and therefore removed.
-// Ternary clauses containing ~p are now either binary or SAT. Those that
-// are SAT are removed when the satisfied literal is processed. 
-// All conditional binary-clauses are replaced with a real binary clause.
-// Note: Ternary clauses containing p watch ~p. Those containing ~p watch p.
-// Note: Those clauses are now either binary or satisfied.
-void Solver::simplifyShort(Literal p) {
-	releaseVec(watches_[p.index()]);
-	releaseVec(watches_[(~p).index()]);
-	shared_->simplify(*this, p);
-}
-
 void Solver::setConflict(Literal p, const Antecedent& a, uint32 data) {
 	++stats.conflicts;
 	conflict_.push_back(~p);
@@ -594,12 +686,13 @@ bool Solver::unitPropagate() {
 	Literal p, q, r;
 	uint32 idx, ignore, DL = decisionLevel();
 	const ShortImplicationsGraph& btig = shared_->shortImplications();
+	const uint32 maxIdx = btig.size();
 	while ( !assign_.qEmpty() ) {
 		p             = assign_.qPop();
 		idx           = p.index();
 		WatchList& wl = watches_[idx];
 		// first: short clause BCP
-		if (!btig.propagate(*this, p)) {
+		if (idx < maxIdx && !btig.propagate(*this, p)) {
 			return false;
 		}
 		// second: clause BCP
@@ -684,6 +777,8 @@ bool Solver::backtrack() {
 		levels_.backtrack = decisionLevel() - 1;
 		undoUntil(backtrackLevel(), true);
 	} while (hasConflict() || !force(lastChoiceInverted, 0));
+	// remember flipped literal for copyGuidingPath()
+	impliedLits_.add(decisionLevel(), ImpliedLiteral(lastChoiceInverted, decisionLevel(), 0));
 	return true;
 }
 
@@ -693,10 +788,10 @@ bool ImpliedList::assign(Solver& s) {
 	const uint32 DL     = s.decisionLevel();
 	VecType::iterator j = lits.begin() + front;
 	for (VecType::iterator it = j, end = lits.end(); it != end; ++it) {
-		if (it->level <= DL && ok) {
-			ok = s.force(it->lit, it->ante.ante(), it->ante.data());
+		if(it->level <= DL) {
+			ok = ok && s.force(it->lit, it->ante.ante(), it->ante.data());
+			if (it->level < DL || it->ante.ante().isNull()) { *j++ = *it; }
 		}
-		if (it->level < DL) { *j++ = *it; }
 	}
 	lits.erase(j, lits.end());
 	level = DL * uint32(!lits.empty());
@@ -717,9 +812,8 @@ void Solver::undoUntil(uint32 level) {
 	while (--num) { undoLevel(sp); }
 	assert(level == decisionLevel());
 }
-
 uint32 Solver::undoUntil(uint32 level, bool popBt) {
-	if (popBt && backtrackLevel() > level && !shared_->project(decision(backtrackLevel()).var())) {
+	if (popBt && backtrackLevel() > level && !varInfo(decision(backtrackLevel()).var()).project()) {
 		setBacktrackLevel(level);
 	}
 	undoUntil(level);
@@ -737,9 +831,10 @@ uint32 Solver::estimateBCP(const Literal& p, int rd) const {
 	self.assign_.setValue(p.var(), trueValue(p));
 	self.assign_.trail.push_back(p);
 	const ShortImplicationsGraph& btig = shared_->shortImplications();
+	const uint32 maxIdx = btig.size();
 	do {
 		Literal x = assign_.trail[i++];  
-		if (!btig.propagateBin(self.assign_, x, 0)) {
+		if (x.index() < maxIdx && !btig.propagateBin(self.assign_, x, 0)) {
 			break;
 		}
 	} while (i < assign_.assigned() && rd-- != 0);
@@ -988,7 +1083,8 @@ uint32 Solver::simplifyConflictClause(LitVec& cc, ClauseInfo& info, ClauseHead*
 		clearSeen(cc[0].var());
 	}
 	// 4. finalize
-	jl = finalizeConflictClause(cc, info);
+	uint32 repMode = cc.size() < std::max(strategy.compress, decisionLevel()+1) ? 0 : strategy.ccRepMode;
+	jl = finalizeConflictClause(cc, info, repMode);
 	// 5. bump vars implied by learnt constraints with small lbd 
 	if (!bumpAct_.empty()) {
 		WeightLitVec::iterator j = bumpAct_.begin();
@@ -1100,7 +1196,7 @@ Antecedent Solver::ccHasReverseArc(Literal p, uint32 maxLevel, uint32 maxNew) {
 	assert(seen(p.var()) && isFalse(p) && level(p.var()) == maxLevel);
 	const ShortImplicationsGraph& btig = shared_->shortImplications();
 	Antecedent ante;
-	if (btig.reverseArc(*this, p, maxLevel, ante)) { return ante; }
+	if (p.index() < btig.size() && btig.reverseArc(*this, p, maxLevel, ante)) { return ante; }
 	WatchList& wl   = watches_[p.index()];
 	WatchList::left_iterator it, end; 
 	for (it = wl.left_begin(), end = wl.left_end();  it != end;  ++it) {
@@ -1134,19 +1230,22 @@ void Solver::ccResolve(LitVec& cc, uint32 pos, const LitVec& reason) {
 //  - literals and decision levels in cc are no longer marked
 //  - if cc.size() > 1: cc[1] is a literal from the asserting level
 // RETURN: asserting level of conflict clause.
-uint32 Solver::finalizeConflictClause(LitVec& cc, ClauseInfo& info) {
+uint32 Solver::finalizeConflictClause(LitVec& cc, ClauseInfo& info, uint32 ccRepMode) {
 	// 2. clear flags and compute lbd
 	uint32  lbd         = 1;
 	uint32  onRoot      = 0;
 	uint32  varLevel    = 0;
 	uint32  assertLevel = 0;
 	uint32  assertPos   = 1;
-	Literal tagLit      = ~sharedContext()->tagLiteral();
+	uint32  maxVar      = cc[0].var();
+	Literal tagLit      = ~tagLiteral();
 	bool    tagged      = false;
 	for (LitVec::size_type i = 1; i != cc.size(); ++i) {
-		clearSeen(cc[i].var());
+		Var v = cc[i].var();
+		clearSeen(v);
 		if (cc[i] == tagLit) { tagged = true; }
-		if ( (varLevel = level(cc[i].var())) > assertLevel ) {
+		if (v > maxVar)      { maxVar = v;    }
+		if ( (varLevel = level(v)) > assertLevel ) {
 			assertLevel = varLevel;
 			assertPos   = static_cast<uint32>(i);
 		}
@@ -1156,9 +1255,48 @@ uint32 Solver::finalizeConflictClause(LitVec& cc, ClauseInfo& info) {
 		}
 	}
 	if (assertPos != 1) { std::swap(cc[1], cc[assertPos]); }
+	if (ccRepMode == SolverStrategies::cc_rep_dynamic) {
+		ccRepMode = double(lbd)/double(decisionLevel()) > .66 ? SolverStrategies::cc_rep_decision : SolverStrategies::cc_rep_uip;
+	}
+	if (ccRepMode) {
+		maxVar = cc[0].var(), tagged = false, lbd = 1;
+		if (ccRepMode == SolverStrategies::cc_rep_decision) {
+			// replace cc with decision sequence
+			cc.resize(assertLevel+1);
+			for (uint32 i = assertLevel; i;){
+				Literal x = ~decision(i--);
+				cc[lbd++] = x;
+				if (x == tagLit)     { tagged = true; }
+				if (x.var() > maxVar){ maxVar = x.var(); }
+			}
+		}
+		else {
+			// replace cc with all uip clause
+			uint32 marked = cc.size() - 1;
+			while (cc.size() > 1) { markSeen(~cc.back()); cc.pop_back(); }
+			for (LitVec::const_iterator tr = assign_.trail.end(), next, stop; marked;) {
+				while (!seen(*--tr)) { ; }
+				bool n = --marked != 0 && !reason(*tr).isNull();
+				clearSeen(tr->var());
+				if (n) { for (next = tr, stop = assign_.trail.begin() + levelStart(level(tr->var())); next-- != stop && !seen(*next);) { ; } }
+				if (!n || level(next->var()) != level(tr->var())) {
+					cc.push_back(~*tr);
+					if (tr->var() == tagLit.var()){ tagged = true; }
+					if (tr->var() > maxVar)       { maxVar = tr->var(); }
+				}
+				else {
+					for (reason(*tr, conflict_); !conflict_.empty(); conflict_.pop_back()) {
+						if (!seen(conflict_.back())) { ++marked; markSeen(conflict_.back()); }
+					}
+				}
+			}
+			lbd = cc.size();
+		}
+	}
 	info.setActivity(ccInfo_.activity());
 	info.setLbd(lbd);
 	info.setTagged(tagged);
+	info.setAux(auxVar(maxVar));
 	return assertLevel;
 }
 
@@ -1372,6 +1510,7 @@ uint64 Solver::updateBranch(uint32 cfl) {
 // The basic DPLL-like search-function
 /////////////////////////////////////////////////////////////////////////////////////////
 ValueRep Solver::search(SearchLimits& limit, double rf) {
+	assert(!isFalse(tagLiteral()));
 	uint64 local = limit.local != UINT64_MAX ? limit.local : 0;
 	rf           = std::max(0.0, std::min(1.0, rf));
 	if (local && decisionLevel() == rootLevel()) { cflStamp_.assign(decisionLevel()+1, 0); }
