@@ -34,13 +34,15 @@ namespace Clasp {
 WeightLitsRep WeightLitsRep::create(Solver& s, WeightLitVec& lits, weight_t bound) {
 	// Step 1: remove assigned/superfluous literals and merge duplicate/complementary ones
 	LitVec::size_type j = 0, other;
+	const weight_t MAX_W= std::numeric_limits<weight_t>::max();
 	for (LitVec::size_type i = 0; i != lits.size(); ++i) {
 		lits[i].first.clearWatch();
 		Literal x = lits[i].first;
-		if (s.value(x.var()) == value_free && lits[i].second != 0) {
+		if (lits[i].second != 0 && s.topValue(x.var()) == value_free) {
 			if (lits[i].second < 0) {
 				lits[i].second = -lits[i].second;
 				lits[i].first  = ~lits[i].first;
+				CLASP_ASSERT_CONTRACT_MSG(bound < 0 || (MAX_W-bound) >= lits[i].second, "bound out of range");
 				bound         += lits[i].second;
 			}
 			if (!s.seen(x.var())) { // first time we see x, keep and mark x
@@ -76,13 +78,14 @@ WeightLitsRep WeightLitsRep::create(Solver& s, WeightLitVec& lits, weight_t boun
 	if (bound < 0) { bound = 0; }
 	// Step 2: compute min,max, achievable weight and clear flags set in step 1
 	weight_t sumW = 0;
-	uint32 minW   = std::numeric_limits<uint32>::max(), maxW = 1;
+	weight_t minW = MAX_W, maxW = 1;
 	weight_t B    = std::max(bound, 1);
 	for (LitVec::size_type i = 0; i != lits.size(); ++i) {
 		assert(lits[i].second > 0);
 		s.clearSeen(lits[i].first.var());
-		if (uint32(lits[i].second) > maxW) { maxW = lits[i].second = std::min(lits[i].second, B);  }
-		if (uint32(lits[i].second) < minW) { minW = lits[i].second; }
+		if (lits[i].second > maxW) { maxW = lits[i].second = std::min(lits[i].second, B);  }
+		if (lits[i].second < minW) { minW = lits[i].second; }
+		CLASP_ASSERT_CONTRACT_MSG((MAX_W - sumW) >= lits[i].second, "Sum of weights out of range");
 		sumW += lits[i].second;
 	}
 	// Step 3: sort by decreasing weight
@@ -104,9 +107,11 @@ WeightLitsRep WeightLitsRep::create(Solver& s, WeightLitVec& lits, weight_t boun
 
 // Propagates top-level assignment.
 bool WeightLitsRep::propagate(Solver& s, Literal W) {
-	if      (bound <= 0)    { return s.addUnary(W, Constraint_t::static_constraint); } // trivially SAT
-	else if (bound > reach) { return s.addUnary(~W, Constraint_t::static_constraint);} // trivially UNSAT
-	else if (s.value(W.var()) == value_free) { return true; }
+	if      (bound <= 0)    { return s.force(W); } // trivially SAT
+	else if (bound > reach) { return s.force(~W);} // trivially UNSAT
+	else if (s.topValue(W.var()) == value_free) { 
+		return true; 
+	}
 	// backward propagate
 	bool bpTrue = s.isTrue(W);
 	weight_t B  = bpTrue ? (reach-bound)+1 : bound;
@@ -154,44 +159,52 @@ uint32 WeightConstraint::WL::refCount() const {
 /////////////////////////////////////////////////////////////////////////////////////////
 // WeightConstraint
 /////////////////////////////////////////////////////////////////////////////////////////
-bool WeightConstraint::newWeightConstraint(SharedContext& ctx, Literal W, WeightLitVec& lits, weight_t bound, Constraint** out) {
-	WeightLitsRep rep = WeightLitsRep::create(*ctx.master(), lits, bound);
-	return newWeightConstraint(ctx, W, rep, out);
+WeightConstraint::CPair WeightConstraint::create(Solver& s, Literal W, WeightLitVec& lits, weight_t bound, uint32 flags) {
+	return WeightConstraint::create(s, W, WeightLitsRep::create(s, lits, bound), flags);
 }
-// Adds constraints for W == (lits == bound)
-bool WeightConstraint::newWeightEqConstraint(SharedContext& ctx, Literal W, WeightLitVec& lits, weight_t bound) {
-	WeightLitsRep rep = WeightLitsRep::create(*ctx.master(), lits, bound);
-	return newWeightConstraint(ctx, W, rep)
-		&&   (++rep.bound, newWeightConstraint(ctx, ~W, rep));
+WeightConstraint::CPair WeightConstraint::create(Solver& s, Literal W, WeightLitsRep rep, uint32 flags) {
+	CPair res;
+	if (!s.sharedContext()->physicalShareProblem()) { flags |= create_no_share; }
+	if (s.sharedContext()->frozen())                { flags |= (create_no_freeze|create_no_share); }
+	res.con[0] = createImpl(s, W, rep, flags);
+	if ((flags & create_eq_bound) != 0 && res.ok()) {
+		// Adds constraints for W == (lits == bound)
+		++rep.bound;
+		res.con[1] = createImpl(s, ~W, rep, flags);
+	}
+	return res;
 }
-bool WeightConstraint::newWeightConstraint(SharedContext& ctx, Literal W, WeightLitsRep& rep, Constraint** out) {
-	if (out) *out = 0;
-	Solver& s = *ctx.master();
-	if (!rep.propagate(s, W)){ return false; }
-	if (!rep.open())         { return true;  }
-	if (rep.bound == 1 || rep.bound == rep.reach) {
+
+WeightConstraint* WeightConstraint::createImpl(Solver& s, Literal W, WeightLitsRep& rep, uint32 flags) {
+	WeightConstraint* conflict = (WeightConstraint*)0x1;
+	bool sat = (flags&create_sat) != 0 && rep.size;
+	if (!rep.propagate(s, W))              { return conflict; }
+	if (rep.unsat() || (rep.sat() && !sat)){ return 0; }
+	if ((rep.bound == 1 || rep.bound == rep.reach) && (flags & create_explicit) == 0) {
 		LitVec clause; clause.reserve(1 + rep.size);
+		Literal bin[2];
 		bool disj = rep.bound == 1; // con == disjunction or con == conjunction
 		bool sat  = false;
 		clause.push_back(W ^ disj);
 		for (LitVec::size_type i = 0; i != rep.size; ++i) { 
-			Literal x = ~clause[0];
-			Literal y = rep.lits[i].first ^ disj;
-			if (x != ~y) {
-				if (x != y)                                       { clause.push_back(~y); }
-				if (!ClauseCreator::create(s, ShortRep(x, y), 0)) { return false; }
+			bin[0] = ~clause[0];
+			bin[1] = rep.lits[i].first ^ disj;
+			if (bin[0] != ~bin[1]) {
+				if (bin[0] != bin[1])                 { clause.push_back(~bin[1]); }
+				if (!s.add(ClauseRep::create(bin, 2))){ return conflict; }
 			}
 			else { sat = true; }
 		}
-		return sat || ClauseCreator::create(s, clause, 0);
+		return (sat || ClauseCreator::create(s, clause, 0)) ? 0 : conflict;
 	}
+	assert(rep.open() || (rep.sat() && sat));
 	bool   hasW = rep.hasWeights();
 	uint32 size = 1 + rep.size;
 	uint32 nb   = sizeof(WeightConstraint) + (size+hasW)*sizeof(UndoInfo);
 	uint32 wls  = sizeof(WL) + (size << uint32(hasW))*sizeof(Literal);
 	void*  m    = 0;
 	WL*    sL   = 0;
-	if (!ctx.physicalShareProblem()) {
+	if ((flags & create_no_share) != 0) {
 		nb        = ((nb + 3) / 4)*4;
 		m         = ::operator new (nb + wls);
 		sL        = new (reinterpret_cast<unsigned char*>(m)+nb) WL(size, false, hasW);
@@ -204,36 +217,34 @@ bool WeightConstraint::newWeightConstraint(SharedContext& ctx, Literal W, Weight
 		sL        = new (t+sizeof(uint32)) WL(size, true, hasW);
 	}
 	assert(m && (reinterpret_cast<uintp>(m) & 7u) == 0);
-	Constraint* c = new (m) WeightConstraint(ctx, W, rep, sL);
-	if (out) *out = c;
-	s.add(c);
-	return true;
+	SharedContext*  ctx = (flags & create_no_freeze) == 0 ? const_cast<SharedContext*>(s.sharedContext()) : 0;
+	WeightConstraint* c = new (m) WeightConstraint(s, ctx, W, rep, sL);
+	if ((flags & create_no_add) == 0) { s.add(c); }
+	return c;
 }
-
-WeightConstraint::WeightConstraint(SharedContext& ctx, Literal W, const WeightLitsRep& rep, WL* out) {
+WeightConstraint::WeightConstraint(Solver& s, SharedContext* ctx, Literal W, const WeightLitsRep& rep, WL* out) {
 	const bool hasW = rep.hasWeights();
-	Solver& s       = *ctx.master();
 	lits_           = out;
 	ownsLit_        = !out->shareable();
 	Literal* p      = lits_->lits;
 	Literal* h      = (Literal*)undo_;
 	weight_t w      = 1;
 	Var      mv     = W.var();
-	bound_[FFB_BTB]	= (rep.reach-rep.bound)+1;// ffb-btb
+	bound_[FFB_BTB]	= (rep.reach-rep.bound)+1; // ffb-btb
 	bound_[FTB_BFB]	= rep.bound;               // ftb-bfb
-	*p++            = ~W;                  // store constraint literal
-	if (hasW) *p++  = Literal::fromRep(w); // and weight if necessary
-	ctx.setFrozen(W.var(), true);          // exempt from variable elimination
-	active_ = s.value(W.var())==value_free // if unassigned
-		? NOT_ACTIVE                         // both constraints are relevant
-		: FFB_BTB+s.isFalse(W);              // otherwise only one is relevant
+	*p++            = ~W;                      // store constraint literal
+	if (hasW) *p++  = Literal::fromRep(w);     // and weight if necessary
+	if (ctx) ctx->setFrozen(W.var(), true);    // exempt from variable elimination
+	active_ = s.topValue(W.var())==value_free  // if unassigned
+		? NOT_ACTIVE                             // both constraints are relevant
+		: FFB_BTB+s.isFalse(W);                  // otherwise only one is relevant
 	for (uint32 i = 0, j = 1, end = rep.size; i != end; ++i, ++j) {
-		*p++ = h[j] = rep.lits[i].first;     // store constraint literal
-		w    = rep.lits[i].second;           // followed by weight 
-		if (hasW) *p++= Literal::fromRep(w); // if necessary
-		addWatch(s, j, FTB_BFB);             // watches  lits[idx]
-		addWatch(s, j, FFB_BTB);             // watches ~lits[idx]
-		ctx.setFrozen(h[j].var(), true);     // exempt from variable elimination
+		*p++ = h[j] = rep.lits[i].first;         // store constraint literal
+		w    = rep.lits[i].second;               // followed by weight 
+		if (hasW) *p++= Literal::fromRep(w);     // if necessary
+		addWatch(s, j, FTB_BFB);                 // watches  lits[idx]
+		addWatch(s, j, FFB_BTB);                 // watches ~lits[idx]
+		if (ctx) ctx->setFrozen(h[j].var(), true);// exempt from variable elimination
 		mv   = std::max(mv, h[j].var());
 	}
 	if (hasW) s.requestData(mv+1);         // weight constraints can become unit more than once
@@ -303,6 +314,9 @@ void WeightConstraint::destroy(Solver* s, bool detach) {
 			s->removeWatch( lits_->lit(i), this );
 			s->removeWatch(~lits_->lit(i), this );
 		}
+		for (uint32 i = s->decisionLevel(); i; --i) {
+			s->removeUndoWatch(i, this);
+		}
 	}
 	if (ownsLit_ == 0) { lits_->release(); }
 	void* mem    = static_cast<Constraint*>(this);
