@@ -66,13 +66,13 @@ BodyAggregateDomain::BodyAggregateDomain(UTerm &&repr, BoundVec &&bounds, Aggreg
         case AggregateFunction::COUNT:
         case AggregateFunction::SUMP:
         case AggregateFunction::MAX: {
-            for (auto &x : bounds) {
+            for (auto &x : this->bounds) {
                 if (x.rel != Relation::GT && x.rel != Relation::GEQ) { positive = false; break; }
             }
             break;
         }
         case AggregateFunction::MIN: {
-            for (auto &x : bounds) {
+            for (auto &x : this->bounds) {
                 if (x.rel != Relation::LT && x.rel != Relation::LEQ) { positive = false; break; }
             }
             break;
@@ -282,12 +282,13 @@ void HeadAggregateDomain::accumulateUnmark(Queue &queue) {
                     y.second.imported = y.second.conds.size();
                 }
             }
+            x.todo = false;
         }
         todo.clear();
         auto headIt(heads.begin());
         for (auto &x : import) { 
+            auto &head = *headIt++;
             if (!x.empty()) {
-                auto &head = *headIt++;
                 for (PredicateDomain::element_type &y : x) { head.first.insert(y); }
                 head.first.mark();
                 head.first.unmark();
@@ -348,6 +349,18 @@ void init(BoundVec const &bounds, IntervalSet<Value> &set) {
     }
 }
 
+BodyAggregateDomain::element_type &BodyAggregateDomain::reserve(Value repr) {
+    auto state = domain.find(repr);
+    if (state == domain.end()) { 
+        state = domain.emplace(std::piecewise_construct, std::forward_as_tuple(repr), std::forward_as_tuple()).first;
+        state->second.init(fun);
+        Gringo::Ground::init(bounds, state->second.bounds);
+        exports.append(*state);
+        state->second.state = BodyAggregateState::UNKNOWN;
+    }
+    return *state;
+}
+
 void BodyAggregateDomain::insert(Value const &repr, ValVec const &tuple, Output::LitVec const &cond, Location const &loc) {
     auto state = domain.find(repr);
     if (state == domain.end()) { 
@@ -444,7 +457,7 @@ void ConjunctionDomain::insert() {
             state->second._fact = false;
             state->second.elems.emplace_back(&elem, std::move(cond));
         }
-   }
+    }
     else if (cond.empty()) { state->second.blocked.emplace(Value()); } // Note: blocks forever
     else {
         state->second._fact = false;
@@ -698,7 +711,7 @@ Output::Literal *BodyAggregateLiteral::toOutput() {
     switch (gLit.naf) {
         case NAF::POS:
         case NAF::NOTNOT: { return !gLit.repr->second.fact(gLit.incomplete) ? &gLit : nullptr; }
-        case NAF::NOT:    { return !gLit.repr->second.isFalse() ? &gLit : nullptr; }
+        case NAF::NOT:    { return gLit.incomplete || !gLit.repr->second.isFalse() ? &gLit : nullptr; }
     }
     assert(false);
     return nullptr;
@@ -831,7 +844,7 @@ void Rule::report(Output::OutputBase &out) {
         if (defines) {
             Value val(defines->repr->eval());
             auto ret(domain->insert(val, false));
-            out.external(*std::get<0>(ret));
+            out.external(*std::get<0>(ret), Output::ExternalType::E_FALSE);
         }
     }
     else {
@@ -1175,6 +1188,7 @@ void ConjunctionAccumulateFact::analyze(Dep::Node &node, Dep &dep) {
 }
 void HeadAggregateAccumulate::analyze(Dep::Node &node, Dep &dep) {
     if (headNum > 0) { dep.provides(node, dom->head(headNum).second, dom->head(headNum).second.getRepr()); }
+    // TODO: find out why exactly I added this dependency
     dep.provides(node, dom->dummy, dom->dummy.getRepr());
     dep.depends(node, dom->dummy);
     _analyze(node, dep, lits);
