@@ -304,6 +304,18 @@ int atomsModel(lua_State *L) {
     }
     return 1;
 }
+int optimizationModel(lua_State *L) {
+    Model const *& model = *(Model const **)luaL_checkudata(L, 1, "gringo.Model");
+    Int64Vec *values = luaCppNew<Int64Vec>(L, "gringo._Int64Vec");
+    protect<void>(L, [&model, values]() { *values = model->optimization(); });
+    lua_createtable(L, values->size(), 0);
+    int i = 1;
+    for (auto x : *values) { 
+        lua_pushinteger(L, x);
+        lua_rawseti(L, -2, i++);
+    }
+    return 1;
+}
 int toStringModel(lua_State *L) { 
     Model const *& model =  *(Model const **)luaL_checkudata(L, 1, "gringo.Model");
     std::string *rep = luaCppNew<std::string>(L, "gringo._string");
@@ -621,6 +633,7 @@ int luaopen_gringo(lua_State* L) {
     static luaL_Reg const modelMeta[] = {
         {"__tostring", toStringModel},
         {"atoms", atomsModel},
+        {"optimization", optimizationModel},
         {"contains", containsModel},
         {nullptr, nullptr}
     };
@@ -651,6 +664,10 @@ int luaopen_gringo(lua_State* L) {
         {"__gc", luaCppDelete<ValVec>},
         {nullptr, nullptr}
     };
+    static luaL_Reg const _Int64VecMeta[] = {
+        {"__gc", luaCppDelete<Int64Vec>},
+        {nullptr, nullptr}
+    };
     static luaL_Reg const _FWStringVecMeta[] = {
         {"__gc", luaCppDelete<FWStringVec>},
         {nullptr, nullptr}
@@ -683,6 +700,7 @@ int luaopen_gringo(lua_State* L) {
     regMeta("gringo.Control",      controlMeta);
     regMeta("gringo._string",      _stringMeta);
     regMeta("gringo._ValVec",      _ValVecMeta);
+    regMeta("gringo._Int64Vec",    _Int64VecMeta);
     regMeta("gringo._FWStringVec", _FWStringVecMeta);
 
 #if LUA_VERSION_NUM < 502
