@@ -104,6 +104,12 @@ int PredicateLiteral::lparseUid(LparseOutputter &out) const {
     assert(false);
     return 0;
 }
+bool PredicateLiteral::invertible() const {
+    return naf != NAF::POS;
+}
+void PredicateLiteral::invert() {
+    naf = inv(naf);
+}
 size_t PredicateLiteral::hash() const { return get_value_hash(typeid(PredicateLiteral).hash_code(), naf, repr->first); }
 bool PredicateLiteral::operator==(Literal const &x) const { 
     PredicateLiteral const *t{dynamic_cast<PredicateLiteral const*>(&x)};
@@ -279,7 +285,8 @@ ULit BodyAggregate::toLparse(LparseTranslator &x) {
     // TODO: check if false
     auto rng(repr->second.range(fun));
     SAuxAtom auxAtom(std::make_shared<AuxAtom>(x.auxAtom()));
-    if (!repr->second.bounds.contains(rng)) {
+    if (repr->second.bounds.contains(rng)) { LparseRule(auxAtom, ULitVec()).toLparse(x); } 
+    else {
         using ULitValVec = std::vector<std::pair<ULit,Value>>;
         ULitValVec elemVec;
         for (auto &y : repr->second.elems) { 
@@ -532,6 +539,31 @@ size_t AssignmentAggregate::hash() const                               { throw s
 bool AssignmentAggregate::operator==(Literal const &) const            { throw std::runtime_error("AssignmentAggregate::operator==: implement me if necessary!"); }
 AssignmentAggregate::~AssignmentAggregate()                            { }
 
+// }}}
+// {{{ definition of ConjunctionElem
+
+ConjunctionElem::ConjunctionElem(PredicateDomain::element_type* head, ULitVec &&body)
+    : head(head)
+    , body(std::move(body)) { }
+
+size_t ConjunctionElem::hash() const {
+    return get_value_hash(size_t(head), body);
+}
+
+bool ConjunctionElem::operator==(ConjunctionElem const &x) const {
+    return head == x.head && is_value_equal_to(body, x.body);
+}
+
+void ConjunctionElem::print(std::ostream &out) const {
+    if (head) { out << head->first; }
+    else      { out << "#false"; }
+    if (!body.empty()) {
+        out << ":";
+        using namespace std::placeholders;
+        print_comma(out, body, ",", std::bind(&Literal::printPlain, _2, _1));
+    }
+}
+
 // }}}
 // {{{ definition of ConjunctionState
 
@@ -545,62 +577,60 @@ bool ConjunctionState::defined() const                     { return _generation
 // }}}
 // {{{ definition of Conjunction
 
-void Conjunction::printElem(std::ostream &out, ConjunctionState::Elem const &x) {
-    if (x.first) { out << x.first->first; }
-    else         { out << "#false"; }
-    if (!x.second.empty()) {
-        out << ":";
-        using namespace std::placeholders;
-        print_comma(out, x.second, ",", std::bind(&Literal::printPlain, _2, _1));
-    }
-}
 void Conjunction::printPlain(std::ostream &out) const {
     if (!repr->second.elems.empty()) {
-        print_comma(out, repr->second.elems, ";", &Conjunction::printElem);
+        print_comma(out, repr->second.elems, ";");
     }
     else { out << "#true"; }
 }
 ULit Conjunction::toLparse(LparseTranslator &x) {
     // TODO: at some point I should try to prove this translation
-    SAuxAtom bdLit(std::make_shared<AuxAtom>(x.auxAtom()));
-    ULitVec bd;
-    for (ConjunctionState::Elem &y : repr->second.elems) {
-        if (!y.second.empty()) {
-            SAuxAtom aux(std::make_shared<AuxAtom>(x.auxAtom()));
-            if (y.first && y.first->second.defined()) {
-                // aux :- x.first.
+    if (!repr->second.bdLit) {
+        repr->second.bdLit = std::make_shared<AuxAtom>(x.auxAtom());
+        ULitVec bd;
+        for (ConjunctionState::Elem &y : repr->second.elems) {
+            if (!y.head && y.body.size() == 1 && y.body.front()->invertible()) {
+                ULit inv(get_clone(y.body.front()));
+                inv->invert();
+                bd.emplace_back(std::move(inv));
+            }
+            else if (!y.body.empty()) {
+                SAuxAtom aux(std::make_shared<AuxAtom>(x.auxAtom()));
+                if (y.head) {
+                    // aux :- x.first.
+                    ULitVec lits;
+                    lits.emplace_back(make_unique<PredicateLiteral>(NAF::POS, *y.head));
+                    LparseRule(aux, std::move(lits)).toLparse(x);
+                }
+                SAuxAtom chk(std::make_shared<AuxAtom>(x.auxAtom()));
+                // chk :- y.second.
                 ULitVec lits;
-                lits.emplace_back(make_unique<PredicateLiteral>(NAF::POS, *y.first));
+                for (auto &z : y.body) { lits.emplace_back(get_clone(z)); }
+                LparseRule(chk, std::move(lits)).toLparse(x);
+                // aux :- ~chk.
+                lits.emplace_back(make_unique<AuxLiteral>(chk, true));
                 LparseRule(aux, std::move(lits)).toLparse(x);
+                if (incomplete && y.head) {
+                    // aux | chk :- ~~x.first.
+                    SAuxAtomVec head;
+                    head.emplace_back(aux);
+                    head.emplace_back(chk);
+                    lits.emplace_back(make_unique<PredicateLiteral>(NAF::NOTNOT, *y.head));
+                    LparseRule(std::move(head), std::move(lits), false).toLparse(x);
+                    // y.second :- chk.
+                    for (auto &z : y.body) { z->makeEqual(make_unique<AuxLiteral>(chk, false), x); }
+                }
+                // body += aux
+                bd.emplace_back(make_unique<AuxLiteral>(aux, false));
             }
-            SAuxAtom chk(std::make_shared<AuxAtom>(x.auxAtom()));
-            // chk :- y.second.
-            ULitVec lits;
-            for (auto &z : y.second) { lits.emplace_back(get_clone(z)); }
-            LparseRule(chk, std::move(lits)).toLparse(x);
-            // aux :- ~chk.
-            lits.emplace_back(make_unique<AuxLiteral>(chk, true));
-            LparseRule(aux, std::move(lits)).toLparse(x);
-            if (incomplete && y.first && y.first->second.defined()) {
-                // aux | chk :- ~~x.first.
-                SAuxAtomVec head;
-                head.emplace_back(aux);
-                head.emplace_back(chk);
-                lits.emplace_back(make_unique<PredicateLiteral>(NAF::NOTNOT, *y.first));
-                LparseRule(std::move(head), std::move(lits), false).toLparse(x);
-                // y.second :- chk.
-                for (auto &z : y.second) { z->makeEqual(make_unique<AuxLiteral>(chk, false), x); }
-            }
-            // body += aux
-            bd.emplace_back(make_unique<AuxLiteral>(aux, false));
+            // body += x.first
+            else if (y.head) { bd.emplace_back(make_unique<PredicateLiteral>(NAF::POS, *y.head)); }
+            else             { return make_unique<AuxLiteral>(repr->second.bdLit, false); }
         }
-        // body += x.first
-        else if (y.first) { bd.emplace_back(make_unique<PredicateLiteral>(NAF::POS, *y.first)); }
-        else              { return make_unique<AuxLiteral>(bdLit, false); }
+        // bdLit :- body.
+        LparseRule(repr->second.bdLit, std::move(bd)).toLparse(x);
     }
-    // bdLit :- body.
-    LparseRule(bdLit, std::move(bd)).toLparse(x);
-    return make_unique<AuxLiteral>(bdLit, false);
+    return make_unique<AuxLiteral>(repr->second.bdLit, false);
 }
 void Conjunction::makeEqual(ULit &&, LparseTranslator &) const {
     throw std::runtime_error("AssignmentAggregate::makeEqual: must not happen!!!");
