@@ -606,9 +606,10 @@ struct LparseHandler : StmHandler {
         if ((unsigned)debug & (unsigned)LparseDebug::PLAIN) { std::cerr << "%"; x.printPlain(std::cerr); }
         x.toLparse(trans);
     }
-    virtual void operator()(PredicateDomain::element_type &head) {
+    virtual void incremental() { out.incremental(); }
+    virtual void operator()(PredicateDomain::element_type &head, ExternalType type) {
         if (!head.second.hasUid()) { head.second.uid(out.newUid()); }
-        out.printExternal(head.second.uid());
+        out.printExternal(head.second.uid(), type);
     }
     virtual void finish(PredDomMap &domains, OutputPredicates &outPreds) {
         out.disposeMinimize() = trans.minimizeChanged();
@@ -687,8 +688,12 @@ struct LparsePlainHandler : StmHandler {
         : trans([&out](Statement const &x)                { x.printPlain(out); })
         , out(out) { }
     virtual void operator()(Statement &x)                 { x.toLparse(trans); }
-    virtual void operator()(PredicateDomain::element_type &head) {
-        out << "#external " << head.first << ".\n";
+    virtual void operator()(PredicateDomain::element_type &head, ExternalType type) {
+        switch (type) {
+            case ExternalType::E_FALSE: { out << "#external " << head.first << ".\n"; break; }
+            case ExternalType::E_TRUE:  { out << "#external " << head.first << "=true.\n"; break; }
+            case ExternalType::E_FREE:  { out << "#external " << head.first << "=free.\n"; break; }
+        }
     }
     virtual void finish(PredDomMap &, OutputPredicates &) { trans.translate(); }
     virtual void atoms(int, std::function<bool(unsigned)> const &, ValVec &, PredDomMap const &, OutputPredicates const &) { }
@@ -705,8 +710,12 @@ struct PlainHandler : StmHandler {
     PlainHandler(std::ostream &out) : out(out)            { }
     virtual ~PlainHandler()                               { }
     virtual void operator()(Statement &x)                 { x.printPlain(out); }
-    virtual void operator()(PredicateDomain::element_type &head) {
-        out << "#external " << head.first << ".\n";
+    virtual void operator()(PredicateDomain::element_type &head, ExternalType type) {
+        switch (type) {
+            case ExternalType::E_FALSE: { out << "#external " << head.first << ".\n"; break; }
+            case ExternalType::E_TRUE:  { out << "#external " << head.first << "=true.\n"; break; }
+            case ExternalType::E_FREE:  { out << "#external " << head.first << "=free.\n"; break; }
+        }
     }
     virtual void finish(PredDomMap &, OutputPredicates &outPreds) { 
         for (auto &x : outPreds) { 
@@ -725,6 +734,9 @@ struct PlainHandler : StmHandler {
 // {{{ definition of PlainLparseOutputter
 
 PlainLparseOutputter::PlainLparseOutputter(std::ostream &out) : out(out) { }
+void PlainLparseOutputter::incremental() {
+    out << "90 0\n";
+}
 void PlainLparseOutputter::printBasicRule(unsigned head, LitVec const &body) {
     out << "1 " << head << " " << body.size();
     unsigned neg(0);
@@ -787,11 +799,17 @@ void PlainLparseOutputter::printDisjunctiveRule(AtomVec const &head, LitVec cons
     for (auto &x : body) { if (x > 0) { out << " " << +x; } }
     out << "\n";
 }
+void PlainLparseOutputter::printExternal(unsigned atomUid, ExternalType type) { 
+    switch (type) {
+        case ExternalType::E_FALSE: { out << "91 " << atomUid << " 0\n"; break; }
+        case ExternalType::E_TRUE:  { out << "91 " << atomUid << " 1\n"; break; }
+        case ExternalType::E_FREE:  { out << "92 " << atomUid << "\n"; break; }
+    }
+}
 unsigned PlainLparseOutputter::falseUid()                                   { return 1; }
 unsigned PlainLparseOutputter::newUid()                                     { return uids++; }
 void PlainLparseOutputter::finishRules()                                    { out << "0\n"; }
 void PlainLparseOutputter::printSymbol(unsigned atomUid, Value v)           { out << atomUid << " " << v << "\n"; }
-void PlainLparseOutputter::printExternal(unsigned)                          { }
 void PlainLparseOutputter::finishSymbols()                                  { out << "0\nB+\n0\nB-\n" << falseUid() << "\n0\n1\n"; }
 PlainLparseOutputter::~PlainLparseOutputter()                               { }
 
@@ -814,8 +832,11 @@ void OutputBase::output(Value const &val) {
         (*handler)(tempRule);
     }
 }
-void OutputBase::external(PredicateDomain::element_type &head) {
-    (*handler)(head);
+void OutputBase::incremental() {
+    handler->incremental();
+}
+void OutputBase::external(PredicateDomain::element_type &head, ExternalType type) {
+    (*handler)(head, type);
 }
 void OutputBase::output(UStm &&x) {
     if (!x->isIncomplete()) { (*handler)(*x); }
@@ -870,6 +891,18 @@ Gringo::AtomState const *OutputBase::find(Gringo::Value val) const {
     }
     return nullptr;
 }
+PredicateDomain::element_type *OutputBase::find2(Gringo::Value val) {
+    if (val.type() == Gringo::Value::ID || val.type() == Gringo::Value::FUNC) {
+        auto it = domains.find(val.sig());
+        if (it != domains.end()) {
+            auto jt = it->second.domain.find(val);
+            if (jt != it->second.domain.end() && jt->second.defined()) {
+                return &*jt;
+            }
+        }
+    }
+    return nullptr;
+}
 
 // }}}
 
