@@ -709,6 +709,19 @@ struct Model {
         }
         return list.release();
     }
+    static PyObject *optimization(Model *self) {
+        Int64Vec values(self->model->optimization());
+        Object list = PyList_New(values.size());
+        if (!list) { return nullptr; }
+        int i = 0;
+        for (auto x : values) {
+            Object val = PyInt_FromLong(x);
+            if (!val) { return nullptr; }
+            if (PyList_SetItem(list, i, val.release()) < 0) { return nullptr; }
+            ++i;
+        }
+        return list.release();
+    }
     static PyObject *str(Model *self, PyObject *) {
         std::string s;
         if (!protect([self, &s]() -> void {
@@ -756,15 +769,20 @@ Model.CSP   -- selects all csp assignments (independent of #show statements)
 The string representation of a model object is similar to the output of models
 by clingo using the default output.
 
-Note that atoms are represented using Fun objects, and that CSP
-assignments are represented using function symbols with name "$" where the the
-first argument is the name of the CSP variable and the second its value.)"},
+Note that atoms are represented using Fun objects, and that CSP assignments are
+represented using function symbols with name "$" where the first argument is
+the name of the CSP variable and the second its value.)"},
     {"contains", (PyCFunction)contains, METH_O,       
 R"(contains(self, a) -> Boolean
     
 Returns true if atom a is contained in the model.
 
 Atom a must be represented using a Fun term.)"},
+    {"optimization", (PyCFunction)optimization, METH_NOARGS,
+R"(contains(self) -> [int]
+    
+Returns the list of optimization values of the model. This corresponds to
+clasp's optimization output.)"},
     {nullptr, nullptr, 0, nullptr}
 };
 
