@@ -40,6 +40,7 @@ class TestInstantiation : public CppUnit::TestFixture {
         CPPUNIT_TEST(test_min_true);
         CPPUNIT_TEST(test_min_open);
         CPPUNIT_TEST(test_rec_count);
+        CPPUNIT_TEST(test_rec_count2);
         CPPUNIT_TEST(test_assign_count);
         CPPUNIT_TEST(test_assign_sum);
         CPPUNIT_TEST(test_assign_sump);
@@ -54,6 +55,7 @@ class TestInstantiation : public CppUnit::TestFixture {
         CPPUNIT_TEST(test_strategicB);
         CPPUNIT_TEST(test_optimize);
         CPPUNIT_TEST(test_neg);
+        CPPUNIT_TEST(test_tuple);
     CPPUNIT_TEST_SUITE_END();
 public:
     typedef std::string S;
@@ -76,6 +78,7 @@ public:
     void test_min_true();
     void test_min_open();
     void test_rec_count();
+    void test_rec_count2();
     void test_assign_count();
     void test_assign_sum();
     void test_assign_sump();
@@ -90,6 +93,7 @@ public:
     void test_strategicB();
     void test_optimize();
     void test_neg();
+    void test_tuple();
 
     virtual ~TestInstantiation();
 };
@@ -578,6 +582,15 @@ void TestInstantiation::test_rec_count() {
             "outedge(U,V) :- linked(U,V,I), not reduced(I).\n"));
 }
 
+void TestInstantiation::test_rec_count2() {
+    Gringo::Test::Messages msg;
+    CPPUNIT_ASSERT_EQUAL(
+        std::string(
+            "a:-1!=#count{1,0,a:a}.\n"),
+        ground(
+            "a:-{a}!=1.\n"));
+}
+
 std::string TestInstantiation::gbie() const {
     return
         "char_to_digit(X,X) :- X=0..9.\n"
@@ -1067,6 +1080,14 @@ void TestInstantiation::test_neg() {
             " q(X):-not not  q(X), p(X).\n"));
 }
 
+void TestInstantiation::test_tuple() {
+    CPPUNIT_ASSERT_EQUAL(
+        std::string(
+            "p(((),())).\n"),
+        ground(
+            "p(((),())).\n"));
+}
+
 TestInstantiation::~TestInstantiation() { }
 
 // }}}
