@@ -30,7 +30,12 @@ class TestLparse : public CppUnit::TestFixture {
     CPPUNIT_TEST_SUITE(TestLparse);
         CPPUNIT_TEST(test_empty);
         CPPUNIT_TEST(test_projectionBug);
+        CPPUNIT_TEST(test_recAntiAggr);
         CPPUNIT_TEST(test_aggregateBug);
+        CPPUNIT_TEST(test_aggregateRecBug);
+        CPPUNIT_TEST(test_headAggregateBug);
+        CPPUNIT_TEST(test_symTabBug);
+        CPPUNIT_TEST(test_mutexBug);
         CPPUNIT_TEST(test_head);
         CPPUNIT_TEST(test_assign);
         CPPUNIT_TEST(test_conjunction);
@@ -53,7 +58,12 @@ public:
 
     void test_empty();
     void test_projectionBug();
+    void test_recAntiAggr();
     void test_aggregateBug();
+    void test_aggregateRecBug();
+    void test_headAggregateBug();
+    void test_symTabBug();
+    void test_mutexBug();
     void test_head();
     void test_assign();
     void test_conjunction();
@@ -100,6 +110,199 @@ void TestLparse::test_aggregateBug() {
         "b(1)."
         "{ b(X) } :- a(X).")));
 }
+void TestLparse::test_aggregateRecBug() {
+    Gringo::Test::Messages msg;
+    CPPUNIT_ASSERT_EQUAL(
+        S("[]"), 
+        IO::to_string(solve(
+        "a :- {a}!=1."
+        )));
+    CPPUNIT_ASSERT_EQUAL(
+        S("[]"), 
+        IO::to_string(solve(
+        "a :- #sum {1:a}!=1."
+        )));
+    CPPUNIT_ASSERT_EQUAL(
+        S("[[b]]"), 
+        IO::to_string(solve(
+            "b :- 0  #sum+ { 1: b }."
+        )));
+    CPPUNIT_ASSERT_EQUAL(
+        S("[[b]]"), 
+        IO::to_string(solve(
+            "b :- 0  #sum { 1: b }."
+        )));
+}
+void TestLparse::test_symTabBug() {
+    Gringo::Test::Messages msg;
+    CPPUNIT_ASSERT_EQUAL(
+        S("[[does(a,0),does(a,1)],[does(a,0),does(b,1)]]"),
+        IO::to_string(solve(
+            "time(0..1).\n"
+            "1 { does(M,T) : legal(M,T) } 1 :- time(T).\n"
+            "legal(a,T) :- time(T).\n"
+            "legal(b,T) :- does(a,0), time(T).\n", 
+            {"does"})));
+}
+void TestLparse::test_recAntiAggr() {
+    CPPUNIT_ASSERT_EQUAL(
+        S("[[p],[r]]"),
+        IO::to_string(solve(
+            "r :- #sum { 1:p } < 1.\n"
+            "p :- not r.")));
+    CPPUNIT_ASSERT_EQUAL(
+        S("[[],[p]]"),
+        IO::to_string(solve("p :- #sum { 1:not p } < 1.")));
+    CPPUNIT_ASSERT_EQUAL(
+        S("[[],[p]]"),
+        IO::to_string(solve("p :- not #sum { 1:p } < 1.")));
+    CPPUNIT_ASSERT_EQUAL(
+        S("[[],[p]]"),
+        IO::to_string(solve("p :- not #sum { 1:not p } > 0.")));
+    CPPUNIT_ASSERT_EQUAL(
+        S("[[],[p]]"),
+        IO::to_string(solve("p :- not not #sum { 1:p } > 0.")));
+    CPPUNIT_ASSERT_EQUAL(
+        S("[[],[p]]"),
+        IO::to_string(solve("p :- not not #sum { 1:not p } < 1.")));
+}
+void TestLparse::test_mutexBug() {
+#ifdef WITH_LUA
+    CPPUNIT_ASSERT_EQUAL(
+        S("[[]]"), 
+        IO::to_string(solve(
+            "#script (lua) \n"
+            "function main(prg)\n"
+            "	prg:ground(\"base\", {})\n"
+            "	prg:ground(\"plan_graph_base\", {})\n"
+            "\n"
+            "	for step = 1,3,1 do\n"
+            "		prg:ground(\"plan_graph_step\", {step})\n"
+            "		prg:solve()\n"
+            "	end\n"
+            "--	Comment line 8 and uncomment line 11\n"
+            "--	prg:solve()\n"
+            "end\n"
+            "#end.\n"
+            "\n"
+            "#program plan_graph_base.\n"
+            "valid_f(FLUENT, 1) :- init(FLUENT).\n"
+            "valid_f1(FLUENT, 1) :- init(FLUENT).\n"
+            "\n"
+            "#program plan_graph_step(time).\n"
+            "\n"
+            "%Encoding 1\n"
+            "\n"
+            "valid_a(ACT, time) :- action(ACT); valid_f(FLUENT, time) : pre(ACT, FLUENT);\n"
+            "			not mutex(F1, F2, time) : req_both(ACT, F1, F2).\n"
+            "\n"
+            "valid_pre(ACT, FLUENT, time) :- valid_a(ACT, time), pre(ACT, FLUENT).\n"
+            "valid_add(ACT, FLUENT, time) :- valid_a(ACT, time), addadd(ACT, FLUENT).\n"
+            "valid_f(FLUENT, time + 1) :- valid_add(_, FLUENT, time).\n"
+            "\n"
+            "mutex_actions(A1, A2, time) :- valid_a(A1, time), valid_a(A2, time), conflicting(A1, A2).\n"
+            "\n"
+            "mutex_a_with_f(A1, P2, time) :- mutex(P1, P2, time), valid_pre(A1, P1, time).\n"
+            "mutex_actions(A1, A2, time) :- mutex_a_with_f(A1, P2, time), valid_pre(A2, P2, time).\n"
+            "%mutex_actions(A1, A2, time) :- mutex(P1, P2, time), valid_pre(A1, P1, time), valid_pre(A2, P2, time).\n"
+            "\n"
+            "mutex_a_sym(A1, A2, time; A2, A1, time) :- mutex_actions(A1, A2, time).\n"
+            "\n"
+            "closes(A2, F1, time) :- valid_a(A2, time), valid_f(F1, time+1); mutex_a_sym(A1, A2, time) : valid_add(A1, F1, time).\n"
+            "mutex(F1, F2, time+1) :- F1 < F2, valid_f(F1, time+1), valid_f(F2, time+1);\n"
+            "			closes(A2, F1, time) : valid_add(A2, F2, time).\n"
+            "\n"
+            "% Encoding 2:\n"
+            "% Identical to encoding 1 except for lines 32 and 33 vs. line 57.\n"
+            "% However these should generate identical groundings, but they don't for some reason\n"
+            "\n"
+            "valid_a1(ACT, time) :- action(ACT); valid_f1(FLUENT, time) : pre(ACT, FLUENT);\n"
+            "			not mutex1(F1, F2, time) : req_both(ACT, F1, F2).\n"
+            "\n"
+            "valid_pre1(ACT, FLUENT, time) :- valid_a1(ACT, time), pre(ACT, FLUENT).\n"
+            "valid_add1(ACT, FLUENT, time) :- valid_a1(ACT, time), addadd(ACT, FLUENT).\n"
+            "valid_f1(FLUENT, time + 1) :- valid_add1(_, FLUENT, time).\n"
+            "\n"
+            "mutex_actions1(A1, A2, time) :- valid_a1(A1, time), valid_a1(A2, time), conflicting(A1, A2).\n"
+            "\n"
+            "%mutex_a_with_f1(A1, P2, time) :- mutex1(P1, P2, time), valid_pre1(A1, P1, time).\n"
+            "%mutex_actions1(A1, A2, time) :- mutex_a_with_f1(A1, P2, time), valid_pre1(A2, P2, time).\n"
+            "mutex_actions1(A1, A2, time) :- mutex1(P1, P2, time), valid_pre1(A1, P1, time), valid_pre1(A2, P2, time).\n"
+            "\n"
+            "mutex_a_sym1(A1, A2, time; A2, A1, time) :- mutex_actions1(A1, A2, time).\n"
+            "\n"
+            "closes1(A2, F1, time) :- valid_a1(A2, time), valid_f1(F1, time+1); mutex_a_sym1(A1, A2, time) : valid_add1(A1, F1, time).\n"
+            "mutex1(F1, F2, time+1) :- F1 < F2, valid_f1(F1, time+1), valid_f1(F2, time+1);\n"
+            "			closes1(A2, F1, time) : valid_add1(A2, F2, time).\n"
+            "\n"
+            "% diff reports if the two encodings are not the same\n"
+            "\n"
+            "diff(A1, A2, time) :- mutex_actions1(A1, A2, time), not mutex_actions(A1, A2, time).\n"
+            "diff(A1, A2, time) :- mutex_actions(A1, A2, time), not mutex_actions1(A1, A2, time).\n"
+            "\n"
+            "#program base.\n"
+            "\n"
+            "#show diff/3.\n"
+            "\n"
+            "preserve_action(preserve(F)) :- fact(F).\n"
+            "action(A) :- preserve_action(A).\n"
+            "pre(preserve(F), F) :- fact(F).\n"
+            "addadd(preserve(F), F) :- fact(F).\n"
+            "\n"
+            "prepre(A, F) :- pre(A, F), not del(A, F).\n"
+            "deldel(A, F) :- del(A, F), not pre(A, F).\n"
+            "predel(A, F) :- pre(A, F), del(A, F).\n"
+            "addadd(A, F) :- add(A, F), not pre(A, F), not del(A, F).\n"
+            "\n"
+            "conflicting(A1, A2) :- pre(A1, F), del(A2, F), A1 != A2.\n"
+            "\n"
+            "req_both(ACT, F1, F2) :- pre(ACT, F1), pre(ACT, F2), F1 < F2.\n"
+            "\n"
+            "type(object).\n"
+            "type(ferry, object).\n"
+            "fact(at_ferry(X)) :- type(X, object).\n"
+            "fact(at(X, Y)) :- type(X, object), type(Y, object).\n"
+            "fact(on(X, Y)) :- type(X, object), type(Y, object).\n"
+            "fact(empty_ferry).\n"
+            "action(debark(X, Y)) :- type(X, object), type(Y, object), auto__(X), place__(Y).\n"
+            "pre(debark(X, Y), on(X, ferry)) :- action(debark(X, Y)).\n"
+            "pre(debark(X, Y), at_ferry(Y)) :- action(debark(X, Y)).\n"
+            "add(debark(X, Y), at(X, Y)) :- action(debark(X, Y)).\n"
+            "add(debark(X, Y), empty_ferry) :- action(debark(X, Y)).\n"
+            "del(debark(X, Y), on(X, ferry)) :- action(debark(X, Y)).\n"
+            "action(sail(X, Y)) :- type(X, object), type(Y, object), place__(X), place__(Y).\n"
+            "pre(sail(X, Y), at_ferry(X)) :- action(sail(X, Y)).\n"
+            "add(sail(X, Y), at_ferry(Y)) :- action(sail(X, Y)).\n"
+            "del(sail(X, Y), at_ferry(X)) :- action(sail(X, Y)).\n"
+            "action(board(X, Y)) :- type(X, object), type(Y, object), place__(Y), auto__(X).\n"
+            "pre(board(X, Y), at(X, Y)) :- action(board(X, Y)).\n"
+            "pre(board(X, Y), at_ferry(Y)) :- action(board(X, Y)).\n"
+            "pre(board(X, Y), empty_ferry) :- action(board(X, Y)).\n"
+            "add(board(X, Y), on(X, ferry)) :- action(board(X, Y)).\n"
+            "del(board(X, Y), at(X, Y)) :- action(board(X, Y)).\n"
+            "del(board(X, Y), empty_ferry) :- action(board(X, Y)).\n"
+            "type(c1, object).\n"
+            "type(b, object).\n"
+            "type(c2, object).\n"
+            "type(a, object).\n"
+            "place__(a).\n"
+            "place__(b).\n"
+            "auto__(c1).\n"
+            "auto__(c2).\n"
+            "init(at(c1, a)).\n"
+            "init(at(c2, a)).\n"
+            "init(at_ferry(a)).\n"
+            "init(empty_ferry).\n"
+            "goal(at(c1, b)).\n"
+            "goal(at(c2, b)).\n"
+            )));
+#endif // WITH_LUA
+}
+void TestLparse::test_headAggregateBug() {
+    Gringo::Test::Messages msg;
+    CPPUNIT_ASSERT_EQUAL(S("[[q(a),r(a)]]"), IO::to_string(solve("1 { q(a); p(X) : z(X) }. r(X) :- q(X).")));
+    CPPUNIT_ASSERT_EQUAL(S("[[q(a),r(a)]]"), IO::to_string(solve("1 { p(X) : z(X); q(a) }. r(X) :- q(X).")));
+}
 void TestLparse::test_head() {
     CPPUNIT_ASSERT_EQUAL(S("[[],[a],[a,b],[b]]"), IO::to_string(solve("{a;b}.")));
     CPPUNIT_ASSERT_EQUAL(S("[[a],[b]]"), IO::to_string(solve("1{a;b}1.")));
