@@ -174,12 +174,13 @@ private:
 	typedef OptionGroup::OptionList                   OptionList;
 public:
 	//! Type for identifying an option within a context
-	typedef OptionList::const_iterator      option_iterator;
+	typedef OptionList::const_iterator option_iterator;
+	typedef PrefixRange                OptionRange;  
 	
 	OptionContext(const std::string& caption = "", DescriptionLevel desc_default = desc_level_default);
 	~OptionContext();
 
-	const std::string& caption() const { return caption_; }
+	const std::string& caption() const;
 	
 	//! Adds the given group of options to this context.
 	/*!
@@ -233,6 +234,9 @@ public:
 	 */
 	option_iterator tryFind(const char* key, FindType t = find_name) const;
 
+	OptionRange findImpl(const char* key, FindType t, unsigned eMask = unsigned(-1)) const { return findImpl(key, t, eMask, caption()); }
+	OptionRange findImpl(const char* key, FindType t, unsigned eMask, const std::string& eCtx) const;
+
 	const OptionGroup& findGroup(const std::string& caption) const;
 	const OptionGroup* tryFindGroup(const std::string& caption) const;
 	
@@ -259,7 +263,6 @@ public:
 	 */
 	bool assignDefaults(const ParsedOptions& exclude) const;
 private:
-	PrefixRange findImpl(const char* key, FindType t) const;
 	void        insertOption(size_t groupId, const SharedOptPtr& o);
 	size_t      findGroupKey(const std::string& name) const;
 	
@@ -281,6 +284,7 @@ public:
 	bool        empty()                        const  { return parsed_.empty(); }
 	std::size_t size()                         const  { return parsed_.size(); }
 	std::size_t count(const std::string& name) const  { return parsed_.count(name); }
+	void        add(const std::string& name)          { parsed_.insert(name); }
 	
 	//! Assigns the parsed values in p to their options.
 	/*!
@@ -294,7 +298,7 @@ public:
 	 *        for a non-composing option or if p contains a value that is 
 	 *        invalid for its option.
 	 */
-	bool        assign(const ParsedValues& p);
+	bool        assign(const ParsedValues& p, const ParsedOptions* exclude = 0);
 private:
 	std::set<std::string> parsed_;
 	int assign(const Option& o, const std::string& value);
@@ -319,37 +323,42 @@ public:
 	
 	//! Adds a value for option opt.
 	void add(const std::string& opt, const std::string& value);
+	void add(const SharedOptPtr& opt, const std::string& value) {
+		parsed_.push_back(OptionAndValue(opt, value));
+	}
 
 	iterator begin() const { return parsed_.begin(); }
 	iterator end()   const { return parsed_.end(); }
 
 	void clear() { parsed_.clear(); }
 private:
-	friend class OptionParser;
-	void add(const SharedOptPtr& opt, const std::string& value) {
-		parsed_.push_back(OptionAndValue(opt, value));
-	}
 	Values parsed_;
 };
 
+class ParseContext {
+public:
+	typedef OptionContext::FindType FindType;
+	virtual ~ParseContext();
+	virtual SharedOptPtr getOption(const char* name, FindType ft)                    = 0;
+	virtual SharedOptPtr getOption(int posKey, const char* tok)                      = 0;
+	virtual void         addValue(const SharedOptPtr& key, const std::string& value) = 0;
+};
+
 //! Base class for options parsers.
 class OptionParser {
 public:
 	typedef OptionContext::FindType FindType;
-	OptionParser(const OptionContext& o, bool allowUnreg);
+	explicit OptionParser(ParseContext& ctx);
 	virtual ~OptionParser();
-	ParsedValues parse();
+	ParseContext& parse();
 protected:
-	// PRE: key.get() != 0
-	void addOptionValue(const SharedOptPtr& key, const std::string& value) {
-		pv_.add(key, value);
-	}
-	SharedOptPtr getOption(const char* name, FindType ft) const;
-	SharedOptPtr tryGetOption(const char* name, FindType ft) const;
+	ParseContext& ctx() const { return *ctx_; }
+	SharedOptPtr  getOption(const char* name, FindType ft) const { return ctx_->getOption(name, ft); }
+	SharedOptPtr  getOption(int posKey, const char* tok)   const { return ctx_->getOption(posKey, tok); }
+	void          addOptionValue(const SharedOptPtr& key, const std::string& value) { ctx_->addValue(key, value); } 
 private:
 	virtual void doParse() = 0;
-	ParsedValues pv_;
-	bool allowUnreg_; 
+	ParseContext* ctx_;
 };
 
 //! Default formatting for options.
@@ -436,6 +445,10 @@ typedef OptionOutputImpl<FileWriter>    FileOut;
  */
 typedef bool (*PosOption)(const std::string&, std::string&);
 
+enum CommandLineFlags {
+	command_line_allow_flag_value = 1u,
+};
+
 /*!
 * Parses the command line starting at index 1 and removes
 * all found options from argv.
@@ -453,7 +466,9 @@ typedef bool (*PosOption)(const std::string&, std::string&);
 */
 ParsedValues parseCommandLine(int& argc, char** argv, const OptionContext& ctx,
 	bool allowUnregistered = true,
-	PosOption posParser = 0);
+	PosOption posParser = 0, unsigned flags = 0);
+
+ParseContext& parseCommandLine(int& argc, char** argv, ParseContext& ctx, unsigned flags = 0); 
 
 /*!
 * Parses the command line given in the first parameter.
@@ -467,7 +482,8 @@ ParsedValues parseCommandLine(int& argc, char** argv, const OptionContext& ctx,
 * \throw SyntaxError if command line syntax is incorrect.
 * \throw UnknownOption if an argument is found that does not match any option.
 */
-ParsedValues parseCommandString(const std::string& cmd, const OptionContext& ctx, bool allowUnreg = false, PosOption posParser = 0);
+ParsedValues parseCommandString(const std::string& cmd, const OptionContext& ctx, bool allowUnreg = false, PosOption posParser = 0, unsigned flags = command_line_allow_flag_value);
+ParseContext& parseCommandString(const char* cmd, ParseContext& ctx, unsigned flags = command_line_allow_flag_value); 
 
 /*!
 * Parses a config file having the format key = value.
